# pkcs lib

##############################################################################
# extension: PKCS management
# Author: Chris Fraire 2016, 2018
##############################################################################

use strict;
use Expect;
use File::Basename;

##############################################################################
# inputs: none.
# outputs: a defined directory if there is a napp-it keystore; else undef.
sub has_pkcs_nappit_keystore {
    my $pkcsdir = _get_napp_it_keystore_folder();
    return has_pkcs_keystore($pkcsdir);
}

##############################################################################
# inputs: a specified directory.
# outputs: the specified directory if it contains a keystore; else undef.
sub has_pkcs_keystore {
    my ($pkcsdir) = @_;
    $pkcsdir or die "Bad args--stopped";

    my $keystoredir = &pkcs_set_softtoken_dir($pkcsdir);
    &exe("/bin/test -e ". quotemeta("$keystoredir/pkcs11_softtoken"));
    my $ex = !$?;
#    print "Is a PKCS keystore in ", $pkcsdir, "? ", $ex ? "Yes" : "No", ".<br>\n";
    return $ex ? $pkcsdir : undef;
}

##############################################################################
# inputs: none.
# outputs: none.
# environment: sets SOFTTOKEN_DIR to the static path of napp-it keystore
#     directory, whether or not it has been initialized.
sub pkcs_set_nappit_softtoken_dir {
    my $pkcsdir = _get_napp_it_keystore_folder();
    return &pkcs_set_softtoken_dir($pkcsdir);
}

##############################################################################
# inputs: a specified directory.
# outputs: none.
# environment: sets SOFTTOKEN_DIR to the specified directory, whether or not
#     it has been initialized with a PKCS keystore.
sub pkcs_set_softtoken_dir {
    my ($pkcsdir) = @_;
    $pkcsdir or die "Bad args--stopped";

    my $keystoredir = "$pkcsdir/pkcs";
    $ENV{'SOFTTOKEN_DIR'} = $keystoredir;
    return $keystoredir;
}

##############################################################################
# inputs: a password used to initialize the napp-it PKCS keystore.
# outputs: 1 if the keystore was successfully initialized; else undef.
# environment: sets SOFTTOKEN_DIR to the static path of napp-it keystore
#     directory.
sub pkcs_nappit_keystore_init {
    my ($password) = @_;
    $password or die "Bad args--stopped";

    my $pkcsdir = _get_napp_it_keystore_folder();
    return pkcs_keystore_init($pkcsdir, $password);
}

##############################################################################
# inputs: a specified directory and a password used to initialize the
#     PKCS keystore.
# outputs: 1 if the keystore was successfully initialized; else undef.
# environment: sets SOFTTOKEN_DIR to the specified directory.
sub pkcs_keystore_init {
    my ($pkcsdir, $password) = @_;
    ($pkcsdir && $password) or die "Bad args--stopped";

    print "initialize PKCS keystore in $pkcsdir<br>\n";

    if (has_pkcs_keystore($pkcsdir)) {
        print "PKCS keystore already initialized!<br>\n";
        return 1;
    }

    my $keystoredir = &pkcs_set_softtoken_dir($pkcsdir);
    -d $keystoredir or do {
        print "create directory $keystoredir<br>\n";
        &exe("/bin/mkdir -m 700 ". quotemeta($keystoredir));
        return if $?;
    };

    # "changeme" is the illumos default initial passphrase
    return pkcs_keystore_passwd($pkcsdir, $password, "changeme");
}

##############################################################################
# inputs: a current password and a new password.
# outputs: 1 if the napp-it keystore password was successfully changed; else
#     undef.
# environment: sets SOFTTOKEN_DIR to the static path of napp-it keystore
#     directory.
sub pkcs_nappit_keystore_passwd {
    my ($password, $oldpassword) = @_;
    ($password && $oldpassword) or die "Bad args--stopped";

    my $pkcsdir = _get_napp_it_keystore_folder();
    return pkcs_keystore_passwd($pkcsdir, $password, $oldpassword);
}

##############################################################################
# inputs: a specified directory, a current password and a new password.
# outputs: 1 if the keystore password was successfully changed; else undef.
# environment: sets SOFTTOKEN_DIR to the specified directory.
sub pkcs_keystore_passwd {
    my ($pkcsdir, $password, $oldpassword) = @_;
    ($pkcsdir && $password && $oldpassword) or die "Bad args--stopped";

    print "change password for PKCS keystore in $pkcsdir<br>\n";

    &pkcs_set_softtoken_dir($pkcsdir);

    my $do = "/usr/bin/pktool setpin keystore=pkcs11";

    my $exp = new Expect;
    $exp->log_stdout(0);
    $exp = Expect->spawn("pfexec $do") or die "Couldn't setpin for keystore: $!\n";

    my $exec_error = 0;
    $exp->expect(3,
        [ 'Enter token passphrase:' =>
                 sub {
                     $exp->send("$oldpassword\r");
                     print "****<br>";
                     exp_continue;
                 }
        ],
        [ 'new passphrase:' =>
                 sub {
                     $exp->send("$password\r");
                     print "****<br>";
                     exp_continue;
                 }
        ],
        [ 'Cannot exec' =>
                 sub {
                     $exec_error = 1;
                 }
        ]
    );
    $exp->soft_close();
    if ($exec_error || $exp->exitstatus()) {
        print "setpin failed!<br>\n" if $exec_error;
        return;
    }

    return 1;
}

##############################################################################
# inputs: a password and a key label.
# outputs: undef if failed to check for key; otherwise 1 if key exists; or
#     0 otherwise.
# environment: sets SOFTTOKEN_DIR to the static path of napp-it keystore
#     directory.
sub has_pkcs_nappit_keystore_key {
    my ($password, $keylabel) = @_;
    ($password && $keylabel) or die "Bad args--stopped";

    my $pkcsdir = _get_napp_it_keystore_folder();
    return has_pkcs_keystore_key($pkcsdir, $password, $keylabel);
}

##############################################################################
# inputs: a specified directory, a password and a key label.
# outputs: undef if failed to check for key; otherwise 1 if key exists; or
#     0 otherwise.
# environment: sets SOFTTOKEN_DIR to the specified directory.
sub has_pkcs_keystore_key {
    my ($pkcsdir, $password, $keylabel) = @_;
    ($pkcsdir && $password && $keylabel) or die "Bad args--stopped";

    &pkcs_set_softtoken_dir($pkcsdir);

    print "look for $keylabel in PKCS keystore in $pkcsdir<br>\n";
    my $do = "/usr/bin/pktool list label=" . quotemeta($keylabel) .
        " objtype=both";

    my $exp = new Expect;
    $exp->log_stdout(0);
    $exp = Expect->spawn("pfexec $do") or die "Couldn't list from keystore: $!\n";

    my $exec_error = 0;
    my $found_keys = 0;
    my $symmetric_keys_hdr = 0;
    $exp->expect(3,
        # note: the prompt for pktool and lofiadm are slightly different w.r.t. spacing
        [ qr/Enter PIN for (.*? softtoken|Sun Metaslot)\s*:\s*/ =>
                 sub {
                     $exp->send("$password\r");
                     print "****<br>";
                     exp_continue;
                 }
        ],
        [ qr/Found \d+ symmetric key/ =>
                 sub {
                     $exp->match() =~ /Found (\d+) symmetric key/;
                     $found_keys = $1;
                 }
        ],
        [ qr/Symmetric keys:/ =>
                 sub {
                     $symmetric_keys_hdr = 1;
                 }
        ],
        [ qr/\d+ \) \s+ AES \s+ \d+ \s+ \Q $keylabel \E\b/x =>
                 sub {
                     $exp->match() =~ /(\d+) \) \s+ AES \s+/x;
                     $found_keys = $symmetric_keys_hdr ? $1 : 0;
                 }
        ],
        [ 'Cannot exec' =>
                 sub {
                     $exec_error = 1;
                 }
        ]
    );
    $exp->soft_close();
    if ($exec_error || $exp->exitstatus()) {
        print "Couldn't list from keystore!<br>\n";
        return;
    }
    return $found_keys > 0;
}

##############################################################################
# inputs: a password and a key label.
# outputs: 1 if a key for the specified label is successfully generated; else
#     undef.
# environment: sets SOFTTOKEN_DIR to the static path of napp-it keystore
#     directory.
sub pkcs_nappit_keystore_genkey {
    my ($password, $keylabel) = @_;
    ($password && $keylabel) or die "Bad args--stopped";

    my $pkcsdir = _get_napp_it_keystore_folder();
    return pkcs_keystore_genkey($pkcsdir, $password, $keylabel);
}

##############################################################################
# inputs: a specified directory, a password and a key label.
# outputs: 1 if a key for the specified label is successfully generated; else
#     undef.
# environment: sets SOFTTOKEN_DIR to the specified directory.
sub pkcs_keystore_genkey {
    my ($pkcsdir, $password, $keylabel) = @_;
    ($pkcsdir && $password && $keylabel) or die "Bad args--stopped";

    &pkcs_set_softtoken_dir($pkcsdir);

    my $has_key = has_pkcs_keystore_key($pkcsdir, $password, $keylabel);
    return if ! defined $has_key;
    return 1 if $has_key;

    # have to use sensitive=n or else lofiadm will not be able to use it
    print "generate $keylabel for PKCS keystore in $pkcsdir<br>\n";
    my $do = "/usr/bin/pktool genkey label=" . quotemeta($keylabel) .
        " keytype=aes keylen=256 sensitive=n extractable=y";

    my $exp = new Expect;
    $exp->log_stdout(0);
    $exp = Expect->spawn("pfexec $do") or die "Couldn't genkey in keystore: $!\n";

    my $exec_error = 0;
    $exp->expect(3,
        [ qr/Enter PIN for (.*? softtoken|Sun Metaslot):\s*/ =>
                 sub {
                     $exp->send("$password\r");
                     print "****<br>";
                     exp_continue;
                 }
        ],
        [ 'Cannot exec' =>
                 sub {
                     $exec_error = 1;
                 }
        ]
    );
    $exp->soft_close();
    if ($exec_error || $exp->exitstatus()) {
        print "Couldn't genkey in keystore!<br>\n";
        return;
    }

    return 1;
}

sub _get_napp_it_keystore_folder {
    my $pkcsdir = "/var/web-gui/_log";
    return $pkcsdir;
}

###############################################
#end of my script, do not delete the line below
1;
