# poef lib

##############################################################################
# extension: PoeF management
# (c) napp-it 2016 from original in "01_create pool oef/action.pl".
# Extended by Chris Fraire 2016 for PKCS keystore and extending pools.
##############################################################################

use strict;
use Expect;
use POSIX;

##############################################################################
# return: 1 on success; 0 on failure.
sub lofi_unmap {
    my ($dev_or_file) = @_;
    $dev_or_file or die "Bad args--stopped";

    print "remove device for $dev_or_file<br>\n";
    my $do = "/usr/sbin/lofiadm -d " . quotemeta($dev_or_file);
    &exe($do);
    return !$?;
}

##############################################################################
# return: device name on success; undef on failure.
sub lofi_map {
    my ($file, $encryption, $password, $keylabel, $pkcsdir) = @_;
    ($file && $encryption && $password) or die "Bad args--stopped";

    # create encrypted devices lofiadm -c $encryption -a /tank/secrets/file1
    # -> password password. If $keylabel is defined, then lofiadm will use a
    # PKCS keystore; otherwise the password is used directly, and it cannot
    # yet be verified as correct until a later zpool operation.

    print "create device for $file<br>\n";
    my $do = "/usr/sbin/lofiadm -c \"$encryption\" -a \"$file\"";
    if ($keylabel) {
        $do .= " -T :::\"$keylabel\"";
        if ($pkcsdir) {
            pkcs_set_softtoken_dir($pkcsdir);
        }
        else {
            pkcs_set_nappit_softtoken_dir();
        }
    }

    my $dev;
    my $exp;   #->raw_pty(1);
    $exp = Expect->spawn("pfexec $do") or die "Couldn't create device: $!\n";

    my $lofi_error;
    $exp->expect(3,
        [ 'passphrase:' =>
                 sub {
                     $exp->send("$password\r");
                     print "****<br>";
                     exp_continue;
                 }
        ],
        # note: the prompt for lofiadm and pktool are slightly different w.r.t. spacing
        [ qr/Enter PIN for (.*? softtoken|Sun Metaslot)\s*:\s*/ =>
                 sub {
                     $exp->send("$password\r");
                     print "****<br>";
                     exp_continue;
                 }
        ],
        [ 'again' =>
                 sub {
                     $exp->send("$password\r");
                     print "****<br>";
                     exp_continue;
                 }
        ],
        [ 'Passphrase must be at least' =>
                 sub {
                     $lofi_error = "Password too short";
                     $exp->send("\cC");
                 }
        ],
        [ qr`/dev/lofi/\d+\r?\n`,
                 sub {
                     $exp->match() =~ m`(/dev/lofi/\d+)`;
                     $dev = $1;
                 }
        ],
        [ 'Cannot exec' =>
                 sub {
                     $lofi_error = "Cannot exec";
                 }
        ],
        [ 'unable to find any cryptographic provider' =>
                 sub {
                     $lofi_error = "No cryptographic provider found";
                 }
        ]
    );
    $exp->soft_close();
    print "<br><br>\n";

    print "lofiadm failed: $lofi_error<br>\n" if $lofi_error;
    print "Did not get lofi device number!<br>\n" if !$dev;
    return $dev;
}

##############################################################################
# return: 1 on success; 0 on failure.
sub lofi_map_pool {
    my ($pool, $folder, $encryption, $password, $keylabel) = @_;
    ($pool && $folder && $encryption && $password) or die "Bad args--stopped";

    # 1. confirm that $pool is not already advertised in /dev/lofi
    if (is_poef_importable($pool)) {
        print "lofi devices for $pool are already attached,",
            " which blocks a password check!<br>\n";
        return;
    }

    # 2. lofi_add files from a directory
    # get files
    my $f = &exe("find " . quotemeta($folder) . " -type f");
    my @files = sort {
        $a =~ m`/(\d+)$`;    my $annnnn = $1;
        $b =~ m`/(\d+)$`;    my $bnnnnn = $1;
        $annnnn <=> $bnnnnn;
    } grep { m`/\d+$` } split(/\n/, $f);

    if (@files ==0) {
        print "no datafiles found<br>\n";
        return;
    }

    my @devs = ();
    my $add_ok = 1;
    print "<div style='position: relative; height: 30px; width: 100%; overflow:auto'>";
    foreach $f (@files) {
        my $dev = lofi_map($f, $encryption, $password, $keylabel);
        if ($dev) {
            push @devs, $dev;
        } else {
           $add_ok = 0;
           last;
        }
    }
    print "</div>";

    if (!$add_ok) {
        foreach my $dev (@devs) { lofi_unmap($dev); }
        return;
    }

    # 3. zpool-import just to see if an expected pool advertises itself in /dev/lofi
    #    afterward. If not, lofi_unmap; otherwise, leave lofi added, to be imported
    #    subsequently
    if (!is_poef_importable($pool)) {
        print "$pool does not show as importable.",
            " Check password and file integrity.<br>\n";
        foreach my $dev (@devs) { lofi_unmap($dev); }
        return;
    }

    return 1;
}

##############################################################################
# return: 1 on success; 0 on failure.
sub lofi_unmap_files {
    my ($pool) = @_;
    $pool or die "Bad args--stopped";

    # get devices to detach
    my $d = &exe("/usr/sbin/lofiadm");
    my @devices = map { s/\s.*//; $_ } grep { m`/\Q$pool\E/` } split(/\n/, $d);

    return lofi_unmap_devices(@devices);
}

##############################################################################
# return: 1 on success; 0 on failure.
sub lofi_unmap_devices {
    my (@devices) = @_;

    my $cmd = "rc=0; ";
    foreach my $d (@devices) {
        $cmd .= "/usr/sbin/lofiadm -d " . quotemeta($d) . " || rc=1; ";
    }
    $cmd .= 'exit $rc';

    &exe("/bin/sh -c " . quotemeta($cmd));
    return !$?;
}

##############################################################################
# return: 1 on success; 0 on failure.
sub poef_import {
    my ($pool) = @_;
    $pool or die "Bad args--stopped";

    print "import pool $pool<br>\n";
    my $do = "/usr/sbin/zpool import -f -d /dev/lofi " . quotemeta($pool);
    my $t = &exe($do);   # return stderr
    my $rc = !$?;

    if (!$rc) { lofi_unmap_files($pool); }

    #reload
    if (-f '/tmp/nappit/zfs.cfg') { unlink('/tmp/nappit/zfs.cfg'); }

    return $rc;
}

##############################################################################
# return: 1 on success; 0 on failure.
sub poef_export {
    my ($pool) = @_;
    $pool or die "Bad args--stopped";

    print "export pool $pool<br>\n";
    my $do = "/usr/sbin/zpool export -f " . quotemeta($pool);
    my $t = &exe($do);   # return stderr
    my $rc = !$?;

    lofi_unmap_files($pool);

    #reload
    if (-f '/tmp/nappit/zfs.cfg') { unlink('/tmp/nappit/zfs.cfg'); }

    return $rc;
}

##############################################################################
# return: 1 if importable; 0 otherwise.
sub is_poef_importable {
    my ($pool) = @_;
    $pool or die "Bad args--stopped";

    print "check if zpool-import shows $pool as exported and ONLINE<br>\n";
    my $do = "/usr/sbin/zpool import -d /dev/lofi";
    my $t = &exe($do);

    return if $?;

    # Extract from output e.g., :
    #    pool: foobar
    #      id: 1873403973325425730
    #   state: ONLINE
    #  action: The pool can be imported using its name or numeric identifier.
    $t =~ /^(\s* pool:\s* \Q$pool\E
            (?:.|\n)+?
            action: .+)/mx or return 0;
    my $m = $1;
    $m =~ s/\n/<br>\n/g;
    print $m, "<br>\n";

    return $m =~ /state:\s* ONLINE/x;
}

##############################################################################
# Return an array of arrays to represent vdevs and their initial device stat,
# which is its filesize.
# E.g., [["32g"]] is a one, non-redundant vdev of one 32g file.
# E.g., [["2g", "2g", "2g"], ["2g", "2g", "2g"]] is two z2 vdevs with three 32g files.
sub poef_calc_vdevs {
    my ($poolsize, $sizespec) = @_;
    ($poolsize && $sizespec && $poolsize =~ /\dGB?$/i) or die "Bad args--stopped";

    $poolsize =~ s/GB$/g/i;
    my $filesize;

    # Basic Pool, single file
    if ($sizespec =~ /Basic/) {
        $filesize = $poolsize;
        return [[$filesize]];
    }

    # raidz2 with 2GB file size
    if ($sizespec =~ /^2G.+Z2/i) {
        $filesize = "2g";

        # calculate for even-disk vdev allocations of a set number of sizes
        my @evenly = ();
        foreach my $ndisk ((4, 6, 10)) {
            my $nvdev = ceil($poolsize / ($ndisk * 2));
            my $vgigs = $nvdev * $ndisk * 2;
            push @evenly, [$vgigs, $nvdev, $ndisk];
        };

        # order to find the best vdev size according to two criteria:
        @evenly = sort {
            # first sort by amount of excessive data space (ASC)
            my $cmp = $a->[0] - $poolsize <=> $b->[0] - $poolsize;
            # next possibly sort by size of vdev (DESC)
            if (!$cmp) { $cmp = $b->[2] <=> $a->[2]; }
            $cmp;
        } @evenly;

        # Size the @vdevs arrays as calculated (with undef contents)
        my @vdevs = ();
        my ($vgigs, $nvdev, $ndisk) = @{$evenly[0]};
        for (my $i = 0; $i < $nvdev; ++$i) {
            my @vdev = ($filesize) x ($ndisk + 2); # add 2 for raidz2
            push @vdevs, \@vdev;
        }
        return \@vdevs;
    }

    # 10% RAID-Z2; except 5GB is explicit 1GB filesize
    if ($sizespec =~ /^10%.*Z2/i) {
        $filesize = "1g";
        my $files = 5;                                  # number of datafiles on a 5GB pool = 5 x 1g
        if ($poolsize > 5) {
            $filesize = ceil($poolsize/10 * 1024) . "m";# use MB which rounds better--12GB * 10% (1229m)
            $files = 10;                                # number of datafiles
        }

        my @vdev = ($filesize) x ($files + 2);  # add 2 for raidz2
        return [\@vdev];
    }

    die "Unknown sizespec $sizespec!";
}

##############################################################################
# Return: a defined hashref on success; or undef on error.
sub poef_get_pooloptions {
   my ($folder) = @_;
    $folder or die "Bad args--stopped";

   # NOTE double quotemeta because of pfexec running sh
   my $qconf = quotemeta("$folder/pooloptions.txt");
   open(PF, "pfexec /bin/sh -c " .
       quotemeta("/bin/test ! -f $qconf || /bin/cat $qconf") . " |") or
       return;

   my (%opts, $line);
   while (defined($line = <PF>)) {
       chomp $line;
       $line =~ /^([^=]+)=(.+)/ && do { $opts{$1} = $2; };
   }
   close(PF) or return undef;

   return \%opts;
}


###############################################
#end of my script, do not delete the line below
1;
