#!/usr/bin/perl
#
# VMware_SOAP Perl script / 2023-02-26 / Ver. 1.04
# 
# https://www.hardwareluxx.de/community/members/cluster_.112838/
# https://www.hardwareluxx.de/community/threads/esx-esxi-hilfethread.748644/post-29758517
#
# Changelog
# ver. 1.00: 2023-02-18 / Initial
# ver. 1.01: 2023-02-19 / New "summary" action + Switch to moid for "list_attached_vms"
# ver. 1.02: 2023-02-21 / New actions (poweron/reboot/kill) in addition to option scope (host/vm/guest), snapshot options now partially optional, added helptexts for actions
# ver. 1.03: 2023-02-25 / Changed the way colorcoding is displayed
# ver. 1.04: 2023-02-26 / Added "plain" modifier to suppress colorized output
#
# Install required linux packages
#
# sudo apt install make build-essential libssl-dev libnet-ssleay-perl libcrypt-ssleay-perl
#
# Install required perl modules
#
# sudo cpan URI Tie::Hash HTTP::Cookies warnings::register utf8 Socket YAML::Node Net::Ping Carp LWP::MemberMixin URI::_punycode LWP::Protocol::https LWP Net::HTTPS HTTP::Response FileHandle warnings URI::_idna Data::Dumper Getopt::Long URI::_server Net::SSLeay IO::Socket::SSL::PublicSuffix IO::Compress::Base::Common IO::Socket::IP File::Spec Fcntl IO::Uncompress::Adapter::Inflate vars URI::_generic File::Spec::Unix SelectSaver Switch IO::Handle overload Filter::Util::Call HTTP::Headers::Util YAML::Mo Net::HTTP::Methods File::Copy IO::Socket::INET LWP::Protocol Exporter::Heavy HTTP::Headers IO::Compress::Zlib::Extra HTTP::Message IO::File parent Text::Balanced URI::https IO::Seekable IO::Uncompress::Gunzip deprecate Exporter LWP::UserAgent POSIX Compress::Raw::Zlib HTTP::Status overloading integer IO::Uncompress::RawInflate YAML bytes base HTTP::Config Clone URI::http URI::_query Config HTTP::Cookies::Netscape Module::Load Errno Net::HTTP IO::Socket::UNIX Cwd IO::Uncompress::Base LWP::Protocol::http Try::Tiny IO::Compress::Gzip::Constants IO::Socket::SSL::Utils Symbol IO::Socket::SSL File::GlobMapper URI::Escape Sub::Util Time::Local Time::HiRes Scalar::Util AutoLoader XSLoader Term::ANSIColor HTTP::Request HTTP::Date File::Glob strict constant
#
# Include Perl modules
#
use strict;
use warnings;
use POSIX qw(strftime);
use Switch;
use Net::Ping;
use Term::ANSIColor;
use LWP::UserAgent;
use HTTP::Request;
use HTTP::Cookies;
use Getopt::Long;
Getopt::Long::Configure("pass_through");
#
# Variabales
#
my $datetime = strftime('%Y-%m-%d %H:%M:%S ',gmtime());
my @implemented_actions = ('summary','ssh_on','ssh_off','poweron','shutdown','reboot','kill','mount','unmount','create_snapshot','remove_snapshot','last_snapshot','revert_snapshot','list_attached_vms','list_all_datastores');
my @mount_parameters = ('mounthost','mountpoint','mounttype','mountaccess','localpath','mountuser','mountpassword');
my @unmount_parameters = ('mounthost','mountpoint');
my @last_snapshot_parameters = ('vm_id');
my @create_snapshot_parameters = ('vm_id','snapname','snapdesc','mem','quiesce');
my @remove_snapshot_parameters = ('snapid','childs');
my @revert_snapshot_parameters = ('snapid','suppress');
my @list_attached_vms_parameters = ('moid');
my @poweron_parameters = ('scope','vm_id');
my @shutdown_parameters = ('scope','vm_id');
my @reboot_parameters = ('scope','vm_id');
my @kill_parameters = ('scope','vm_id');
my @host_parameters = ('host','user','password','plain');
our ($verbose,$debug) = (0,0);	# global option variable with default value (false)
my $help = 0;					# option variable with default value (false)
my $todieornot = 0;
my ($host,$user,$password) = ('','','');
my ($mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword) = ('','','','','readWrite','','','');
my ($vm_id,$snapname,$snapid,$snapdesc,$suppress,$mem,$quiesce,$childs) = ('','','','',0,1,0,0);
my ($scope) = ('dummy');
my ($plain) = (0);
my ($request,$message,$response,$retval,$cookie);
my %action_helper = (
	"summary", 				"Prints summary incl. states of attached datastores and vms for host\n".
							"Datastore shows status / VM shows heartbeat state\n\n".
							"Datastore health color coding:\n".
							color('red').    "⬛ ".color('reset')."red     = Critial event\n".
							color('ansi202')."⬛ ".color('reset')."orange  = Major event\n".
							color('yellow'). "⬛ ".color('reset')."yellow  = Minor event\n".
							color('ansi14'). "⬛ ".color('reset')."blue    = Unknown\n".
							color('green').  "⬛ ".color('reset')."green   = Normal/oaky\n\n".
							"VM heartbeat color coding:\n".
							color('ansi60'). "⬛ ".color('reset')."grey    = VMware Tools are not installed or not running\n".
							color('red').    "⬛ ".color('reset')."red     = No heartbeat. Guest operating system may have stopped responding\n".
							color('yellow'). "⬛ ".color('reset')."yellow  = Intermittent heartbeat. A Yellow status may be caused by heavy guest OS usage\n".
							color('green').  "⬛ ".color('reset')."green   = Guest operating system is responding normally\n",
	"ssh_on", 				"Enables SSH access on host (Service name 'TSM-SSH')",
	"ssh_off", 				"Disables SSH access on host (Service name 'TSM-SSH')",
	"poweron",				"scope vm:    Powers on 'vm_id' virtual machine",
	"shutdown", 			"scope host:  Shuts down a host. If the command is successful, then the host has been shut down\n".
							"             Parameter 'vm_id' not needed in this scope\n".
							"scope vm:    Powers off (hard) 'vm_id' virtual machine\n".
							"scope guest: Issues shutdown command to the guest operating system (gracefully)",
	"reboot", 				"scope host:  Reboots a host\n".
							"             Parameter 'vm_id' not needed in this scope\n".
							"scope vm:    Resets power (hard) on 'vm_id' virtual machine\n".
							"scope guest: Issues reboot command to the guest operating system (gracefully)",
	"kill", 				"scope vm:    Do an immediate power off of 'vm_id' virtual machine\n".
							"             This issues a SIGKILL to the vmx process of 'vm_id' virtual machine\n".
							"             Pending synchronous I/Os may not be written out!",
	"mount", 				"Creates a new network-attached storage datastore\n".
							"Possible types: CIFS / NFS / NFS41\n".
							"mountuser/mountpassword optional for CIFS / NFS41",
	"unmount", 				"Removes a network-attached datastore from a host\n".
							"Only if no access pending on this datastore (e.g. registered VMs)",
	"create_snapshot", 		"Creates a new snapshot of 'vm_id' virtual machine",
	"remove_snapshot", 		"Removes 'snapid' snapshot and deletes any associated storage",
	"last_snapshot", 		"List current (i.e. latest) snapshot for 'vm_id' virtual machine",
	"revert_snapshot", 		"Change the execution state of a virtual machine to the state of 'snapid' snapshot",
	"list_attached_vms", 	"Lists all attached VMs of 'moid' datastore",
	"list_all_datastores", 	"Lists all datastores available on a given host system"
);
my %parameter_helper = (
	"host", 			"--host:          VMware ESXi IP address",
	"user", 			"--user:          ESXi username - usually root",
	"password", 		"--password:      ESXi users password",
	"plain", 			"--plain:         Boolean parameter for action summary which defines to show colors for state or not\n".
						"                 OPTIONAL - <default --no-plain, i.e. show colors>",
	"mounthost",		"--mounthost:     IP address of system to be mounted",
	"mountpoint", 		"--mountpoint:    Network mount point (e.g.: NFS /mnt/HDD/mount)",
	"moid", 			"--moid:          Unique identifier of the internal datastore. Can be obtained with summary",
	"mounttype", 		"--mounttype:     The type of mount to be established (CIFS / NFS / NFS41)",
	"mountaccess", 		"--mountaccess:   readOnly or readWrite <default readWrite>",
	"localpath", 		"--localpath:     The VMware path of the created mountpoint (i.e. Datastore-Name)",
	"mountuser", 		"--mountuser:     OPTIONAL input for CIFS or NFS41 - User to be used for mount",
	"mountpassword", 	"--mountpassword: OPTIONAL input for CIFS or NFS41 - Users password to be used for mount",
	"vm_id", 			"--vm_id:         ID of the required VirtualMachine. Can be obtained by using summary/list_attached_vms",
	"scope", 			"--scope:         Scope this action has to be applied to (host / vm / guest)",
	"snapname", 		"--snapname:      Name of the Snapshot. Either one word or use \"Whatever name you want to use\"\n".
						"                 OPTIONAL - <default prefix datetime>",
	"snapdesc", 		"--snapdesc:      Description of the Snapshot. Either one word or use \"Whatever description you want to use\"\n".
						"                 OPTIONAL",
	"snapid",	 		"--snapid:        Identifier of the Snapshot. Can be obtained with summary/last_snapshot",
	"suppress",			"--(no-)suppress: Boolean parameter for suppressing VM Power-On after revert to Snapshot\n".
						"                 OPTIONAL - <default --no-suppress, i.e. false>", 
	"mem", 				"--(no-)mem:      Boolean parameter for whether RAM of guest should be included in Snapshot\n".
						"                 OPTIONAL - <default --mem, i.e. true>",
	"quiesce", 			"--(no-)quiesce:  Boolean parameter for whether guest filesystem is stopped or not - Only if VMware-Tools installed in guest\n".
						"                 OPTIONAL - <default --no-quiesce, i.e. false>",
	"childs", 			"--(no-)childs:   Boolean parameter for whether child Snapshots should be removed as well\n".
						"                 OPTIONAL - <default --no-childs, i.e. false>"
);
# Helper
{
	local(@ARGV) = @ARGV;
	GetOptions ('verbose!' => \$verbose,
				'debug!' => \$debug,
				'help' => \$help);
}
usage() if $help;
usage() if ($#ARGV == -1);
#
# Scriptstart
#
print ("#--- getting action\n") if $::debug;
my $action = $ARGV[0];
TrimSpaces($action);
if ( not grep(/\b$action\b/, @implemented_actions) ) {
	privatedie ("\nNot implemented action $action defined.\nUse '$0 --help' to get a list of implemented actions.", __LINE__);
}
$verbose = 1 if ($#ARGV == 0);
if ($#ARGV == 0) { ParameterHelper($verbose,""); exit;};
print ("#--- getting options\n") if $::debug;
GetOptions ('host=s' => \$host,
			'user=s' => \$user,
			'password=s' => \$password,
			'scope=s' => \$scope,
			'mounthost=s' => \$mounthost,
			'mountpoint=s' => \$mountpoint,
			'moid=s' => \$moid,
			'mounttype=s' => \$mounttype,
			'mountaccess=s' => \$mountaccess,
			'localpath=s' => \$localpath,
			'mountuser=s' => \$mountuser,
			'mountpassword=s' => \$mountpassword,
			'vm_id=s' => \$vm_id,
			'snapname=s' => \$snapname,
			'snapid=s' => \$snapid,
			'snapdesc=s' => \$snapdesc,
			'mem!' => \$mem,
			'quiesce!' => \$quiesce,
			'childs!' => \$childs,
			'plain!' => \$plain,
			'suppress!' => \$suppress);

$snapname=$datetime.$snapname;
$snapdesc=$snapdesc." ";
$todieornot = $todieornot + checkParameters("host",$host) if ($action =~ /^(.*)$/);
if (not $todieornot >= 2) {checkIP("host",$host);}
$todieornot = $todieornot + checkParameters("user",$user) if ($action =~ /^(.*)$/);
$todieornot = $todieornot + checkParameters("password",$password) if ($action =~ /^(.*)$/);
$todieornot = $todieornot + checkParameters("scope",$scope) if ($action =~ /^(poweron|shutdown|reboot|kill)$/);
$todieornot = $todieornot + checkParameters("mounthost",$mounthost) if ($action =~ /^(mount|unmount)$/);
if (not $todieornot >= 2) {checkIP("mounthost",$mounthost) if ($action =~ /^(mount|unmount)$/);}
$todieornot = $todieornot + checkParameters("mountpoint",$mountpoint) if ($action =~ /^(mount|unmount)$/);
$todieornot = $todieornot + checkParameters("mounttype",$mounttype) if ($action =~ /^(mount)$/);
$todieornot = $todieornot + checkParameters("mountaccess",$mountaccess) if ($action =~ /^(mount)$/);
$todieornot = $todieornot + checkParameters("moid",$moid) if ($action =~ /^(list_attached_vms)$/);
$todieornot = $todieornot + checkParameters("localpath",$localpath) if ($action =~ /^(mount)$/);
if (not $scope eq "host") {$todieornot = $todieornot + checkParameters("vm_id",$vm_id) if ($action =~ /^(create_snapshot|last_snapshot|poweron|shutdown|reboot|kill)$/);}
$todieornot = $todieornot + checkParameters("snapname",$snapname) if ($action =~ /^(create_snapshot)$/);
$todieornot = $todieornot + checkParameters("snapid",$snapid) if ($action =~ /^(remove_snapshot|revert_snapshot)$/);
$todieornot = $todieornot + checkParameters("snapdesc",$snapdesc) if ($action =~ /^(create_snapshot)$/);
$todieornot = $todieornot + checkParameters("mem",$mem) if ($action =~ /^(create_snapshot)$/);
$todieornot = $todieornot + checkParameters("quiesce",$quiesce) if ($action =~ /^(create_snapshot)$/);
$todieornot = $todieornot + checkParameters("childs",$childs) if ($action =~ /^(remove_snapshot)$/);
$todieornot = $todieornot + checkParameters("plain",$plain) if ($action =~ /^(summary)$/);
$todieornot = $todieornot + checkParameters("suppress",$suppress) if ($action =~ /^(revert_snapshot)$/);
print "\nParameter checking - return code is $todieornot" if $::debug;
if ( $todieornot >= 2 ) { ParameterHelper(1,"\n\n"); exit;}
#
# Check scope dependant actions for definition & map scope action to single action if applicable
#

my ($firstaction) = $action;
if ($action =~ /^(poweron|shutdown|reboot|kill)$/) {
	switch ($scope) {
		case "host" {
			if ($action =~ /^(shutdown|reboot)$/) {
				$action=$action."host";
			} else { print "\nNot implemented action $action in scope $scope\n"; exit; }
		}
		case "vm" {
			if ($action =~ /^(poweron|shutdown|reboot|kill)$/) {
				$action=$action."vm";
			} else { print "\nNot implemented action $action in scope $scope\n"; exit; }
		}
		case "guest" {
			if ($action =~ /^(shutdown|reboot)$/) {
				$action=$action."guest";
			} else { print "\nNot implemented action $action in scope $scope\n"; exit; }
		}
		case "dummy" {
			print "\nMissing scope for action $action. Please define with --scope host/vm/guest\n"; 
			exit;
		}
	}
}
#
# Create init Messages
#
$message = createHelloMessage($user,$password);
$response = sendRequest($host,$message);
$retval = checkReponse($response);
#
# If connection works proceed
#
if( $retval eq 1 ) {
	my $cookie = extractCookie($response);
	if ( $action eq "summary" ) {
		createSummary($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs,$cookie);
	} else {
		$message = createMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
		print "\n#--- Creating and sending $action command to $host ...\n" if $::verbose;
		$response = &sendRequest($host,$message,$cookie);
		$retval = checkReponse($response);
		if($retval eq 1) {
			if ($action eq "list_attached_vms") {
				print ("\n---------Trying to extract vm ids for \"".$mounthost.":".$mountpoint."\n") if $::debug;
				my @vms = extract_vm_ids($response->as_string);		
				if (not  @vms ) {
					print "\nThere are no attached VMs for this datastore \"".$mounthost.":".$mountpoint."\"\n\n";
				} 
				foreach (@vms) {
					print "\nAttached VM ID \"".$_."\"";
					$action = "get_vm_name";
					$message = createMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$_,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
					print "\n#--- Creating and sending $action command to $host ...\n" if $::verbose;
					$response = &sendRequest($host,$message,$cookie);
					$retval = checkReponse($response);
					if($retval eq 1) {
						print ("\n---------Trying to extract vm name \n") if $::debug;
						my $vmn = extract_vm_name($response->as_string);
						print " = \"$vmn\"";
					} else {
						print "\t" . ( $plain ? "" : color("red")."⬛ ".color('reset') ) . "Sent $action message but did not get confirmation back from $host\n\n";
					}
				}
				print "\n\n";
			}
			if ($action eq "list_all_datastores") {
				print ("\n---------Trying to extract datastore id \n") if $::debug;
				my @vms = extract_vm_ids($response->as_string);
				foreach (@vms) {
					print "\nAttached Datastores \"".$_."\"";
				}
				print "\n\n";
			}
			if ($action eq "last_snapshot") {
				print ("\n---------Trying to extract latest snapshot \n") if $::debug;
				my $last_snap = extract_vm_last_snap($response->as_string);
				if (not defined $last_snap ) {
					print "There is no Snapshot for VM with ID \"$vm_id\"\n\n";
				} else {
					print "Last Snapshot for VM with ID \"$vm_id\" = \"$last_snap\"\n\n";
				}
			}
		} else {
			print "\t" . ( $plain ? "" : color("red")."⬛ ".color('reset') ) . "Sent $action message but did not get confirmation back from $host\n\n";
		}
	}
	if($retval eq 1) {
			print "\t" . ( $plain ? "" : color("green")."⬛ ".color('reset') ) .  "Successfully initiated $firstaction on $host\n\n";
	} else {
			print "\t" . ( $plain ? "" : color("red")."⬛ ".color('reset') ) . "Sent $firstaction message but did not get confirmation back from $host\n\n";
	}
} else {
	print "\t" . ( $plain ? "" : color("red")."⬛ ".color('reset') ) . "Failed to issue hello message to $host\n\n";
}
################################################################################################################################################################
#
# SUBS
#
################################################################################################################################################################
sub createSummary {
	my ($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs,$cookie) = @_;
	$mountpoint = "";
	$mounthost = "";
	$action = "list_all_datastores";
	$message = createMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
	print "\n#--- Creating and sending $action command to $host ...\n" if $::verbose;
	$response = &sendRequest($host,$message,$cookie);
	$retval = checkReponse($response);
	if($retval eq 1) {
		print ("\n---------Trying to extract datastore id \n") if $::debug;
		my @vmd = '';
		@vmd = extract_vm_ids($response->as_string);
		print "\n<Host>\"".$host."\"\n";
		foreach (@vmd) {
			$moid = $_;
			my $dsstatebox = '';
			my $dsstate = '';
			my $dsname = '';
			my $dstype = '';
			$action = "get_datastore_details";
			$message = createMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
			print "\n#--- Creating and sending $action command to $host ...\n" if $::verbose;
			$response = &sendRequest($host,$message,$cookie);
			$retval = checkReponse($response);
			if($retval eq 1) {
				$dsstate = extract_ds_state($response->as_string);
				switch ( $dsstate ) {
					case "gray" { $dsstatebox=color("ansi60")."⬛ ".color('reset') }
					case "red" { $dsstatebox=color("red")."⬛ ".color('reset') }
					case "yellow" { $dsstatebox=color("yellow")."⬛ ".color('reset') }
					case "green" { $dsstatebox=color("green")."⬛ ".color('reset') }
					case "orange" { $dsstatebox=color("ansi202")."⬛ ".color('reset') }
					case "light blue" { $dsstatebox=color("ansi14")."⬛ ".color('reset') }
				}
				$dsname = extract_ds_name($response->as_string);
				$dstype = extract_ds_type($response->as_string);
			} else {
				print "\t" . ( $plain ? "" : color("red") ."⬛ ".color('reset') ) .  "Sent $action message but did not get confirmation back from $host\n\n";
			}
			$action = "list_attached_vms";
			$message = createMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
			print "\n#--- Creating and sending $action command to $host ...\n" if $::verbose;
			$response = &sendRequest($host,$message,$cookie);
			$retval = checkReponse($response);
			if($retval eq 1) {
				my @vms = '';
				@vms = extract_vm_ids($response->as_string);
				if (not  @vms ) {
					print "  |\n";
					print "  |-<DataStore>MOID: \"".$moid."\" ; Name: \"$dsname\" ; Type: \"$dstype\" ; " . ( $plain ? "" : $dsstatebox ) . "Status: $dsstate\n";
					print "  |  |\n";
					print "  |  |- No attached VMs\n";
					print "  |  |\n";
					print "  | </DataStore>\n";
				} else {
					print "  |\n";
					print "  |-<DataStore>MOID: \"".$moid."\" ; Name: \"$dsname\" ; Type: \"$dstype\" ; " . ( $plain ? "" : $dsstatebox ) . "Status: $dsstate\n";
					foreach (@vms) {
						print "  |  |\n";
						print "  |  |-<VM>ID: ".$_."";
						$vm_id = $_;
						my $vmstate = '';
						my $vmstatebox = '';
						$action = "get_vm_state";
						$message = createMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
						print "\n#--- Creating and sending $action command to $host ...\n" if $::verbose;
						$response = &sendRequest($host,$message,$cookie);
						$retval = checkReponse($response);
						if($retval eq 1) {
							$vmstate = extract_vm_state($response->as_string);
							switch ( $vmstate ) {
								case "gray" { $vmstatebox=color("ansi60")."⬛ ".color('reset') }
								case "red" { $vmstatebox=color("red")."⬛ ".color('reset') }
								case "yellow" { $vmstatebox=color("yellow")."⬛ ".color('reset') }
								case "green" { $vmstatebox=color("green")."⬛ ".color('reset') }
							}
						} else {
							print "\t" . ( $plain ? "" : color("red") ."⬛ ".color('reset') ) . "Sent $action message but did not get confirmation back from $host\n\n";
						}
						$action = "get_vm_name";
						$message = createMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
						print "\n#--- Creating and sending $action command to $host ...\n" if $::verbose;
						$response = &sendRequest($host,$message,$cookie);
						$retval = checkReponse($response);
						if($retval eq 1) {
							print ("\n---------Trying to extract vm name \n") if $::debug;
							my $vmn = extract_vm_name($response->as_string);
							print " ; Name: \"$vmn\" ; " . ( $plain ? "" : $vmstatebox ) . "State: $vmstate\n";
							$action = "last_snapshot";
							print ("\n---------Trying to extract latest snapshot \n") if $::debug;
							$message = createMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
							print "\n#--- Creating and sending $action command to $host ...\n" if $::verbose;
							$response = &sendRequest($host,$message,$cookie);
							$retval = checkReponse($response);
							if($retval eq 1) {
								print ("\n---------Trying to extract latest snapshot \n") if $::debug;
								my $last_snap = extract_vm_last_snap($response->as_string);
								if (not defined $last_snap ) {
									print "  |  |  |- No Snapshot available\n";
								} else {
									print "  |  |  |-<Snapshot current>$last_snap</Snapshot current>\n";
									print ("\n---------Trying to extract snapshot tree\n") if $::debug;
									my @snaps = extract_vm_snaps($response->as_string);	
									if ( @snaps ) {
											print "  |  |  |  |-<Snapshot Tree>\n";
										foreach ( sort {$b cmp $a} @snaps ) {
											my ($snap) = $_;
											print ("\n---------Trying to extract snapshot details\n") if $::debug;
											my @snapdet = extract_vm_snap_detail($response->as_string,$snap);
											print "  |  |  |  |  |-<Snapshot>ID:   $snap\n";
											print "  |  |  |  |  |  |-Name:        $snapdet[1]\n";
											print "  |  |  |  |  |  |-Description: $snapdet[2]\n";
											print "  |  |  |  |  |  |-Created:     $snapdet[4]\n";
											print "  |  |  |  |  |  |-State:       $snapdet[5]\n";
											print "  |  |  |  |  |  |-Quiesced:    $snapdet[6]\n";
											print "  |  |  |  |  | </Snapshot>\n";
											print "  |  |  |  |  |\n";
										}
											print "  |  |  |  | </Snapshot Tree>\n";
									}
								}
							} else {
								print "\t" . ( $plain ? "" : color("red")."⬛ ".color('reset') ) . "Sent $action message but did not get confirmation back from $host\n\n";
							}
						} else {
							print "\t" . ( $plain ? "" : color("red") ."⬛ ".color('reset') ) .  "Sent $action message but did not get confirmation back from $host\n\n";
						}
						print "  |  | </VM>\n";
					}						
					print "  |  |\n";
					print "  | </DataStore>\n";
				}
			} else {
				print "\t" . ( $plain ? "" : color("red") ."⬛ ".color('reset') ) .  "Sent $action message but did not get confirmation back from $host\n\n";
			}
		}		
		print "  |\n";	
		print "</Host>\n";
	} else {
		print "\t" . ( $plain ? "" : color("red") ."⬛ ".color('reset') ) .  "Sent $action message but did not get confirmation back from $host\n\n";
	}
	################# END OF SUMMARY
}
sub checkParameters {
	my($opt_name,$opt_value) = @_;
	if ( $opt_value eq "") { print ("Expected option \"$opt_name\" with no value\n"); return 2; };
	TrimSpaces($opt_value);
	print ("Option name \"$opt_name\" with value \"$opt_value\"\n") if $::verbose;
	return 0;
}

sub checkIP {
	my($opt_name,$opt_value) = @_;
	if ( $opt_value eq "") { privatedie("Expected option \"$opt_name\" with no value", __LINE__) };
	TrimSpaces($opt_value);
	if($opt_value =~ /^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$/) {
		checkIPalive($opt_value) if $opt_name eq "host";
		return 0;
	} else {
		privatedie ("Option name \"$opt_name\" with value \"$opt_value\" does not match a valid IP address format.", __LINE__);
	}
}

sub checkIPalive {
	my($ip) = @_;
	my $p = Net::Ping->new( "tcp", 5);
	$p->port_number(scalar(getservbyname("http", "tcp")));
	if ( $p->ping("$ip") ) {
		$p->close();
		print "Host \@IP: $ip is alive.\n\n" if $::verbose;
		return 0;
	} else {
		$p->close();
		privatedie ("Host \@IP: $ip is not reachable.", __LINE__);
	}
	
}

sub TrimSpaces {
	foreach (@_) {
		s/^\s+|\s*$//g
	}
}

sub XMLtoSTDOUT {
		my ($msg) = @_;
        $msg=~ s/></>\n</g;
		return $msg;
}

sub MessageToXML {
	foreach (@_) {
		s/\s+</</g
	}
}

sub createHelloMessage {
        my ($user,$pass) = @_;
		print "\n#--- Creating Hello Message\n" if $::debug;
        my $msg = "<soap:Envelope xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\"><soap:Body><Login xmlns=\"urn:vim25\"><_this xsi:type=\"SessionManager\" type=\"SessionManager\" serverGuid=\"\">ha-sessionmgr</_this><userName>$user</userName><password>$pass</password><locale>en_US</locale></Login></soap:Body></soap:Envelope>";
        return $msg;
}

sub sendRequest {
	my ($host,$msg,$cookie) = @_;
	my $host_to_connect = "https://" . $host . "/sdk";
	my $userAgent = LWP::UserAgent->new(agent => 'VMware VI Client/4.0.0', ssl_opts => {SSL_verify_mode => '0', verify_hostname => '0'});
	print "\n#--- Sending request to \"$host\"\n" if $::debug;
	my $request = HTTP::Request->new(POST => $host_to_connect);
	$request->header(SOAPAction => '"urn:vim25/4.0"');
	$request->content($msg);
	$request->content_type("text/xml; charset=utf-8");
	if(defined($cookie)) {
		$cookie->add_cookie_header($request);
	}
	my $rsp = $userAgent->request($request);
}

sub extractCookie {
	my ($rsp) = @_;
	my $cookie_jar = HTTP::Cookies->new;
	print "\n#--- Extracting Session Cookie\n" if $::debug;
	$cookie_jar->extract_cookies($rsp);

	return $cookie_jar;
}

sub checkReponse {
	my ($resp) = @_;
	my $ret = -1;
	if($resp->code == 200) {
			print "\n#--- Server response\n" . ( $plain ? (XMLtoSTDOUT($resp->as_string)) : color("green") . (XMLtoSTDOUT($resp->as_string)) . color("reset") ) . "\n" if $::debug;
			return 1;
	} else {
			print "\n" . ( $plain ? (XMLtoSTDOUT($resp->error_as_HTML)) : color("red") . (XMLtoSTDOUT($resp->error_as_HTML)) . color("reset") ) . "\n" if $::verbose;
			return $ret;
	}
}

sub createMessage {
	my ($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs) = @_;
	my $msg = "<soap:Envelope xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\"><soap:Body>";
	$msg = $msg . createActionMessage($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs);
	$msg = $msg . "</soap:Body></soap:Envelope>";
	MessageToXML($msg);
	print "\n#--- Assembled message\n".(XMLtoSTDOUT($msg))."\n" if $::debug;
	return $msg;
}

sub createActionMessage {
	my ($action,$mounthost,$mountpoint,$moid,$mounttype,$mountaccess,$localpath,$mountuser,$mountpassword,$vm_id,$snapname,$snapdesc,$snapid,$suppress,$mem,$quiesce,$childs) = @_;
	my ($msg) = "";
	switch ($action) {
		case "ssh_on" { 
			$msg = "<StartService xmlns=\"urn:vim25\">
						<_this xsi:type=\"HostServiceSystem\" type=\"HostServiceSystem\">serviceSystem</_this>
							<id>TSM-SSH</id>
					</StartService>";
		}
		case "ssh_off" { 
			$msg = "<StopService xmlns=\"urn:vim25\">
						<_this xsi:type=\"HostServiceSystem\" type=\"HostServiceSystem\">serviceSystem</_this>
							<id>TSM-SSH</id>
					</StopService>";
		}		
		case "poweronvm" { 
			$msg = "<PowerOnVM_Task xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachine\" type=\"VirtualMachine\">".$vm_id."</_this>
					</PowerOnVM_Task>";
		}		
		case "shutdownhost" { 
			$msg = "<ShutdownHost_Task xmlns=\"urn:vim25\">
						<_this xsi:type=\"HostSystem\" type=\"HostSystem\" serverGuid=\"\">ha-host</_this>
							<force>true</force>
					</ShutdownHost_Task>";
		}		
		case "shutdownvm" { 
			$msg = "<PowerOffVM_Task xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachine\" type=\"VirtualMachine\">".$vm_id."</_this>
					</PowerOffVM_Task>";
		}		
		case "shutdownguest" { 
			$msg = "<ShutdownGuest xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachine\" type=\"VirtualMachine\">".$vm_id."</_this>
					</ShutdownGuest>";
		}	
		case "reboothost" { 
			$msg = "<RebootHost_Task xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachine\" type=\"VirtualMachine\">ha-host</_this>
						<force>true</force>
					</RebootHost_Task>";
		}	
		case "rebootvm" { 
			$msg = "<ResetVM_Task xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachine\" type=\"VirtualMachine\">".$vm_id."</_this>
					</ResetVM_Task>";
		}		
		case "rebootguest" { 
			$msg = "<RebootGuest xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachine\" type=\"VirtualMachine\">".$vm_id."</_this>
					</RebootGuest>";
		}		
		case "killvm" { 
			$msg = "<TerminateVM xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachine\" type=\"VirtualMachine\">".$vm_id."</_this>
					</TerminateVM>";
		}
		case "list_attached_vms" {
			my ($mountid) = "";
			if ( $mounthost eq "" && $mountpoint eq "" ) {
				$mountid = $moid;
			} else {
				$mountid = $mounthost.":".$mountpoint;
			}
			$msg = "<RetrieveProperties xmlns=\"urn:vim25\">
						<_this xsi:type=\"PropertyCollector\" type=\"PropertyCollector\">ha-property-collector</_this>
							<specSet>
								<propSet>
									<type>Datastore</type>
									<all>false</all>
									<pathSet>vm</pathSet>
								</propSet>
								<objectSet>
									<obj type=\"Datastore\">".$mountid."</obj>
									<skip>false</skip>
								</objectSet>
								<reportMissingObjectsInResults>true</reportMissingObjectsInResults>
							</specSet>
					</RetrieveProperties>";
		}
		case "get_datastore_details" {
			my ($mountid) = "";
			if ( $mounthost eq "" && $mountpoint eq "" ) {
				$mountid = $moid;
			} else {
				$mountid = $mounthost.":".$mountpoint;
			}
			$msg = "<RetrieveProperties xmlns=\"urn:vim25\">
						<_this xsi:type=\"PropertyCollector\" type=\"PropertyCollector\">ha-property-collector</_this>
							<specSet>
								<propSet>
									<type>Datastore</type>
									<all>true</all>
								</propSet>
								<objectSet>
									<obj type=\"Datastore\">".$mountid."</obj>
									<skip>false</skip>
								</objectSet>
								<reportMissingObjectsInResults>true</reportMissingObjectsInResults>
							</specSet>
					</RetrieveProperties>";
		}
		case "list_all_vms" {
			$msg = "<RetrieveProperties xmlns=\"urn:vim25\">
						<_this xsi:type=\"PropertyCollector\" type=\"PropertyCollector\">ha-property-collector</_this>
							<specSet>
								<propSet>
									<type>HostSystem</type>
									<all>false</all>
									<pathSet>vm</pathSet>
								</propSet>
								<objectSet>
									<obj type=\"HostSystem\">ha-host</obj>
									<skip>false</skip>
								</objectSet>
								<reportMissingObjectsInResults>true</reportMissingObjectsInResults>
							</specSet>
					</RetrieveProperties>";
		}
		case "list_all_datastores" {
			$msg = "<RetrieveProperties xmlns=\"urn:vim25\">
						<_this xsi:type=\"PropertyCollector\" type=\"PropertyCollector\">ha-property-collector</_this>
							<specSet>
								<propSet>
									<type>HostSystem</type>
									<all>false</all>
									<pathSet>datastore</pathSet>
								</propSet>
								<objectSet>
									<obj type=\"HostSystem\">ha-host</obj>
									<skip>false</skip>
								</objectSet>
								<reportMissingObjectsInResults>true</reportMissingObjectsInResults>
							</specSet>
					</RetrieveProperties>";
		}
		case "last_snapshot" {
			$msg = "<RetrieveProperties xmlns=\"urn:vim25\">
						<_this xsi:type=\"PropertyCollector\" type=\"PropertyCollector\">ha-property-collector</_this>
							<specSet>
								<propSet>
									<type>VirtualMachine</type>
									<all>false</all>
										<pathSet>snapshot</pathSet>
								</propSet>
								<objectSet>
									<obj type=\"VirtualMachine\">".$vm_id."</obj>
									<skip>false</skip>
								</objectSet>
								<reportMissingObjectsInResults>false</reportMissingObjectsInResults>
							</specSet>
					</RetrieveProperties>";
		}
		case "get_vm_name" {
			$msg = "<RetrieveProperties xmlns=\"urn:vim25\">
						<_this xsi:type=\"PropertyCollector\" type=\"PropertyCollector\">ha-property-collector</_this>
							<specSet>
								<propSet>
									<type>VirtualMachine</type>
									<all>false</all>
									<pathSet>name</pathSet>
								</propSet>
								<objectSet>
									<obj type=\"VirtualMachine\">".$vm_id."</obj>
									<skip>false</skip>
								</objectSet>
								<reportMissingObjectsInResults>false</reportMissingObjectsInResults>
							</specSet>
					</RetrieveProperties>";
		}
		case "get_vm_state" {
			$msg = "<RetrieveProperties xmlns=\"urn:vim25\">
						<_this xsi:type=\"PropertyCollector\" type=\"PropertyCollector\">ha-property-collector</_this>
							<specSet>
								<propSet>
									<type>VirtualMachine</type>
									<all>false</all>
									<pathSet>guestHeartbeatStatus</pathSet>
								</propSet>
								<objectSet>
									<obj type=\"VirtualMachine\">".$vm_id."</obj>
									<skip>false</skip>
								</objectSet>
								<reportMissingObjectsInResults>false</reportMissingObjectsInResults>
							</specSet>
					</RetrieveProperties>";
			

		}
		case "create_snapshot" {
			$msg = "<CreateSnapshot_Task xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachine\" type=\"VirtualMachine\">".$vm_id."</_this>
							<name>".$snapname."</name>
							<description>".$snapdesc."</description>
							<memory>".$mem."</memory>
							<quiesce>".$quiesce."</quiesce>
					</CreateSnapshot_Task>";
		}		
		case "remove_snapshot" {
			$msg = "<RemoveSnapshot_Task xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachineSnapshot\" type=\"VirtualMachineSnapshot\">".$snapid."</_this>
							<removeChildren>".$childs."</removeChildren>
					</RemoveSnapshot_Task>";
		}		
		case "revert_snapshot" {
			$msg = "<RevertToSnapshot_Task xmlns=\"urn:vim25\">
						<_this xsi:type=\"VirtualMachineSnapshot\" type=\"VirtualMachineSnapshot\">".$snapid."</_this>
							<suppressPowerOn>".$suppress."</suppressPowerOn>
					</RevertToSnapshot_Task>";
		}		
		case "mount" {
			if ( defined $mountuser && defined $mountpassword) {
				$msg = "<CreateNasDatastore xmlns=\"urn:vim25\">
						<_this xsi:type=\"HostDatastoreSystem\" type=\"HostDatastoreSystem\">ha-datastoresystem</_this>
							<spec>
								<remoteHost>".$mounthost."</remoteHost>
								<remotePath>".$mountpoint."</remotePath>
								<localPath>".$localpath."</localPath>
								<accessMode>".$mountaccess."</accessMode>
								<type>".$mounttype."</type>
								<userName>".$mountuser."</userName>
								<password>".$mountpassword."</password>
							</spec>
					</CreateNasDatastore>";
			} else {
				$msg = "<CreateNasDatastore xmlns=\"urn:vim25\">
							<_this xsi:type=\"HostDatastoreSystem\" type=\"HostDatastoreSystem\">ha-datastoresystem</_this>
								<spec>
									<remoteHost>".$mounthost."</remoteHost>
									<remotePath>".$mountpoint."</remotePath>
									<localPath>".$localpath."</localPath>
									<accessMode>".$mountaccess."</accessMode>
									<type>".$mounttype."</type>
								</spec>
						</CreateNasDatastore>";
			}
		}		
		case "unmount" {
			$msg = "<RemoveDatastore xmlns=\"urn:vim25\">
						<_this xsi:type=\"HostDatastoreSystem\" type=\"HostDatastoreSystem\">ha-datastoresystem</_this>
							<datastore xsi:type=\"Datastore\" type=\"Datastore\">".$mounthost.":".$mountpoint."</datastore>
					</RemoveDatastore>";
		}
	}
	return $msg;
}

sub extract_vm_ids {
	
	my ($msg) = @_;
	my (@matched);
	$msg =~ /<val(.*?)\/val>/;
	$msg = $1;
	while ($msg =~ />(\d[^<]*)</g) {
		push @matched, $1;
	}
	return @matched;
}

sub extract_vm_name {	
	my ($msg) = @_;
	$msg =~ /<val(.*?)\/val>/;
	$msg = $1;
	my ($matched) = $msg =~ />(.[^<]*)</;
	return $matched;
}

sub extract_vm_last_snap {
	my ($msg) = @_;
	$msg =~ /<currentSnapshot(.*?)\/currentSnapshot>/;
	$msg = $1;
	my ($matched) = $msg =~ />(.[^<]*)</;
	return $matched;
}

sub extract_vm_snaps {
	my ($msg) = @_;
	my (@matched);
	$msg =~ /<rootSnapshotList>(.*?)<\/rootSnapshotList>/;
	$msg = $1;
	while ($msg =~/VirtualMachineSnapshot">(.*?)<\/snapshot/g) {
		push @matched, $1;
	}
	return @matched;
}

sub extract_vm_state {
	my ($msg) = @_;
	$msg =~ /<val(.*?)\/val>/;
	$msg = $1;
	my ($matched) = $msg =~ />(.[^<]*)</;
	return $matched;
}

sub extract_ds_name {
	my ($msg) = @_;
	$msg =~ /<propSet><name>info<\/name(.*?)\/propSet>/;
	$msg = $1;
	$msg =~ /<name(.*?)\/name>/;
	$msg = $1;
	my ($matched) = $msg =~ />(.[^<]*)</;
	return $matched;
}

sub extract_ds_type {
	my ($msg) = @_;
	$msg =~ /<name>info(.*?)\/propSet>/;
	$msg = $1;
	my ($matched) = $msg =~ /<type>(.[^<]*)<\/type/;
	return $matched;
}

sub extract_ds_state {
	my ($msg) = @_;
	$msg =~ /<name>overallStatus<\/name>(.*?)\/propSet>/;
	$msg = $1;
	$msg =~ /<val(.*?)\/val>/;
	my ($matched) = $msg =~ />(.[^<]*)</;
	return $matched;
}

sub extract_vm_snap_detail {
	my ($msg,$snap) = @_;
	my (@matched);
	push @matched, $snap;
	$msg =~ /<snapshot type="VirtualMachineSnapshot">${snap}(.*?)<\/quiesced>/;
	$msg = $1;
	$msg =~ s/(\d)T/$1 @ /g;
	$msg =~ s/.(\d+)Z/ /g;
	push @matched, $msg =~ /name>(.*?)<\/name/;
	push @matched, $msg =~ /description>(.*?)<\/description/;
	push @matched, $msg =~ /id>(.*?)<\/id/;
	push @matched, $msg =~ /createTime>(.*?)<\/createTime/;
	push @matched, $msg =~ /state>(.*?)<\/state/;
	push @matched, $msg =~ /quiesced>(.*?)$/;
	return @matched;
}

sub privatedie {
	my($diemsg,$dieline) = @_;
	print ($diemsg . ( $::debug ? "\n" . ( $plain ? "[@ ". __FILE__ . " in line " . $dieline . " ]" : color("red") . "[@ ". __FILE__ . " in line " . $dieline . " ]" . color("reset") ) : '' ) . "\n\n");
	exit;
}

sub ParameterHelper {
	my ($verbose,$newlines) = @_;
	if ( $verbose ) {
	print ($newlines."Called action: $action\n");
	print ("\n".$action_helper{$action}."\n");
	print ("\nExpected arguments:\n" . join(", ", @host_parameters) ); 
	switch ($action) {
		case "ssh_on" { print "\n"; foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); } }
		case "ssh_off" { print "\n"; foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); } }
		case "summary" { print "\n"; foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); } }
		case "list_all_datastores" { print "\n"; foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); } }
		case "mount" { print (", " . join(", ", @mount_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@mount_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "unmount" { print (", " . join(", ", @unmount_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@unmount_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "create_snapshot" { print (", " . join(", ", @create_snapshot_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@create_snapshot_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "remove_snapshot" { print (", " . join(", ", @remove_snapshot_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@remove_snapshot_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "revert_snapshot" { print (", " . join(", ", @revert_snapshot_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@revert_snapshot_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "last_snapshot" { print (", " . join(", ", @last_snapshot_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@last_snapshot_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "list_attached_vms" { print (", " . join(", ", @list_attached_vms_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@list_attached_vms_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "poweron" { print (", " . join(", ", @poweron_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@poweron_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "shutdown" { print (", " . join(", ", @shutdown_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@shutdown_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "reboot" { print (", " . join(", ", @reboot_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@reboot_parameters) { print ("\n".$parameter_helper{$_}); } } 
		case "kill" { print (", " . join(", ", @kill_parameters) . "\n"); foreach (@host_parameters) { print ("\n".$parameter_helper{$_}); }; foreach (@kill_parameters) { print ("\n".$parameter_helper{$_}); } } 
	}
	print ("\n\n");
	}
}

sub usage {
    my($usage) = <<EndOfHelp;
-----------------------------------------------------------------------------------
Usage: $0 <action> <parameters> <--verbose> <--debug>
       $0 --help

Use $0 <action> to get a detailed list of possible and needed parameters.

Implemented actions:

EndOfHelp
	print $usage;
	foreach (@implemented_actions) { print $_."\n"; }
    exit
}
print "\n\n---END--- at line ".__LINE__."\n\n" if $::debug;