#!"\xampp\perl\bin\perl.exe"

###############################

# admin-wlib.pl = cgi-lib          Copyright (c) 1993-1999 Steven E. Brenner
# + basic functions of napp-it    Copyright (c) 2010-2024 Gnther Alka
# 15.jan 2024 win
################################

# Parts of cgi-lib@pobox.com
# $Id: cgi-lib.pl,v 2.18 1999/02/23 08:16:43 brenner Exp $
#
# Perl Routines to Manipulate CGI input
# cgi-lib@pobox.com
# $Id: cgi-lib.pl,v 2.18 1999/02/23 08:16:43 brenner Exp $
#
# Copyright (c) 1993-1999 Steven E. Brenner
# Unpublished work.
# Permission granted to use and modify this library so long as the
# copyright above is maintained, modifications are documented, and
# credit is given for any use of the library.
#
# Thanks are due to many people for reporting bugs and suggestions

# For more information, see:
#     http://cgi-lib.stanford.edu/cgi-lib/

$cgi_lib'version = sprintf("%d.%02d", q$Revision: 2.18 $ =~ /(\d+)\.(\d+)/);


# Parameters affecting cgi-lib behavior
# User-configurable parameters affecting file upload.
$cgi_lib'maxdata    = 131072;    # maximum bytes to accept via POST - 2^17
$cgi_lib'writefiles =      0;    # directory to which to write files, or
                                 # 0 if files should not be written
$cgi_lib'filepre    = "cgi-lib"; # Prefix of file names, in directory above

# Do not change the following parameters unless you have special reasons
$cgi_lib'bufsize  =  8192;    # default buffer size when reading multipart
$cgi_lib'maxbound =   100;    # maximum boundary length to be encounterd
$cgi_lib'headerout =    0;    # indicates whether the header has been printed


# ReadParse
# Reads in GET or POST data, converts it to unescaped text, and puts
# key/value pairs in %in, using "\0" to separate multiple selections

# Returns >0 if there was input, 0 if there was no input
# undef indicates some failure.

# Now that cgi scripts can be put in the normal file space, it is useful
# to combine both the form and the script in one place.  If no parameters
# are given (i.e., ReadParse returns FALSE), then a form could be output.

# If a reference to a hash is given, then the data will be stored in that
# hash, but the data from $in and @in will become inaccessable.
# If a variable-glob (e.g., *cgi_input) is the first parameter to ReadParse,
# information is stored there, rather than in $in, @in, and %in.
# Second, third, and fourth parameters fill associative arrays analagous to
# %in with data relevant to file uploads.

# If no method is given, the script will process both command-line arguments
# of the form: name=value and any text that is in $ENV{'QUERY_STRING'}
# This is intended to aid debugging and may be changed in future releases

sub ReadParse {
  # Disable warnings as this code deliberately uses local and environment
  # variables which are preset to undef (i.e., not explicitly initialized)
  local ($perlwarn);
  $perlwarn = $^W;
  $^W = 0;

  local (*in) = shift if @_;    # CGI input
  local (*incfn,                # Client's filename (may not be provided)
         *inct,                 # Client's content-type (may not be provided)
         *insfn) = @_;          # Server's filename (for spooled files)
  local ($len, $type, $meth, $errflag, $cmdflag, $got, $name);

  binmode(STDIN);   # we need these for DOS-based systems
  binmode(STDOUT);  # and they shouldn't hurt anything else
  binmode(STDERR);

  # Get several useful env variables
  $type = $ENV{'CONTENT_TYPE'};
  $len  = $ENV{'CONTENT_LENGTH'};
  $meth = $ENV{'REQUEST_METHOD'};

  if ($len > $cgi_lib'maxdata) { #'
      &CgiDie("cgi-lib.pl: Request to receive too much data: $len bytes\n");
  }

  if (!defined $meth || $meth eq '' || $meth eq 'GET' ||
      $meth eq 'HEAD' ||
      $type eq 'application/x-www-form-urlencoded') {
    local ($key, $val, $i);

    # Read in text
    if (!defined $meth || $meth eq '') {
      $in = $ENV{'QUERY_STRING'};
      $cmdflag = 1;  # also use command-line options
    } elsif($meth eq 'GET' || $meth eq 'HEAD') {
      $in = $ENV{'QUERY_STRING'};
    } elsif ($meth eq 'POST') {
        if (($got = read(STDIN, $in, $len) != $len))
          {$errflag="Short Read: wanted $len, got $got\n";};
    } else {
      &CgiDie("cgi-lib.pl: Unknown request method: $meth\n");
    }

    @in = split(/[&;]/,$in);
    push(@in, @ARGV) if $cmdflag; # add command-line parameters

    foreach $i (0 .. $#in) {
      # Convert plus to space
      $in[$i] =~ s/\+/ /g;

      # Split into key and value.
      ($key, $val) = split(/=/,$in[$i],2); # splits on the first =.

      # Convert %XX from hex numbers to alphanumeric
      $key =~ s/%([A-Fa-f0-9]{2})/pack("c",hex($1))/ge;
      $val =~ s/%([A-Fa-f0-9]{2})/pack("c",hex($1))/ge;

      # Associate key and value
      $in{$key} .= "\0" if (defined($in{$key})); # \0 is the multiple separator
      $in{$key} .= $val;
    }

  } elsif ($ENV{'CONTENT_TYPE'} =~ m#^multipart/form-data#) {
    # for efficiency, compile multipart code only if needed
$errflag = !(eval <<'END_MULTIPART');

    local ($buf, $boundary, $head, @heads, $cd, $ct, $fname, $ctype, $blen);
    local ($bpos, $lpos, $left, $amt, $fn, $ser);
    local ($bufsize, $maxbound, $writefiles) =
      ($cgi_lib'bufsize, $cgi_lib'maxbound, $cgi_lib'writefiles);


    # The following lines exist solely to eliminate spurious warning messages
    $buf = '';

    ($boundary) = $type =~ /boundary="([^"]+)"/; #";   # find boundary
    ($boundary) = $type =~ /boundary=(\S+)/ unless $boundary;
    &CgiDie ("Boundary not provided: probably a bug in your server")
      unless $boundary;
    $boundary =  "--" . $boundary;
    $blen = length ($boundary);

    if ($ENV{'REQUEST_METHOD'} ne 'POST') {
      &CgiDie("Invalid request method for  multipart/form-data: $meth\n");
    }

    if ($writefiles) {
      local($me);
      stat ($writefiles);
      $writefiles = "/tmp" unless  -d _ && -w _;
      # ($me) = $0 =~ m#([^/]*)$#;
      $writefiles .= "/$cgi_lib'filepre";
    }

    # read in the data and split into parts:
    # put headers in @in and data in %in
    # General algorithm:
    #   There are two dividers: the border and the '\r\n\r\n' between
    # header and body.  Iterate between searching for these
    #   Retain a buffer of size(bufsize+maxbound); the latter part is
    # to ensure that dividers don't get lost by wrapping between two bufs
    #   Look for a divider in the current batch.  If not found, then
    # save all of bufsize, move the maxbound extra buffer to the front of
    # the buffer, and read in a new bufsize bytes.  If a divider is found,
    # save everything up to the divider.  Then empty the buffer of everything
    # up to the end of the divider.  Refill buffer to bufsize+maxbound
    #   Note slightly odd organization.  Code before BODY: really goes with
    # code following HEAD:, but is put first to 'pre-fill' buffers.  BODY:
    # is placed before HEAD: because we first need to discard any 'preface,'
    # which would be analagous to a body without a preceeding head.

    $left = $len;
   PART: # find each part of the multi-part while reading data
    while (1) {
      die $@ if $errflag;

      $amt = ($left > $bufsize+$maxbound-length($buf)
              ?  $bufsize+$maxbound-length($buf): $left);
      $errflag = (($got = read(STDIN, $buf, $amt, length($buf))) != $amt);
      die "Short Read: wanted $amt, got $got\n" if $errflag;
      $left -= $amt;

      $in{$name} .= "\0" if defined $in{$name};
      $in{$name} .= $fn if $fn;

      $name=~/([-\w]+)/;  # This allows $insfn{$name} to be untainted
      if (defined $1) {
        $insfn{$1} .= "\0" if defined $insfn{$1};
        $insfn{$1} .= $fn if $fn;
      }

     BODY:
      while (($bpos = index($buf, $boundary)) == -1) {
        if ($left == 0 && $buf eq '') {
          foreach $value (values %insfn) {
            unlink(split("\0",$value));
          }
          &CgiDie("cgi-lib.pl: reached end of input while seeking boundary " .
                  "of multipart. Format of CGI input is wrong.\n");
        }
        die $@ if $errflag;
        if ($name) {  # if no $name, then it's the prologue -- discard
          if ($fn) { print FILE substr($buf, 0, $bufsize); }
          else     { $in{$name} .= substr($buf, 0, $bufsize); }
        }
        $buf = substr($buf, $bufsize);
        $amt = ($left > $bufsize ? $bufsize : $left); #$maxbound==length($buf);
        $errflag = (($got = read(STDIN, $buf, $amt, length($buf))) != $amt);
        die "Short Read: wanted $amt, got $got\n" if $errflag;
        $left -= $amt;
      }
      if (defined $name) {  # if no $name, then it's the prologue -- discard
        if ($fn) { print FILE substr($buf, 0, $bpos-2); }
        else     { $in {$name} .= substr($buf, 0, $bpos-2); } # kill last \r\n
      }
      close (FILE);
      last PART if substr($buf, $bpos + $blen, 2) eq "--";
      substr($buf, 0, $bpos+$blen+2) = '';
      $amt = ($left > $bufsize+$maxbound-length($buf)
              ? $bufsize+$maxbound-length($buf) : $left);
      $errflag = (($got = read(STDIN, $buf, $amt, length($buf))) != $amt);
      die "Short Read: wanted $amt, got $got\n" if $errflag;
      $left -= $amt;


      undef $head;  undef $fn;
     HEAD:
      while (($lpos = index($buf, "\r\n\r\n")) == -1) {
        if ($left == 0  && $buf eq '') {
          foreach $value (values %insfn) {
            unlink(split("\0",$value));
          }
          &CgiDie("cgi-lib: reached end of input while seeking end of " .
                  "headers. Format of CGI input is wrong.\n$buf");
        }
        die $@ if $errflag;
        $head .= substr($buf, 0, $bufsize);
        $buf = substr($buf, $bufsize);
        $amt = ($left > $bufsize ? $bufsize : $left); #$maxbound==length($buf);
        $errflag = (($got = read(STDIN, $buf, $amt, length($buf))) != $amt);
        die "Short Read: wanted $amt, got $got\n" if $errflag;
        $left -= $amt;
      }
      $head .= substr($buf, 0, $lpos+2);
      push (@in, $head);
      @heads = split("\r\n", $head);
      ($cd) = grep (/^\s*Content-Disposition:/i, @heads);
      ($ct) = grep (/^\s*Content-Type:/i, @heads);

      ($name) = $cd =~ /\bname="([^"]+)"/i; #";
      ($name) = $cd =~ /\bname=([^\s:;]+)/i unless defined $name;

      ($fname) = $cd =~ /\bfilename="([^"]*)"/i; #"; # filename can be null-str
      ($fname) = $cd =~ /\bfilename=([^\s:;]+)/i unless defined $fname;
      $incfn{$name} .= (defined $in{$name} ? "\0" : "") .
        (defined $fname ? $fname : "");

      ($ctype) = $ct =~ /^\s*Content-type:\s*"([^"]+)"/i;  #";
      ($ctype) = $ct =~ /^\s*Content-Type:\s*([^\s:;]+)/i unless defined $ctype;
      $inct{$name} .= (defined $in{$name} ? "\0" : "") . $ctype;

      if ($writefiles && defined $fname) {
        $ser++;
        $fn = $writefiles . ".$$.$ser";
        open (FILE, ">$fn") || &CgiDie("Couldn't open $fn\n");
        binmode (FILE);  # write files accurately
      }
      substr($buf, 0, $lpos+4) = '';
      undef $fname;
      undef $ctype;
    }

1;
END_MULTIPART
    if ($errflag) {
      local ($errmsg, $value);
      $errmsg = $@ || $errflag;
      foreach $value (values %insfn) {
        unlink(split("\0",$value));
      }
      &CgiDie($errmsg);
    } else {
      # everything's ok.
    }
  } else {
    &CgiDie("cgi-lib.pl: Unknown Content-type: $ENV{'CONTENT_TYPE'}\n$ENV{'REQUEST_METHOD'}");
  }

  # no-ops to avoid warnings
  $insfn = $insfn;
  $incfn = $incfn;
  $inct  = $inct;

  $^W = $perlwarn;

  return ($errflag ? undef :  scalar(@in));
}



# PrintHeader
# Returns the magic line which tells WWW that we're an HTML document

sub PrintHeader {
  return "Content-type: text/html\n\n";
}


# HtmlTop
# Returns the <head> of a document and the beginning of the body
# with the title and a body <h1> header as specified by the parameter
sub HtmlTop
{
  local ($title) = @_;

  return <<END_OF_TEXT;
<html>
<head>
<title>$title</title>
</head>
<body>
<h1>$title</h1>
END_OF_TEXT
}


# HtmlBot
# Returns the </body>, </html> codes for the bottom of every HTML page
sub HtmlBot
{
  return "</body>\n</html>\n";
}


# SplitParam
# Splits a multi-valued parameter into a list of the constituent parameters

sub SplitParam
{
  local ($param) = @_;
  local (@params) = split ("\0", $param);
  return (wantarray ? @params : $params[0]);
}



# MyBaseUrl
# Returns the base URL to the script (i.e., no extra path or query string)
sub MyBaseUrl {
  local ($ret, $perlwarn);
  $perlwarn = $^W; $^W = 0;
  $ret = 'http://' . $ENV{'SERVER_NAME'} .
         ($ENV{'SERVER_PORT'} != 80 ? ":$ENV{'SERVER_PORT'}" : '') .
         $ENV{'SCRIPT_NAME'};
  $^W = $perlwarn;
  return $ret;

}


# MyFullUrl
# Returns the full URL to the script (i.e., with extra path or query string)
sub MyFullUrl {
  local ($ret, $perlwarn);
  $perlwarn = $^W; $^W = 0;
  $ret = 'http://' . $ENV{'SERVER_NAME'} .
         ($ENV{'SERVER_PORT'} != 80 ? ":$ENV{'SERVER_PORT'}" : '') .
         $ENV{'SCRIPT_NAME'} . $ENV{'PATH_INFO'} .
         (length ($ENV{'QUERY_STRING'}) ? "?$ENV{'QUERY_STRING'}" : '');
  $^W = $perlwarn;
  return $ret;
}


# MyURL
# Returns the base URL to the script (i.e., no extra path or query string)
# This is obsolete and will be removed in later versions
sub MyURL  {
  return &MyBaseUrl;
}


# CgiError
sub CgiError {
  local (@msg) = @_;
  local ($i,$name);

  if (!@msg) {
    $name = &MyFullUrl;
    @msg = ("Error: script $name encountered fatal error\n");
  };

  if (!$cgi_lib'headerout) { #')
    print &PrintHeader;
    print "<html>\n<head>\n<title>$msg[0]</title>\n</head>\n<body>\n";
  }
  print "<tt>$msg[0]</tt>\n";
  foreach $i (1 .. $#msg) {
    print "<p>$msg[$i]</p>\n";
  }
  print "<br><hr><tt><b>Fragen Sie bei Bedarf Ihren Systembetreuer</b></tt>\n";

  $cgi_lib'headerout++;
}


# CgiDie
# Identical to CgiError, but also quits with the passed error message.

sub CgiDie {
  local (@msg) = @_;
  &CgiError (@msg);
  die @msg;
}


# PrintVariables
# Nicely formats variables.  Three calling options:
# A non-null associative array - prints the items in that array
# A type-glob - prints the items in the associated assoc array
# nothing - defaults to use %in
# Typical use: &PrintVariables()

sub PrintVariables {
  local (*in) = @_ if @_ == 1;
  local (%in) = @_ if @_ > 1;
  local ($out, $key, $output);

  $output =  "\n<dl compact>\n";
  foreach $key (sort keys(%in)) {
    foreach (split("\0", $in{$key})) {
      ($out = $_) =~ s/\n/<br>\n/g;
      $output .=  "<dt><b>$key</b>\n <dd>:<i>$out</i>:<br>\n";
    }
  }
  $output .=  "</dl>\n";

  return $output;
}

# PrintEnv
# Nicely formats all environment variables and returns HTML string
sub PrintEnv {
  &PrintVariables(*ENV);
}











 #########################################
## my library##

# private punktionen
# lines below Copyright Gnther Alka 2009-2019
###############



## called from console
sub Readsocket {
##

    my ($v,@v,@k,$k,$par);

    $par=$_[0];
    $par=~s/\'//g;

    $par=~s/action.*//;
    #print "par: $par\n\n";

    # 'cli=1::id=xx yy::l1=abc::...'


    @v=split(/::/,$par);
    foreach my $v (@v) {
           @k=();
           @k=split(/=/,$v);
           if (($k[0] ne "") && ($k[1] ne "")) {
                 $in{$k[0]}=$k[1];
           }
    }
    $sys{'Readsocket'}="1";


    # debug
    # &print_hash(%in);
}



########       # execute system commands   ex. &exe("ls -a","n","ref");  n=no error   2>&1; ref="0" > no log
sub exe {       # ``
##########
  my ($r,$e,$n,$err,$t,$ref);
  if (!(defined $sys{'auto_os'})) { $sys{'auto_os'}=""; }

  $e="$_[0]";
  $e=~s/[^a-zA-Z0-9 \(\)\~\-\_\/\=\|\.\,\`\'\"\;\:\@\?\+\>\<\n\r\*\%\&\$\\]//gs;  #pipe erlaubt ?

  $ref=$_[2];
  if (($ref ne "0") && ($ref ne "")) {
      $ref=" ($ref)";
  }

  $t=&get_caller("p1");           # caller parent 1
  $t=~s/\n+$//s;

  #log opt with reference
  if (($in{'l1'} ne "")  && ($ref ne "0")) {
       &log_add("$t\nexe$ref\: $e ");
  }

  $err="";
  if (defined($_[1])) {
    $err="$_[1]";
  }


  if ($sys{'auto_os'} =~/win/i) {                 # win, apache with admin priv
          $r=&socket($e);
          #log
          if (($in{'l1'} ne "")  && ($ref ne "0")) {
               &log_add("$r ");
          }
          return ($r);
  }


  #opt parse execute
  ###################
  # ex pipe only with egrep ..

  $e=~s/^\s*sudo\s+//;                             # console command without sudo/pfexec
  $e=~s/^\s*pfexec\s+//;

  # Comstar + RSF1
  # use /opt/HAC/RSF-1/bin/stfmha instead of itadm and stmfadm  !! require more work or a dedicated Comstar HA menu

   # if ($e=~/^itadm|^stmfadm/) {
   #     if (-f "/var/web-gui/_log/ha-comstar.txt") {
   #           if (-f "/opt/HAC/RSF-1/bin/stmfha") {
   #              $e=~s/^itadm/\/opt\/HAC\/RSF-1\/bin\/stmfha/;
   #              $e=~s/^stmfadm/\/opt\/HAC\/RSF-1\/bin\/stmfha/;
   #           }
   #     }
   # }


  if ($sys{'auto_os_type'} =~/_n/i) {              # linux, nexenta at least v2 stores needed ls and chmod in /usr/sun/bin

       #### ls und chmod: sun version
       if ($e=~/^ls |^chmod /) {
         $e="/usr/sun/bin/$e";
       }
       $e=~s/\| /\| sudo /gs;                      # all pipes with sudo

  } else {

      if (($sys{'auto_os_type'} =~/_i/i) && ($e=~/^ls |^chmod /)) {          # ilumian use ls and chmod from /usr/bin
         $e="/usr/bin/$e";
      }

      $e=~s/\| /\| sudo /gs;                      # all pipes with sudo
  }


# sudo on openindiana missed /usr/sbin in path, so add it now manually, Linus add /sbin
if (!($e=~/^\//)) {                              # only if no path in executable   ex '/xx/yy/command '
    # read all files in /usr/sbin
    @t=();

    if ($sys{'auto_os'} eq "Linux") {
        &read_dir("/sbin","t","",__LINE__);
    } else {
        &read_dir("/usr/sbin","t","",__LINE__);
    }
    #

    $e.=" ";                                     # in case of simple commands
    foreach $t (@t)   {
       if ($t=~/\//) {  next; }                  # directory
       $t=~s/\+/\\+/g;                           # due to error about bonnie++
       $e=~s/^$t /$sys{'sbinpath'}$t /;          # add path first expression
       $e=~s/\| $t /\| $sys{'sbinpath'}$t /g     # add path after pipes
    }
}


  ## do it
  if ($err eq "n") {                             # no error
          $r=`sudo $e 2>/dev/null`;                          #
  } else {                                       # incl error
         $r=`sudo $e 2>&1`;
  }

  $r=~s/\s+$//s;                                 # no spaces or \n\r at end

  #log
  if (($in{'l1'} ne "")  && ($ref ne "0")) {
       &log_add("$r ");
  }

  # bug with sudo in omni 151017: output starts with Last Login
  if ($r=~/Last login:/s) {
       @t=();
       @t=split(/\n/,$r);
       my @new=();
       foreach my $t (@t) {
          if (!($t=~/Last login:/)) { push (@new,$t); }
       }
       $r=join("\n",@new);
  }

  return ($r);

}


#########
sub stop { print "Content-type: text/html\n\nstop:<hr>$_[0]"; exit; }
########

#############
# logwindows
#############
sub log_begin {
#
     #create these divs always

     $sys{'log_ok'}="1";  # log-add allowed

     #logwindow
     print <<EoF;
<!-- center-->
<div id='center_log' style='position: absolute; z-index:2000; top: 150px; left: 10px; padding:0px; margin:0px auto; width: 100%; text-align:center;'>

  <div id='mylog_close'>
    <div style='position:absolute; z-index: 2002; width: 870px; text-align:center; margin-top:2px'><a title="move me" href="#" onmousedown="dragStart(event, 'center_log')"><b>Monitor extension</b></a></div>
    <div style='position:absolute; z-index: 2001; width: 900px; text-align:right'>
       <a href='#' onclick=\"\$('#mylog_close').animate({height: 'hide', opacity: 'hide'}, 'fast');\$('.logwindow').animate({height: 'hide', opacity: 'hide'}, 'fast');\" title='close'><img src='/_doc/close.png' alt=''></a>
    </div>
  </div>

  <div id='mylog_cmd' class='logwindow'><br>Monitor-Extension: Command Log
  </div>

  <div id='minilog' style='position:fixed; bottom:0px; display:none; width: 100%; height: 45px; z-index: 20'> minilog<br>
  </div>
  <div id='miniover' style='position:fixed; bottom:0px; width: 100%; height: 6px; z-index: 21'> <br>
  </div>


EoF

     # minilog footer only on mouseover
     if ($in{'l1'} ne "") {
        print <<EoF;
  <script type="text/javascript">
    \$("#minilog").animate({height: 'show', opacity: 'show'}, 'slow');
  </script>
EoF
    }



    # show log in editmode with monitor ext
    if ((-f "/tmp/nappit/napp-it_edit_on") && ($in{'l1'} ne "")) {
      print <<EoF;
  <!-- add hash logs -->
  <div id='mylog_in' class='logwindow'><br>Hash-values \%in:<hr>
  </div>

  <div id='mylog_sys' class='logwindow'><br>Hash-values \%sys:<hr>
  </div>

  <div id='mylog_disk' class='logwindow'><br>Hash-values \%disk:<hr>
  </div>

  <div id='mylog_zfs' class='logwindow'><br>Hash-values \%zfs:<hr>
  </div>

  <div id='mylog_current' class='logwindow'><br>Hash-values \%current:<hr>
  </div>

 <div id='mylog_txt' class='logwindow'><br>Hash-values \%txt:<hr>
 </div>

EoF
    }
    print "</div><!-- end center -->\n\n"; # end center

    my $ok="0";
    # (show Log when edit=on + moni ext)
    # show always in menu disk, disk-smart and sas on these submenus

    $t=$txt{$sys{'menu'}};
    if ($t=~/=/) {
          $t=~s/.*=//;                                          # ex disk details=AL
    } else {
          $t="";
    }

    # on with extension c or m
    if (-f "/tmp/napp-it_edit_on") {
        if (($t=~/L/) && ($cfg{'ver'}=~/c|m/) ) { $ok="1" };   #editmode          # menu.txt contains L (show full log during run)

        if ($ok) {
        print <<EoF;
        <script type="text/javascript">
        \$("#mylog_close").animate({height: 'show', opacity: 'show'}, 'fast');
        \$("#mylog_cmd").animate({height: 'show', opacity: 'show'}, 'fast');
        </script>
EoF
        }
    }
}

#############
sub log_add_mini {
   my $add=$_[0];                           # minilofg only

   if (!($sys{'log_ok'})) { return; }      # do not log prior log_begin

   my @p0=caller(0);
   my @p1=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask ex $p1[3] l$p1[2]
   my @p2=caller(2);
   my @p3=caller(3);
   my @p4=caller(4);

                  my $t="$p0[1] $p0[3] $p0[2]";        $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p1[2] > 0) { $t.=" <- $p1[1] $p1[3] $p1[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p2[2] > 0) { $t.=" <- $p2[1] $p2[3] $p2[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p3[2] > 0) { $t.=" <- $p3[1] $p3[3] $p3[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p4[2] > 0) { $t.=" <- $p4[1] $p4[3] $p4[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   $t=~s/main::/\&/g;

   $add="$t $add";

   $add=~s/>/\&gt;/gs;                     # avoid active html
   $add=~s/</\&lt;/gs;

   $add=~s/\&lt;hr\&gt;/<hr>/gs;           # keep <hr>

   $add=~s/\r//gs;

   $add=~s/\\/\\\\/gs;
   $add=~s/\//\\\//gs;
   $add=~s/\|/\\\|/gs;

   $add=~s/\'/\\'/gs;
   $add=~s/\"/\\"/gs;

   $add=~s/\n/<br>/gs;
   print <<EoF;

   <script type="text/javascript">
     \$("#minilog").append(" $add");
     \$("#minilog").animate({ scrollTop: \$("#minilog").prop("scrollHeight") }, 10);
   </script>
EoF
#&log_scroll;     $("#myDiv").attr({ scrollTop: $("#myDiv").attr("scrollHeight") });
}

#############
sub log_add {
   my $add=$_[0];

   if (!($sys{'log_ok'})) { return; }      # do not log prior log_begin

   my @p0=caller(0);
   my @p1=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask ex $p1[3] l$p1[2]
   my @p2=caller(2);
   my @p3=caller(3);
   my @p4=caller(4);

                  my $t="$p0[1] $p0[3] $p0[2]";        $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p1[2] > 0) { $t.=" <- $p1[1] $p1[3] $p1[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p2[2] > 0) { $t.=" <- $p2[1] $p2[3] $p2[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p3[2] > 0) { $t.=" <- $p3[1] $p3[3] $p3[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p4[2] > 0) { $t.=" <- $p4[1] $p4[3] $p4[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   $t=~s/main::/\&/g;

   $add="--- $t ---\n\n $add";

   $add=~s/>/\&gt;/gs;                     # avoid active html
   $add=~s/</\&lt;/gs;

   $add=~s/\&lt;hr\&gt;/<hr>/gs;           # keep <hr>

   $add=~s/\r//gs;

   $add=~s/\\/\\\\/gs;
   $add=~s/\//\\\//gs;
   $add=~s/\|/\\\|/gs;

   $add=~s/\'/\\'/gs;
   $add=~s/\"/\\"/gs;

   $add=~s/\n/<br>/gs;
   print <<EoF;

   <script type="text/javascript">
     \$("#mylog_cmd").append("<hr>$add<br>");
     \$("#mylog_cmd").animate({ scrollTop: \$("#mylog_cmd").prop("scrollHeight") }, 50);

     \$("#minilog").append("$add");
     \$("#minilog").animate({ scrollTop: \$("#minilog").prop("scrollHeight") }, 10);
   </script>
EoF
#&log_scroll;     $("#myDiv").attr({ scrollTop: $("#myDiv").attr("scrollHeight") });
}

########### called at page </body>
sub log_end {

   my ($in,$sys,$disk,$zfs,$current,$txt);
   if ((-f "/tmp/nappit/napp-it_edit_on") && ($in{'l1'} ne "")) {       # add hashinfo
         $in=&return_hash(%in);

         $sys=&return_hash(%sys);
         $disk=&return_hash(%disk);
         $zfs=&return_hash(%zfs);
         $current=&return_hash(%current);
         $txt=&return_hash(%txt);

         # add agents, cpu and ram info
         my $stat="\ncurrent prstat:\n";
         if ($sys{'auto_os'} ne "Linux")   {
          # $stat.=`sudo $sys{'binpath'}prstat -c 1 1`;
         }

         $stat.="\n\ncurrent ps axw\n";
         $stat.=`sudo $sys{'binpath'}ps aw | grep web-gui | grep var`;

         $stat=~s/\n/<br>/gs;


         print <<EoF;
  <script type="text/javascript">
     \$("#mylog_cmd").append("$stat");
     \$("#mylog_in").append("$in");
     \$("#mylog_sys").append("$sys");
     \$("#mylog_disk").append("$disk");
     \$("#mylog_zfs").append("$zfs");
     \$("#mylog_current").append("$current");
     \$("#mylog_txt").append("$txt");
  </script>

EoF
   }
   &log_add("\n\n - End logging, details see monitor extension + topmenu edit-\n");

   sleep(1);
   print <<EoF;
  <script type="text/javascript">
     \$("#minilog").val("more >> topmenu Edit + monitor extension.");

     \$("#mylog_close").animate({height: 'hide', opacity: 'hide'}, 'fast');
     \$("#mylog_cmd").animate({height: 'hide', opacity: 'hide'}, 'fast');
     \$("#minilog").animate({opacity: 'hide'}, 'slow');

     \$("#miniover").mouseenter(function() {
         \$("#minilog").animate({height: 'show', opacity: 'show'}, 'slow');
     });

     \$("#miniover").mouseleave(function() {
         \$("#minilog").animate({height: 'hide', opacity: 'hide'}, 'slow');
     });

  </script>
EoF
}

######################## parameter: list, key, parameters ex "$r","LU Name: ","a,b,c"
sub multi_list2table {
#######################
# convert a list to a table
# if parameter ends on a ! it is displayed as alink with text and prop of parameter
#  link is defined in $current{'mlt_link'}

my (@v,$v,@l,$l,$t,$r,@p,$p,%p,$ref,$k);

my $list=$_[0];
my $key =$_[1];
my $par =$_[2];
   $par=~s/^,+//g;
   $par=~s/^,+//g;
my @par=split(/,/,$par);
   #

   if ($key =~/\!$/) {                   # Key:! no key, key value is left, other with spaces, !! start with second line
     $t=$key;
     $t=~s/\!+$//;
     @l=split(/\n/,$list);

#&print_array(@l);
     for (my $i=0;$i<@l;$i++) {
          $l=$l[$i];
          if ($key =~/\!\!$/) {
             if ($i==0) { next; }      # skip first line
          }
          if (!($l =~/^\s+/)) {        # key value
             $l=~s/\s.*//;             # first word
             $l="$t$l";                # set key ex "LU: "
          }
          push (@v,$l);
     }
     $list=join("\n",@v);
     $key=~s/\!+$//;
   }


#&mess($list);
#&print_array(@v);

   #
   @v=();
   @v=split(/$key/,$list);          #Data sets

#&print_array(@v);

   #LU Name: 600144F0904FAB650000576423870005
   #    Operational Status : Online
   #    Provider Name      : sbd
   #    View Entry Count   : 1
   #     View-entry 0      : Host group 'all' Target group 'all' LUN '4'

   #
   #LU Name: 600144F0904FAB650000576423970006
   #    Operational Status : Online
   #    Provider Name      : sbd
   #    View Entry Count   : 1
   #..

   $key=~s/: *//;  # keyname only

   $r="<b>$key<b>\t<b>$par"; $r.="</b>\n";
   $r=~s/,/<\/b>\t<b>/g;                       # headline with par bold
   $r=~s/!//g;                                 # link marker

   for (my $i=1;$i<@v;$i++) {                  # data sets, each key ex LU:

        %p=();                                 # new Data

        $ref=$v[$i];
        $ref=~s/\n.*//s;
        $r.="$ref\t";                          # ex 600144F0904FAB650000576423970006

        $v=$v[$i];

        @l=();                                 # lines of each v set
        @l=split(/\n/,$v);
        foreach $l (@l) {
           $l=~s/^\s+//;                       # no leading spaces
           $l=~s/\s*:\s+/: /g;                 # key (opt value):value

           @p=();
           @p=split(/: /,$l);
           $p[0]=~s/^\s+//;
           $p[0]=~s/\s+$//;
           $p[1]=~s/\s+//;

           foreach my $par (@par) {
               $k=$par; $k=~s/!$//;
               if ($p[0]=~/^$k/i) {
                   $p[0]=~s/^$k//i;

                   if ($p[0] ne "") {  # data after par
                      $p[1]="($p[0]) $p[1]";       # ex View-entry 0      : with View-entry as par
                   }
                   if ($p{$par} eq "") {
                      $p{$par}="$p[1]";
                   } else {   # multiple values
                      $p{$par}.="<br>\&nbsp;$p[1]";
                   }
               }
           }
        }

#&print_array(@l);

        # add dataline with or without values
        foreach my $par (@par) {
            if (($par=~/!$/) && ($current{'mlt_link'} ne "")) {          # link
                $t=$current{'mlt_link'};
                $t=~s/=text=/$p{$par}/;

                $k=$par;
                $k=~s/!$//;
                $t=~s/=prop=/$k/;

                $t=~s/=list=/$ref/;
                $r.="$t\t";
            } else {
                $r.="$p{$par}\t";
            }
        }
        $r=~s/\t+$//s;
        $r.="\n";
   }
   return ($r);
}


########## Fehlermeldung mit Fehler in err.log speichen und opt drucker via print
sub log {
##########   options m=monitor, l=joblog, p=pront, r=remotelog, a=alert
  my $mess=$_[0];

   my @p0=caller(0);
   my @p1=caller(1); # parent:  package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask ex $p1[3] l$p1[2]
   my @p2=caller(2);
   my @p3=caller(3);
   my @p4=caller(4);
   my @data;

                  my $t="$p0[1] $p0[3] $p0[2]";        $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p1[2] > 0) { $t.=" <- $p1[1] $p1[3] $p1[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p2[2] > 0) { $t.=" <- $p2[1] $p2[3] $p2[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p3[2] > 0) { $t.=" <- $p3[1] $p3[3] $p3[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   if ($p4[2] > 0) { $t.=" <- $p4[1] $p4[3] $p4[2]"; } $t=~s/\/.*zfsos\// /; $t=~s/\/.*scripts\// /;
   $t=~s/main::/\&/g;

  &getdate;      # update
  $mess=~s/\|/\| $sys{'hour'}.$sys{'min'}\:$sys{'sec'} $t<hr>/;     # add time and caller

  my $opt=$_[1];

  my $err="C:\\xampp\napp-it\\_log/job.log";              # err and mess
  my $curr="C:\xampp\\napp-it\\_log\\current_job.log";
  my $rem="C:\xampp\\napp-it\\_log\\remote.log";           # log remote requests

  my $moni="C:\xampp\\napp-it\\_log\\monitor.log";         # log monitor
  if (-M $moni > 1)  { unlink($moni); &exe("touch $moni"); &exe("chmod 666 $moni");   }                 # needed for remote moni log

  my $pre="## $sys{'date_all'}";
  if ($opt =~/a/) { $pre="<font style='color:red'>###</font> "; }


  my ($t,$l2,$l1,$c);
  if (-f $err) {
       $t = (-s "$err");                           # size max 100k
       if ($t >= 100000) {
           #$c=`cat $err`;

           open (F,"<$err");
             flock(F, LOCK_EX);
             @data=<F>;
           close (F);
           $c=join("",@data);


           $c=substr($c,50000);
           open (PF, ">$err");   flock(PF, LOCK_EX); print PF ".... $c";  close (PF);
       }
  }
  if (-f $curr) {
       $t = (-s "$curr");                           # size max 100k
       if ($t >= 100000) {
           #$c=`cat $curr`;
           open (F,"<$curr");
             flock(F, LOCK_EX);
             @data=<F>;
           close (F);
           $c=join("",@data);

           $c=substr($c,50000);
           open (PF, ">$curr");  flock(PF, LOCK_EX);  print PF ".... $c";  close (PF);
       }
  }

  $mess=~s/\t/ /gs;
  # logfile
  if ($opt =~/l/) {
           $t="$mess";
           $t=~s/\n/<br>/gs;

           open (PF, ">>$err");  flock(PF, LOCK_EX); print PF "$pre $sys{'date'}\t$t\n";  close (PF);
           open (PF, ">>$curr"); flock(PF, LOCK_EX); print PF "$pre $sys{'date'}\t$t\n";  close (PF);
  }

  # monitor (menu jobs monitor), newest first, 50 entries
  if ($opt =~/m/) {
      my @data;
      $t=$mess; $t=~s/\n/<br>/gs;
      if ($opt =~/a/) { $t="<font style='color:red'>###</font> $t "; }   # alert: red
      open (F,"<$moni");
            flock(F, LOCK_EX);
            @data=<F>;
      close (F);
      unshift(@data,"$t\n");

      if (@data > 50) {
             $t=@data-50;
             splice(@data,50,$t);
      }
      $t=join("",@data);
      open (PF, ">$moni");
          flock(PF, LOCK_EX);
          print PF "$t";
      close (PF);
  }

  #remote action
  if ($opt =~/r/) {
      @data=();
      if (-f $rem) {
         open (F,"<$rem");
               flock(F, LOCK_EX);
               @data=<F>;
         close (F);
      }

      $t=$mess; $t=~s/\n/<br>/gs;
      unshift(@data,"$t\n");

      if (@data > 100) {
             $t=@data-100;
             splice(@data,100,$t);
      }

      $t=join("",@data);
      open (PF, ">$rem");
          flock(PF, LOCK_EX);
          print PF "$t";
      close (PF);
  }

# print to shell
  if ($opt =~/p/) {
    if ($opt =~/a/) {
       $mess="### \n$mess";
     } else {
       $mess="--- \n$mess";
     }

    if (length($mess) > 320) {
          $mess=substr($mess,0,320); $mess.=" ..\n";
    }
    print "$mess\n\n";
  }

} #eos
#########

################
sub messages_box {
################
  my $txt=$_[0];
  print <<EoF;
<div id="messages">
$txt
</div>
EoF
}

################
sub messages_fadein {
################
  print <<EoF;
   <script>
   \$("#messages").fadeIn("fast");
   </script>
EoF
}

################
sub messages_fadeout {
################
  print <<EoF;
   <script>
   \$("#messages").fadeOut("fast");
   </script>
EoF
}

################  # html formatted output or "pn" ex p1p2 parameter for unformatted with \n
sub get_caller {
################
  my $p;

  my @p0=caller(0); # me                  # package, filename, line, subroutine, hasargs, wantarray, evaltext, is_require, hints, hitmask
  my @p1=caller(1); # parent1
  my @p2=caller(2); # parent2
  my @p3=caller(3); # parent3
  my @p4=caller(4); # parent4

  #called via cli
  if ($in{'id'} eq "") {
      $p= "\ncaller : $p1[0], $p1[1], line $p1[2]\n";
      return ($p);
  }

  $p1[1]=~s/\/var.*\/zfsos//;  # reduce path
  $p2[1]=~s/\/var.*\/zfsos//;
  $p3[1]=~s/\/var.*\/zfsos//;
  $p4[1]=~s/\/var.*\/zfsos//;


  # only part
  &log_add($_[0]);

  if ($_[0]=~/p\d/) {
     if (($_[0]=~/p4/) && ($p4[2] ne "")) { $p="$p4[0], $p4[1], line $p4[2]\n"; }
     if (($_[0]=~/p3/) && ($p3[2] ne "")) { $p="$p3[0], $p3[1], line $p3[2]\n"; }
     if (($_[0]=~/p2/) && ($p2[2] ne "")) { $p="$p2[0], $p2[1], line $p2[2]\n"; }
     if (($_[0]=~/p1/) && ($p1[2] ne "")) { $p="$p1[0], $p1[1], line $p1[2]\n"; }
     return($p);
  }

  #html
  my $c="style='padding-right:10px'";
  my $c1="style='padding-right:10px; border-bottom:1px dashed'";

  $p="<table><tr><td $c1> package </td><td $c1> file </td><td $c1> subroutine </td><td $c1> line </td></tr>\n";
  if ($p4[2] ne "") { $p.="<tr><td $c>$p4[0] </td><td $c>$p4[1] </td><td $c>$p4[3] </td><td>$p4[2]</td></tr>\n";  }
  if ($p3[2] ne "") { $p.="<tr><td $c>$p3[0] </td><td $c>$p3[1] </td><td $c>$p3[3] </td><td>$p3[2]</td></tr>\n";  }
  if ($p2[2] ne "") { $p.="<tr><td $c>$p2[0] </td><td $c>$p2[1] </td><td $c>$p2[3] </td><td>$p2[2]</td></tr>\n";  }
  if ($p1[2] ne "") { $p.="<tr><td $c>$p1[0] </td><td $c>$p1[1] </td><td $c>$p1[3] </td><td>$p1[2]</td></tr>\n";  }

  $p.="</table>\n";
  return("$p");
}


########## Fehlermeldung ausgeben
sub mess {
##########
  my $abs=$_[2];           # abs - >position absolure
  if ($abs=~/abs/) {
      $abs="position: absolute; z-index:3000; width: 99%; ";
  }

  if ($_[1] ne "") {
       print "<br><br><b>$_[1]</b><hr>$_[0]\n";
       exit;
  }

  print "<div style=\"background-color: #d77b03;$abs\">\n";
  print "<!--nows--><pre><br><h1 style=\"color:white\">&nbsp;oops...</h1>\n<FONT SIZE=\"2\" FACE=\"Courier New,Courier\">\n<div style=\"margin-left: 170px; color:white; white-space: pre\">\n";
  print "$_[0]<br>\n";
  print "\n<a href=\"javascript\:history.back()\"><br><font color=\"#ffffff\"><b><u>$txt{'back'}</u></b></font><\/a></div></font>\n<br><br></div></pre>\n";


  $t=&get_caller('p1');

  &log_add($t);
  $t.="<br>";
  $t.=&get_caller('p2');
  $t.="<br>";
  $t.=&get_caller('p3');

  print "<div style='color:#666666; font-size:9px; text-align:right; width:99%'>$t</div>";

  &log_end;
  exit;

  ##&error("$_[0]","mess");
}
#########


#########
sub error {
##########
 # print "Content-type: text/html\n\n";
 # print "<HTML>\n";
 # print "<HEAD>\n";
 # print "<TITLE>Hinweis/ message</TITLE>\n";
 # print "</HEAD>\n";

 my $t=$_[0];
 $t=~s/\n/<br>/gs;


  if ($_[1] eq "mess") {
    print "<div style=\"background-color: #d77b03;z-index:2000\">\n";
  } else {
    print "<div style=\"background-color: #ba2f28;z-index:2000\">\n";
  }
  print "<!--nows--><br><h1 style=\"color:white\">&nbsp;oops...</h1>\n<FONT SIZE=\"2\" FACE=\"Courier New,Courier\">\n<div style=\"margin-left: 170px; color:white; white-space: pre\">\n";
  if ($_[1] ne "mess") {
    print "$txt{'mess_err'}\<br><br><hr>$t<hr><br><br>\n";
  } else {
    print "$t<br>\n";
  }
  print "\n<a href=\"javascript\:history.back()\"><br><font color=\"#ffffff\"><b><u>$txt{'back'}</u></b></font><\/a></div></font>\n<br><br></div><br>\n";

  print &get_caller;
  &log_end;

  exit;
}


##############
sub netmask2n {
##############
   my $mask = $_[0]; # '255.255.255.0' - > /24;
   use Socket;
   my $t=unpack('B*', inet_aton($mask));
   $t=~s/0//g;
   $t=length($t);
   return($t);
}

##############
sub n2netmask {
##############
    my $netbit = $_[0];                 # 24 or 1.1.1.1/24-> '255.255.255.0';
    $netbit=~s/.*\///;

    # $netbit="10";

    my $_bit         = ( 2 ** (32 - $netbit) ) - 1;
    my ($full_mask)  = unpack( "N", pack( "C4", split(/\./, '255.255.255.255') ) );
    my $netmask      = join( '.', unpack( "C4", pack( "N", ( $full_mask ^ $_bit ) ) ) );

    return $netmask;   # 255.255.255.0

}

############
sub con2html { #console to html
   $_[0]=~s/\n|\r/\n<br>/gs;
   $_[0]=~s/ /\&nbsp\;/gs;
   return ($_[0]);
}


sub ltrim {
#########
 $_[0]=~s/^\s+//;
 return ($_[0]);
}


# entfernt fllende Leerzeichen
sub rtrim {
 $_[0]=~s/\s+$//;
 return ($_[0]);
}

# entfernt fhrende und fllende Leerzeichen
sub alltrim {
  my $tmp=$_[0];
  $tmp=&ltrim($tmp);
  $tmp=&rtrim($tmp);
  return ($tmp);
}


sub t2f {
#########   text to form

        my $x=$_[0];
        $x=~s///gs;
        $x=~s/<n>/\n/gs;
        if ($_[1] eq "text") { $x=~s/\n/\n/gs; }

        $x=~s/\t/  &rarr;  /gs;                       #tab

        $x=~s/\&lt\;/&lt;\'/gs;                    #entitities
        $x=~s/\&gt\;/&gt;\'/gs;
        $x=~s/\&quot\;/&quot;\'/gs;

        $x=~s/</&lt;/gs;                           #char
        $x=~s/>/&gt;/gs;

        $x=~s/\"/\&quot\;/gs;
        $x=~s/\&nbsp\;/\&uml\;/gs;
        return ($x);
}


sub f2t {
######### form 2 text
        my $x=$_[0];
        $x=~s/\n/<n>/gs;
        $x=~s/\r//gs;

        $x=~s/ *\&rarr\; */\t/gs;                   #tab
        $x=~s/ *\&\#8594\; */\t/gs;                 #tab

        $x=~s/\&lt;/</gs;                           #char
        $x=~s/\&gt;/>/gs;
        $x=~s/\&quot\;/\"/gs;

        $x=~s/<\'/&lt;/gs;                         #entitities
        $x=~s/>\'/&gt;/gs;
        $x=~s/\"\'/&quot;/gs;

        $x=~s/\//gs;
        $x=~s/\/\&sect\;/gs;
        $x=~s/\&uml\;/\&nbsp\;/gs;
        ## $x=~s/\|/\&\#124\;/gs;
        return ($x);
}

# name nach filename
sub name2file {
#############
 $_[0]=~s//oe/g;
 $_[0]=~s//ue/g;
 $_[0]=~s//ae/g;
 $_[0]=~s//Oe/g;
 $_[0]=~s//Ue/g;
 $_[0]=~s//Ae/g;
 $_[0]=~s/[^a-zA-Z0-9\~\_\=\!\\%\@\-\.\,\ \\\/:]//gs;           # win c;
 $_[0]=~s/^\.+//g;

 return ($_[0]);
}

#############
sub name2path {
###############
    my $p=$_[0];
    if ($p=~/ /) {
      if (!($p=~/\'/)) {
         $p="\'"."$p"."\'";              # if path contains blank
      }
    }
    $p=~s/\/+/\//g;
    return ($p);
}


#kein Pipe-Zeichen | >> ~I~  >> ~3~
sub nopipe {
############
 $_[0]=~s/\|/\~I\~/g;
 return ($_[0]);
}

# String nach html
###########
sub txt2htm {
 local (@zeilen,@tmp,$tmp,$i,$z);

 @zeilen=split(/<n>/,$_[0]);
 foreach $z (@zeilen) {

         if ($z=~/^\s*<meta/) {next;}
         if ($z=~/^\s*<title/) {next;}

         $z=~s//\&auml\;/g;
         $z=~s//\&ouml\;/g;
         $z=~s//\&uuml\;/g;
         $z=~s//\&Auml\;/g;
         $z=~s//\&Ouml\;/g;
         $z=~s//\&Uuml\;/s;

         $z=~s/\/\&ldquo\;/g;           #Satzzeichen
         $z=~s/\/\&bdquo\;/g;
         $z=~s/\/\&quot\;/g;
         $z=~s/\/\&quot\;/g;
         $z=~s/\/\&quot\;/g;
         $z=~s/\/\&quot\;/g;
         $z=~s/\/-/g;

 } #/z


 $_[0]=join("<n>",@zeilen);
 return ($_[0]);
}

############
sub htmlcode {        #quellcode lesbarer machen parameter: arrayname      par: array,n (n=bc: textarea nicht einruecken)
############

 ##return;
 my ($tmp,$i,$t,$n,@werte,$z,$h,$xml,$test);
 local @tmp;
 if (!(@{$_[0]} > 0)) { return; }

 $xml="0";

 #div immer in neuer zeile
 $h="1";
 foreach $tmp (@{$_[0]}) {
           if ($tmp=~/<body/) { $h="0"; }
           if ($h) { next; }

           $tmp=~s/\t//gs;                   #keine tabs
           if ($tmp=~/<div.+<div/) {
                      $tmp=~s/<div/\n<div/g;
           }
           $tmp=~s/<textarea/\n<textarea/gis;                     #immer nl
           $tmp=~s/<\/textarea>/<\/textarea>\n/gis;               #immer nl

 }
 #&show_array($_[0]);
 $tmp=join("",@{$_[0]});



 #allg
 $tmp=~s/<b><\/b>//gs;
 $tmp=~s/<strong><\/strong>//gs;


 #img mit alt und border, & als linkpar
 @tmp=split(/</,$tmp); #tags
#if ($tmp=~/_grau/)   { &error("stophtmlcode<br> $tmp");}

 for ($i=1;$i<@tmp;$i++) {

         if ($tmp[$i]=~/^!DOCTYPE/i) {
                 if ($tmp[$i]=~/XHTML/i) {
                        $xml="1";
                }
                next;
        }

        if ($xml) {
                $tmp[$i]=~s/^br>/br \/>/;
                  if ($tmp[$i]=~/^hr>/)        { $tmp[$i]=~s/>/ \/>/; }                           # <hr />
                  if ($tmp[$i]=~/^meta/)        {
                   if (!($tmp[$i]=~/ \/>/)) {
                        $tmp[$i]=~s/>/ \/>/;                                                    # <meta />
                   }
                    }
        }  else {
                        $tmp[$i]=~s/ \/\>/>/gs;
        }


        #& in href als &amp;
        if ($tmp[$i]=~/^a /i) {
            @werte=();
            @werte=split(/\">/,$tmp[$i]);                        # bis 3.8.08 />/ -> bug falls in link z.b. weiter >>

            $werte[0]=~s/href=\#\"/href=\"\#\"/g;                 #a href=#" fehler
            $werte[0]=~s/\&/\&amp\;/g;
            $werte[1].="";
            $tmp[$i]=join("\">",@werte);
            next;
        }

        if (!($tmp[$i]=~/^img/i)) { next;}

        if (!($tmp[$i]=~/alt\=/i))             { $tmp[$i]=~s/^img /img alt=\"\" /s; }
        if (!($tmp[$i]=~/border=\"/i))         { $tmp[$i]=~s/^img /img border=\"0\" /s; }

        if (!($tmp[$i]=~/\&/))                 { $tmp[$i]=~s/\&/\&amp\;/g; }                     # kein & in img
        if ($xml) {
          if (!($tmp[$i]=~/ \/>/))        { $tmp[$i]=~s/>/ \/>/; }                           # <img.. />
        }

 }

 $tmp=join("<",@tmp);


 $tmp=~s/\r//gs;
 $tmp=~s/\n/<n>/gs;
 $tmp=~s/<n><br/<br/gs;
 $tmp=~s/<n>/\n/gs;

 @{$_[0]}=();
 @{$_[0]}=split(/\n/,$tmp);

 #blockelement immer in neier zeile ausser in textarea
 $n="1";
 foreach $tmp (@{$_[0]}) {
            if ($tmp=~/<textarea/)         { $n="0"; }

              if ($n) {
                 $tmp=~s/<select/\n<select/gis;                    #immer nl
                 $tmp=~s/<\/select/\n<\/select/gis;                #immer nl

                 $tmp=~s/<div/\n<div/gis;                         #immer nl
                 $tmp=~s/<\/div/\n<\/div/gis;

                 $tmp=~s/<table/\n<table/gis;
                 $tmp=~s/<\/table/\n<\/table/gis;

                 $tmp=~s/<tr/\n<tr/gis;
                 $tmp=~s/<\/tr/\n<\/tr/gis;

                 $tmp=~s/<td/\n<td/gis;
                 $tmp=~s/<\/td/\n<\/td/gis;

                 $tmp=~s/<th/\n<th/gis;
                  $tmp=~s/<\/th/\n<\/th/gis;
           }

           if ($tmp=~/<\/textarea/)         { $n="1"; }
 }
 $tmp=join("\n",@{$_[0]});
 $tmp=~s/\n<img/<img/gs;              #damit kein Abstand bei mehreren pic nebeneinander
 @{$_[0]}=split(/\n/,$tmp);

 $t="";
 $i="0";
 $h="1";
 $n="1";
 foreach $tmp (@{$_[0]}) {
   $tmp.="\n";
   if ($tmp eq "") {next; }

   if ($tmp=~/<body/) { $h="0"; }
   if ($h) { next; }

   $tmp=~s/^ +/ /;                        #keine "   "

   #textarea immer linksb.
   if ($tmp=~/<textarea/)                 { $n="0"; }
   if ($n) {

           if ($tmp=~/<div|<table|<tr|<td|<th|<script|^ *<style|<ul|<ol|<form|<select/i)         {

                   $t="\t" x $i;
                $tmp="$t$tmp";
                if (!($tmp=~/<\/div|<\/table|<\/tr|<\/td|<\/th|<\/script|^ *<\/style|<\/ul|<\/ol|<\/form|<\/select/i))         {
                  $i++;
                }
                next;
           }

           if ($tmp=~/<\/div|<\/table|<\/tr|<\/td|<\/th|<\/script|^ *<\/style|<\/ul|<\/ol|<\/form|<\/select/i)         {

                    $i--;
                if ($i<1) {$i="0";}
                $t="\t" x $i;
                $tmp="$t$tmp";

                next;
           }

                $t="\t" x $i;
                $tmp="$t$tmp";

   }
   if ($tmp=~/<\/textarea/)         { $n="1"; }

 }
 #&show_array($_[0]);

}

sub htm2txt {
###########
# html nach name

 $_[0]=~s/\&amp\;/\&/g;
 $_[0]=~s/\&auml\;//g;
 $_[0]=~s/\&ouml\;//g;
 $_[0]=~s/\&uuml\;//g;
 $_[0]=~s/\&Auml\;//g;
 $_[0]=~s/\&Ouml\;//g;
 $_[0]=~s/\&Uuml\;//g;
 $_[0]=~s/\&quot\;/\"/g;
 $_[0]=~s/\&szlig\;//g;
 $_[0]=~s/\&sect\;/\/g;
 return ($_[0]);
}



  sub show_hash {      #ex &show_hash("name",__LINE__);                         # hash immer als referenz
############## sub hash anzeigen (hashname,zeilennr)
  my ($k,$v,$tmp,@tmp,$out);
  my $t="$_[1]$_[2]";
  print "Content-type: text/html\n\n<br><br><br><tt>-show_hash $t<hr>\n";
  while (($k,$v) = each %{$_[0]}) {
        $k=~s/\n/\\n/g;    $v=~s/\n/\\n/g;
        $v=~s/</\&lt\;/g;  $v=~s/>/\&gt\;/g;
        if ($t=~/p/)  { $v=~s/\r/\n/gs; $v=~s/\n/&para;\n/gs; }                 # zeige \n als Paragraph
        push (@tmp,"$k =\&gt\;$v\&lt\;=<br>\n");
  }
  @tmp=sort(@tmp);
  foreach $tmp (@tmp) {
     print $tmp;
  }
  print "<hr></tt>\n";
  exit;
}

# hash as string for logwindow
sub return_hash {
################
 my %h = @_;
 my ($k,$v,@tmp,$s);
 $s=<<EoF;
<b>Global Hashvalues for current (sub)menu item.</b> The Hash is build in three stages:<br>
Stage 1: Libraries like get-disk.pl are preloaded in some menus or on demand and delivers basic information<br>
Stage 2: These infos are processed within library function or menu actions<hr><br>
EoF

 $s.= "<table width='900px'>";
 while (($k,$v) = each %h) {
     $v=~s/\n/\\n/gs;
     $v=~s/<br>/\&lt;br&gt;/gs;
     $v=~s/\t/\t\&rarr;/gs;         # arrow right
     push (@tmp,"<tr><td valign='top' width='300px' style='border-bottom: 1px dashed'>$k</td><td nowrap valign='top' style='border-bottom: 1px dashed; padding-bottom: 5px'> -&gt; </td><td valign='top' style='border-bottom: 1px dashed'>\'$v\'</td></tr>");
 }
 @tmp=sort(@tmp);
 $w=join("\n",@tmp);
 $s.= "$w\n</table>";

 $s=~s/\"/\\"/gs;
 $s=~s/\n|\r//gs;

 return ($s);

}

###############
sub log_hash {
    my %h = @_;
    my ($k,$v,@tmp,$w);
    while (($k,$v) = each %h) {
       push(@tmp,"$k -> \'$v\'");
    }
    @tmp=sort(@tmp);
    $w=join("\n",@tmp);
    return($w);
}

###############
sub print_hash { #parameter %
###############
 my %h = @_;
 my ($k,$v,@tmp,$w,$a);

 if ($in{'id'} eq "") {
      while (($k,$v) = each %h) {
           $w=$k." "."." x 40;
           $w=substr($w,0,41);
           push (@tmp,"$w -> \'$v\' \n");
      }
      @tmp=sort(@tmp);
      print "<pre>\&print_hash\n\n";
      foreach my $t (@tmp) {
        print $t;
      }
      print &get_caller;
      print "\n</pre>";
      exit;

 } else {

       print "<p class='tt' style='padding-top:80px'><b>\&print_hash()</b><hr>\n<table width='1200px'>\n";
       while (($k,$v) = each %h) {
            $v=~s/\</\&lt;/gs;
            $v=~s/\>/\&gt;/gs;
            $v=~s/\n/\\n/gs;
            $v=~s/<br>/\&lt;br&gt;/gs;
            $v=~s/\t/\t\&rarr;/gs;         # arrow right
            # print "$k \t-> \'$v\'\n";
            push (@tmp,"<tr><td valign='top' width='300px' style='border-bottom: 1px dashed'>$k</td><td nowrap valign='top' style='border-bottom: 1px dashed; padding-bottom: 5px'> -&gt; </td><td valign='top' style='border-bottom: 1px dashed'>\'$v\'</td></tr>");
       }
       @tmp=sort(@tmp);
       $w=join("\n",@tmp);

       print "$w\n</table><br>";
       print &get_caller;
       print "<br><br><br><br>";

       &log_end;
       exit;
 }

}

#############
sub print_var { #parameter $
###############
   my $t=$_[0];
   my $a;

   #cli
   if ($in{'id'} eq "") {
        print "\nvar -> \'$t\' \n";
        print &get_caller;
        exit;
   }

   $t=~s/\</\&lt;/gs;
   $t=~s/\>/\&gt;/gs;
   $t=~s/\n/\\n/gs;
   $t=~s/<br>/\&lt;br&gt;/gs;
   $t=~s/\t/\&rarr;/gs;         # arrow right
   print "<!--nows--><p style='padding-top:80px'><tt>\&print_var(\$var) - > \'$t\' \n</tt></pre><br><br><hr>";
   print &get_caller;
   print "<br><br><br><br>";

   &log_end;
   exit;
}

###############
sub print_array { #parameter @
###############
  my @t=@_;
  my $t;
  # @t=sort(@t); # no

  if ($in{'id'} eq "") {
      print "<pre>\&print_array\n\n";

      for (my $i=0;$i<@t; $i++) {
           print "[$i] -> '$t[$i]\'\n";
      }
      print "\n</pre>";
      print &get_caller;
      exit;

   } else {
      #
      print "<!--nows--><pre class='tt' style='padding-top:80px'>sub print_array<hr>\n\n";
      for (my $i=0;$i<@t; $i++) {
          $t[$i]=~s/\</\&lt;/gs;
          $t[$i]=~s/\>/\&gt;/gs;
          $t[$i]=~s/\t/\t\&rarr;/gs;         # arrow right
          $t[$i]=~s/\n/\\n/gs;
          $t[$i]=~s/<br>/\&lt;br&gt;/gs;
          print "[$i] -> '$t[$i]\'\n";
      }
      print "\n</pre><br>";
      print &get_caller;
      print "<br><br><br><br>";
      &log_end;
      exit;
   }
}



sub show_array {
############## array anzeigen (array(name),zeilennr,opt                         #array immer als referenz
  my ($i,$out);  my $t="$_[1]$_[2]";
  print "Content-type: text/html\n\n<br><br><br><tt>show_array: $t<hr>\n";
  for ($i=0;$i<@{$_[0]}; $i++) {
    ${$_[0]}[$i]=~s/\n/\\n/g;
    ${$_[0]}[$i]=~s/\t/<-tab->/g;
    ${$_[0]}[$i]=~s/</\&lt\;/g;
    ${$_[0]}[$i]=~s/>/\&gt\;/g;
    $out=${$_[0]}[$i];
    if ($t=~/p/)  { $out=~s/\r/\n/gs; $out=~s/\n/&para;\n/gs; }                   # zeige \n als Paragraph
    print "<!--nows-->($i)->$out<-<br>\n";
  }
  print "<hr></tt>\n";
  exit;
}

sub show_html {
############## html anzeigen (arrayname)
  my $i;
  &htmlcode($_[0]);
  print "Content-type: text/html\n\n";
  for ($i=0;$i<@{$_[0]}; $i++) {
    ${$_[0]}[$i]=&txt2htm(${$_[0]}[$i]);
    print "${$_[0]}[$i]";
  }
  exit;
}

sub show_file {
############## textfile anzeigen (filename)
  my $r=`cat $_[0]`;
  $r=~s/\n/<br>/gs;
  print "<tt>$r</tt>";
  exit;
}



############
sub read_dir { # &read_dir("dir","arrayname","Muster","Zeilennr")
############ directory lesen ergebnis in array "arrayname" (files+directories!!) oder Muster=/
 my @par = @_;
 my $tmp;
 if ($par[2] eq "") { $par[2]="\.\*"; } #beliebig

 my $file;
 my @liste;
 #$par[0]=&name2file($par[0]);
 local (*F);
 if ($par[0] eq "") { &CgiDie("(lib-561 read_dir: $par[2]) Programmfehler: Es wurde kein Verzeichnis angegeben (ref $_[3])"); }
 if ($par[1] eq "") { &CgiDie("(lib-562 read_dir: $par[2]) Programmfehler: Es wurde kein Array angegeben (ref $_[3])"); }
 if (!(-d $par[0])) { return; &CgiDie("(lib-1356 read_dir: $par[0]) Programmfehler: Kein Directory (ref $_[3])"); }

 opendir(F, $par[0]) || &CgiDie("(lib-564 read_dir: $par[2]) Programmfehler: Verzeichnis $par[0] gesperrt");
 while(defined($file=readdir F)) {
        next if (($file =~/^\.$/) || ($file =~/^\.\.$/));
        $tmp="$par[0]\/$file"; if (-d $tmp) { $file="$file\/"; }        # $file/ falls directory
        if ($par[2]=~/\./) {        #.xxx
                if ($file=~/$par[2]$/) {  push (@{$par[1]},$file); }
        } else {
                if ($file=~/$par[2]/) {  push (@{$par[1]},$file); }
        }
 }
 closedir(F);
 return;
}

#############    read array/hash file
sub read_ahfile {
# file nach hash data read_ahfile(file,hash/arrayname,art,aufruf), art =h/b/+ (hash,array,beides, anfgen), aufruf aus zeile
# sub array + hash aus Textdatei erzeugen
# hash namens $_[1] aus textdatei erzeugen (dort stehen in jeder Zeile die Wortpaare:  key|Wert|key2|wert2)
# ev. auch array namens $_[1] aus textdatei erzeugen
# var: nur wenn var in zeile enthalten (key oder value)
  my @tmp = @_;
  local ($h,$tmp,$k,$v,@werte,$i);
  local (*F);
  if ($tmp[2] eq "") { $tmp[2]="b"; }
  if (!($tmp[2] =~/\+/)) {
      @{$tmp[1]}=();
  }
  # bergabeprfung
  if (($tmp[0] eq "")|| ($tmp[1] eq "")) { &CgiDie("(lib read_ahfile $tmp[0],$tmp[1],$tmp[2],$tmp[3],$tmp[4]) error: no file or no array given"); }
  $tmp[0]=~s/\\/\//g;

  if (!(-e $tmp[0])) {&CgiDie("lib read_ahfile, file $tmp[0] not found");}


       #lock
       #my $SEMAPHORE=$tmp[0];  $SEMAPHORE.="lock";
       #open(LOCK, ">$SEMAPHORE")|| &error("(read_ahfile user.lock) Datei $SEMAPHORE <br>kann nicht gesperrt werden, versuchen Sie es spaeter nochmal!");
       #flock(LOCK, LOCK_EX);

          open (F, "<$tmp[0]") || &CgiDie("lib read_ahfile (file $tmp[0]) readerror");
          flock(F, LOCK_EX);
                while (<F>)  {
                   $_=~s/\015?\012|\015/\n/g;    # egal ob mac, win oder x #
                   # $_=~s/\013?\010|\013/\n/gs;    # egal ob mac, win oder x wrong
                   push (@{$tmp[1]},$_);
                 }
          close (F);
       # close (LOCK);
       # unlink($SEMAPHORE);


        if ($tmp[2] =~/h|b/)   {                        #auch (multi)hash
           if (!($tmp[2] =~/\+/)) {
              %{$tmp[1]}=();
           }
           foreach $tmp (@{$tmp[1]}) {
                if ($tmp=~/^\#/) {next;}
                if (!($tmp=~/\|/)) {next;}
                chomp($tmp);
                @werte=();
                @werte=split(/\|/,$tmp);
                $i=1;

                foreach $h (@werte) {
                       if ($i>1) {
                           $v=$h;
                           $k=~s/[^A-Za-z0-9\_\-\.]\\\///g;
                           if ($k=~/\w+/) {
                              ${$tmp[1]}{$k}=$v;                  # key=value
                           }
                           $i=1;
                       } else {
                          $k=$h;
                          $i++;
                       }
               }
           }
        }
        return;
}

###############
sub read_file {
###############  datei als parameter , return file as string
 my $f=$_[0];
 if (!(-f $f)) { return ""; }
 @t=();
 &read_ahfile($f,'t','a');        #read_ahfile(file,hash/arrayname,art,aufruf), art =h/b
 my $s=join('',@t);
 return ($s);
}

############
sub str2file {  # string to file string,file,append
############
  my $s=$_[0];    #string
  my $f=$_[1];    #file
  my $a=$_[2];    #append
  @t=();
  $t[0]=$s;
  &array2file("t",$f,$a);
}

##############
sub array2file {
# schreibt array in file (arrayname,file,modus,zeilennr,var1,var2) (modus =a heisst anfuegen)
# z.B. &array2file("db",$file2,"a","113",html,bn);
# falls var1 eq html > txt2html
# falls var2 eq bn dann <br> => <br>\n
# falls var2 eq nl dann <br> => <br> und \n => <n>
        my @tmp = @_;
        my $i;
        local (*F);

        if (!(defined($tmp[2]))) { $tmp[2]=""; }
        if (!(defined($tmp[3]))) { $tmp[3]=""; }
        if (!(defined($tmp[4]))) { $tmp[4]=""; }
        if (!(defined($tmp[5]))) { $tmp[5]=""; }

        # bergabeprfung
        if ($_[1] eq "") {
          &CgiDie("(sub array2file $tmp[3]) Programmfehler: Es wurde keine Datei angegeben");
        }
        if ($_[0] eq "") {
          &CgiDie("(sub array2file $tmp[3]) Programmfehler: Es wurde kein Array angegeben");
        }
         $tmp[1]=&name2file($tmp[1]);

          $tmp[6]=$tmp[1];                             #dateiname
         $tmp[6]=~s/.*\/cgi-bin/\.\.\./;

         if ($tmp[4] eq "html") {
                &htmlcode($_[0]);
        }

        #lock
        #my $SEMAPHORE=$tmp[1];  $SEMAPHORE.="lock";
        #open(LOCK, ">$SEMAPHORE")|| &error("(array2file.lock) Datei $SEMAPHORE <br>kann nicht gesperrt werden, versuchen Sie es spaeter nochmal!");
        #flock(LOCK, LOCK_EX);

        if ($tmp[2] ne "a") {
          open(F, ">$tmp[1]") || &error("(sub array2file<br>Datei $tmp[1] konnte nicht geschrieben werden.)<br>");
        } else {
          open(F, ">>$tmp[1]") || &error("(sub array2file<br>An Datei $tmp[1] konnte nichts angefuegt werden.)<br>");
        }
        flock(F, LOCK_EX);
        for ($i=0;$i<@{$tmp[0]};$i++) {
                if ($tmp[4] eq "html") { ${$tmp[0]}[$i]=&txt2htm(${$tmp[0]}[$i]); }
                if ($tmp[5] eq "bn") { ${$tmp[0]}[$i]=~s/<br>/<br>\n/g; }
                if ($tmp[5] eq "nl") { ${$tmp[0]}[$i]=~s/\n/<n>/g; }
                print F "${$tmp[0]}[$i]";
        }
        close(F);

        #close (LOCK);
        #unlink($SEMAPHORE);

  return 0;

}

#############
sub hash2file {
# schreibt hash in file (hashname,file,modus,zeilennr,lf) (modus =a heisst anfuegen)
# lf=br > \n >> <br> und \r > "" und "|" > ~I~
# falls lf eq nl dann <br> => <br> und \n => <n> und \r > "" und "|" > ~I~
# z.B. &hash2file("db",$file2,"a","113","br");
        my @tmp = @_;
        my $tmp;
        my $k;
        my $v;
         my @out;
        local (*F);

        # bergabeprfung
        if ($_[1] eq "") {
          &CgiDie("(lib hash2file $tmp[3]) Programmfehler: Es wurde keine Datei angegeben");
        }
        if ($_[0] eq "") {
          &CgiDie("(lib hash2file $tmp[3]) Programmfehler: Es wurde kein Hash angegeben");
        }
         $tmp[1]=&name2file($tmp[1]);
            $tmp[5]=$tmp[1];                             #dateiname
         $tmp[5]=~s/.*\/cgi-bin/\.\.\./;

        if ($tmp[2] ne "a") {
          open(F, ">$tmp[1]") || &CgiDie("(lib hash2file $tmp[3]) Datei $tmp[5] konnte nicht geschrieben werden.)");
        } else {
          open(F, ">>$tmp[1]") || &CgiDie("(lib hash2file $tmp[3]) An Datei $tmp[5] konnte nichts angefuegt werden.)");
        }
        flock(F, LOCK_EX);
        while (($k,$v) = each %{$_[0]}) {
            if ($tmp[4] eq "br") { $v=~s/\n/<br>/gs; $v=~s/\r//gs; $v=~s/\|/~I~/gs;}
            if ($tmp[4] eq "nl") { $v=~s/\n/<n>/gs; $v=~s/\r//gs; $v=~s/\|/~I~/gs;}
            push (@out,"$k\|$v\|\n")
            #print F "$k\|$v\|\n";
        }
        @out=sort(@out);
        foreach $tmp (@out) { print F "$tmp";}
        close(F);
  return 0;
}

############
sub get_dir { # hide: &get_dir("dir") via sudo
############ directory lesen ergebnis in array "arrayname" (files+directories!!) oder Muster=/
 my ($file,@liste,$d,@d,$r);
 $d=$_[0];
 my $par=$_[1];
 if ($par eq "") { $par="mp"; }                    # ls as commalist with / at dirs

 $d=~s/\'/\'\\\'\'/g;                              # allow ' in name
 $r=`sudo /usr/bin/ls -$par \'$d\' 2>&1`;          # ls list
 chomp($r);
 if ($r=~/^ls:/) { return(); }
 if ($par=~/m/) { $r=~s/, */\n/gs; }

 @d=split(/\n/,$r);
 foreach my $t (@d) {
    if ($t eq "") { next; }
    if ($t=~/no such file/i) { next; }            # supress error messages
    push(@liste,$t);
 }

 #debug
 # print "<hr>$d<hr>"; &print_array(@liste);
 return(@liste);
}

################
sub get_language {  #localisation
##################
 #hash %txt (localisation)

 #debug
 #print "Content-type: text/html\n\n";
 # &print_hash(%sys);

       #get menu number like 05.01, used ex in Logwindow when =L
        my ($p,$t,$f,$i,$folder);
        $p="$in{'l1'}";
        $p=~ tr/A-Z/a-z/; $p=~s/=.*//;
        if ($p=~/^\d+\_/) {
            $f=$p; $f=~s/^\d+\_//; $f=~s/ .*//;      # keep first word
            $p=~s/\_.*//;                            # keep nr
        }

        if ($in{'l2'} ne "") {
          $t=$in{'l2'};
#&mess($t);
          $t=~ tr/A-Z/a-z/;  $t=~s/=.*//;
          if ($t=~/^\d+\_/) {
               $i=$t; $i=~s/^\d+\_//; $i=~s/ .*//;      # keep first word
               $t=~s/\_.*//;
          }
          $p.=".$t";
          $f.="_$i";
        }
        if ($in{'l3'} ne "") {
          $t=$in{'l3'};  $t=~s/=.*//;                   # options
          $t=~ tr/A-Z/a-z/;
          if ($t=~/^\d+\_/) {
               $i=$t; $i=~s/^\d+\_//; $i=~s/ .*//;      # keep first word
               $t=~s/\_.*//;
          }
          $p.=".$t";
          $f.="_$i";
        }
        $sys{'menu'}="m$p";

        $f=~s/napp\-it/about/;                           # base menu is now about
        $f=lc($f);
        $sys{'menufile'}="$f";


        #get menu-files
        if ($cfg{'select_lang'} eq "") { $cfg{'select_lang'}="win"; }

        # 1. always get default/basic
        $t="C:\\xampp\\napp-it\\data\\napp-it\\zfsos\\_lib\\lang\\en\\about_basic.txt";
        if (-f $t) { $t=&exe("type $t"); &get_language2($t);  }
        #&mess($t);

        # 2. always read en menu
        $t="C:\\xampp\\napp-it\\data\\napp-it\\zfsos\\_lib\\lang\\enabout_menus.txt";
        if (-f $t) { $t=&exe("type $t"); &get_language2($t);  }

        # 3. menus, force MY!
        $t="C:\\xampp\\napp-it\\_my\\zfsos\\_lib\\lang\\MY!\\about_menus.txt";
        if (-d "C:\xampp\\napp-it\\_my\\zfsos\\_lib\\lang/MY!\\") {
           $t=&exe("type $t"); &get_language2($t);
           $cfg{'select_lang'}="MY!";
        }

        #4. menu lang
        $folder="C:\\xampp\\napp-it\\data\\napp-it\\zfsos\\_lib\\lang\\$cfg{'select_lang'}\\";             # default ex de
        if ($current{'force_en'} ne "1") {
          if (! (($cfg{'select_lang'} eq "MY!") || ($cfg{'select_lang'} eq "en")) ) {
            if ($cfg{'select_lang'}=~/[A-Z]+/) {
               $folder="C:\xampp\\napp-it\\_my\\zfsos\\_lib\\lang\\$cfg{'select_lang'}\\";               # my languages ex IT
            }
            $t="$folder"."about_menus.txt";   if (-f $t) { $t=&exe("type $t"); &get_language2($t);   }
# &mess($t);
          }
        }

        # txt single menu, default en, load always
        $t="C:\\xampp\\napp-it\\data\\napp-it\\zfsos\\_lib\\lang\\en\\$sys{'menufile'}.txt";
#&mess($t);
        if (-f $t) {
           $t=&exe("type $t"); &get_language2($t);
        }

        # txt lang or my (basic + menu)
        if ($cfg{'select_lang'} eq "MY!") {
              $t="C:\\xampp\\napp-it\\_my\\zfsos\\_lib\\lang\\MY!\\about-basic.txt";
              if (-f $t) {  $t=&exe("type $t"); &get_language2($t); }

              $t="C:\\xampp\\napp-it\\_my\\zfsos\\_lib\\lang/MY!\\$sys{'menufile'}.txt";
              if (-f $t) {  $t=&exe("type $t"); &get_language2($t); }
        } else {   # load lang
             if ($current{'force_en'} ne "1") {
                 $t=$folder."about_basic.txt";
                 if (-f $t) { $t=&exe("type $t"); &get_language2($t);  }

                 $t=$folder."$sys{'menufile'}.txt";
                 if (-f $t) { $t=&exe("type $t"); &get_language2($t);  }
             }
        }

}

#################
sub get_language2 {
#################
   my $l=$_[0];
   my (@l,@p);
        #menue l1
        #syntax <<t01>>This is text 01..
        @l=split(/<</,$l);
        foreach $t (@l) {
                 @p=();
                 @p=split(/>>/,$t);
                 $p[1]=~s/^ //;     # first space
                 $p[1]=~s/\s+$//s;  # trailing spaces an nl
                 if ($p[0]=~/\_inFo$/) { next; }  # only for editing  inFo to allow info

                 if ($p[0] ne "") { $txt{$p[0]}=$p[1]; }
        }
}

#################     synclib,grouplib
sub port_from_id {
     # port from id > 50000
    my $id=$_[0];
    my $port=substr($id,-4,4);
    my $sum="";
    my @sum=split(//,$id);
    foreach my $char (@sum) { $sum=$sum+ord($char); }
    $port=$sum+50000+$port;
    return($port);
}

##################   inter process communication
sub xxset_ps {

# parameter wait, id,
# with refresh (delete other jobs with same id)
my $wait=$_[0]; $wait=~s/\D//g;    if (!($wait >=0))  { return(-1); }
my $id=$_[1];   $id=~s/\s+//g;     if ($id eq "")     { return(-1);  }
my $time=();
#print "$wait $id $refresh"; return;

return;
#set new ps
#system("sudo nohup /usr/bin/sh /var/web-gui/tools/set-ps.sh $wait $id $time &");

# update
  my $p=`/usr/bin/ps axw | grep "set-process.sh" | grep "$id"`;
  if ($p eq "") { return (0); }
  my $i="0";
  my @p=split(/\n/,$p);
  foreach my $p (@p) {
     $p=~s/^\s+//; $p=~s/\s.*//;
     #`sudo kill -9 $p  >/dev/null 2>&1`;
     $i++;
  }
  return($i);
}

                                           ## Datumsfunktionen##
###############

###############
sub setkey {                              #
############# this is not a hidden procedure, do not use without a valid license ###
  my $in=$_[0];
  $in=~s/\s+$//gs;
  if ($in eq "")                  { &mess('Please enter a key'); }

  if ($in=~/\|/) {
     $in=~s/\|+$//;
     $in=~s/\|+ .*$//;
     $in=~s/^.*\|//;
  }

  $t=&getv2($in);
  if ($t=~/ev/i)                 { &mess('Please enter a valid key'); }

  $in=~s/\r/\n/gs;
  #my @l=split(/\n/,$in);
  #&print_array(@l);
  #&mess($t);

 &socks("echo '' > C:\xampp\\napp-it\\_log\\appliance.log");
  my $old=&socks("type /var/web-gui/_log/appliance.log");
  if (!($old=~/\n$/s)) {
       open (PF, ">>C:\xampp\\napp-it\\_log\\appliance.log");
            print PF "\n";                                     # \n at the end
       close (PF);
  }

  my $k=$in;

  $k=~s/^\s+//; $k=~s/\s+$//;
  # enter cfg value ?
  if ($k=~/\|/) {
         $k=~s/\|+$//;
         $k=~s/.*\|+//;
  }

  if ($k=~/\w+::\w+/)  {
       $k=~s/ +$//g;
       if ($k eq "")                  { &mess('Please enter a key'); }
       if (!($k=~/\w+::\w+/))         { &mess('Invalid key $k \nplease request a key'); }
       my $i=&getv($k);
       if ($i =~/^ev/i)               { &mess("Invalid key $k \nplease request a new key."); }
       my $h=$k; $h=~s/ .*//; $h=~s/::.*//;

       $old=&socks("type /var/web-gui/_log/appliance.log");
       if ($old=~/app_$h\|$in\|/s) {
         # already registered

       } else {

         open (PF, ">>C:\xampp\\napp-it\\_log\\appliance.log");
            print PF "app_$h\|$in\|\n";
         close (PF);
         $in{'l2'}="";
       }
  }

  # app_name, app_logo ..
  if ($k=~/^app_\w+\|/)  {
       open (PF, ">>C:\xampp\\napp-it\\_log\\appliance.log");
            print PF "$k\n";
       close (PF);
  }


   # remove evalkey for pro key
   my $e;
   if ($old=~/ eval|\_eval/s) {
         if ($_[0]=~/^complete|^acl|^replicate|^monitor|^pro/) {
             $old=&socks("type C:\xampp\\napp-it\\_log\\appliance.log");
             @t=(); @t=split(/\n/,$old);
             open (PF, ">/var/web-gui/_log/appliance.log");
               foreach my $t (@t) {
                      if ($t=~/ eval|\_eval/) { $e.="$t\n"; next; }
                      print PF "$t\n";
               }
             close (PF);
         }
   }
   $in{'l2'}="";
   if ($e ne "") {
      open (PF, ">>C:\xampp\\napp-it\\_log\\appliance.oldkey");
            print PF "$e";
      close (PF);
   }

   print "Thank you very much for your contribution.";
   sleep(2);

   &reload;
}

###
#################################
sub get_pageid {  # page id like 15.01.htm
  my $l1=$in{'l1'}; $l1=~s/\_.*//;
  my $l2=$in{'l2'}; $l2=~s/\_.*//;
  my $l3=$in{'l3'}; $l3=~s/\_.*//;
  $l1=~s/\D//g; $l2=~s/\D//g; $l3=~s/\D//g;


  my $id;
  if ($l1 > 0) {
      $id="$l1";
      if ($l2 > 0) {
         $id=".$l2";
         if ($l3 > 0) {
            $id=".$l3";
         }
      }
  } else {
      $id="00";
  }

  $id.=".htm";
  return($id);
}

sub getpd {  # getpd
  my $f="C:\xampp\\napp-it\\_log\\appliance.log";
  my ($k,$r,@v,@w,$out,$t,$valid);
  if (-f $f) {
      $r=&socks("cat $f");
      my @w=split(/\n/,$r);
      foreach my $w (@w) {
              if (!($w=~/\d+\.\d+\.\d+::/)) { next; }

                  chomp($w);

                  @v=();
                  @v=split(/\|/,$w);

                  $t=&getv($v[1]);
                  if ($t=~/ev/i) {
                     $valid="<font style='color:red'>not valid</font>";
                  } else {
                    $valid="";

                  }
                  $out.="$v[0]\t$v[1]\t";

                  $k=$v[1];
                  $k=~s/::.*//;
                  $k=~s/.*\- *//;
                  $k=~s/[^0-9\.]//g;                          # date

                  $t=&date2diff($sys{'date'},$k);

                  if ($valid ne "") { $t=$valid; }  else { $t.=" days"; }     # valid?
                  $valid="";

                  $out.="$k";

                  #check against installed ext
                  #if (($w=~/^app_complete/)  && (!($cfg{'ver'}=~/c/)) )   { $k='0'; $t='-1';  }
                  #if (($w=~/^app_pro/)       && (!($cfg{'ver'}=~/c|p/)) ) { $k='0'; $t='-1';  }
                  #if (($w=~/^app_replicate/) && (!($cfg{'ver'}=~/c|r/)) ) { $k='0'; $t='-1';  }
                  #if (($w=~/^app_acl/)       && (!($cfg{'ver'}=~/c|a/)) ) { $k='0'; $t='-1';  }
                  #if (($w=~/^app_monitor/)   && (!($cfg{'ver'}=~/c|m/)) ) { $k='0'; $t='-1';  }

                  if ($k=~/2099/) {
                    $out.="\tunlimited\n";
                  } else {
                          if ($t < 0 ) {
                             $out.="\t<font style='color:red'><b>not valid</b></font>\n";

                          } else {

                               if ($t < 60) {
                                    $out.="\t<font style='color:#eb8212'><b>$t</b></font>\n";
                               } else {
                                    $out.="\t$t\n";
                               }
                          }
                  }
      }
  }

  # hide license, no update info
  if ($txt{'app_about_show_license'}=~/no/) {
     $t="";
  }  else {
      if ($cfg{'select_lang'}=~/my|app/i) {
        $t="";
      } else {
        if ($out eq "") {
          $t=" <a href='http://napp-it.org/extensions/quotation_en.html' target='_blank'><i><u>Update to napp-it PRO</u></i></a>";
        } else {
          $t=" <a href='http://napp-it.org/extensions/quotation_en.html' target='_blank'>Order/renew online</a>";
        }
      }
  }

  if ($out eq "") {
       return("<br><b style='background-color:#dddddd'>Installed Extensions : none</b> $t");
  }

  $out=~s/::\w+\t/....\t/gs;

  $out="<b>Installed Extensions</b>\t<b>Key</b>\t<b>Until</b>\t<b>Valid</b>&nbsp;&nbsp; $t<br>\n$out";
  $out=&list2table($out,'950px');
  $out=~s/\n+$//s;
  $out.="<br>";

  return($out);
}

##############
sub exe_fork {
############## execute a perlscript under /var/web-gui/data/napp-it/zfsos/_lib/scripts/ via agent-manager

&mess("exe_fork in admin.pl: needed?");
#return;

  my $script=$_[0];                      # scriptname without .pl
  $script=~s/[^a-z0-9\_\-]//gs;          # no special char
  if ($script eq "")  { return; }
  $script.=".pl";

  my $par=$_[1];



  my $par_pid = $$;
  my $child_pid = fork();

  if($child_pid) {
      return;             # parent
  }  else {               # child
      `sudo sh /var/web-gui/data/napp-it/zfsos/_lib/scripts/agent_daemon.sh $script $par`;
      exit;
  }
}

############ par: option value, current value
sub get_sel_option {
##############
  my $select=$_[0];
  my $current=$_[1];
  if ($current eq "") { return("<option>$select</option>"); }
  if ($select=~/$current/i) {
    return("<option selected>$select</option>");
  }  else {
    return("<option>$select</option>");
  }
}


sub get_short_a {
###############  # par char, length f/l (first,last, default=first)
        my $c=$_[0];  # char
        my $l=$_[1];  # lenf/l ex 20l (20last)
        my $w;        # what: f/l


        if ($l=~/l/) { $w="l"; } else { $w="f";} #default first
        $l=~s/\D//g;
        if ($l eq "") { $l=20; }  #default

        if ($w eq "l") {
                if (length($c) > $l) {
                    $c=substr($c,length($c)-$l,$l);  # last n
                    $c="<a href='#' style='text-decoration: none; font-weight: normal; color: black' title=\'$_[0]\'>..$c</a>";
                }
        } else {
                if (length($c) > $l) {
                    $c=substr($c,0,$l);  # first n
                    $c="<a href='#' style='text-decoration: none; font-weight: normal; color: black' title=\'$_[0]\'>$c..</a>";
                }

        }
        return($c);
}

sub get_os {
##########
       if (-d "c:/xampp") {
                $sys{'auto_os'}="Windows";
                $sys{'os'}="Windows";
                $sys{'auto_os_type'}="win";
                return;
       }


        my $t=`uname -a`;
        my $v;
        my @t=();
        @t=split(/ /,$t);
        if ($t=~/[0-9]/)    {
                $sys{'auto_os'}="Linux";       # defaults
                $sys{'auto_os_type'}="lin";
                $sys{'binpath'}="/bin/";
                $sys{'sbinpath'}="/sbin/";

                if ($t[3]=~/darwin/is)   { $sys{'auto_os'}="Darwin";  $sys{'auto_os_type'}="dar"; }
                if ($t[0]=~/sunos/is)    {
                     $sys{'auto_os'}="*solaris*"; $sys{'auto_os_type'}="ill_o";
                     $sys{'binpath'}="/usr/bin/";                                                                     # path to chmmod, ls -> use them instead gnu ones
                     $sys{'sbinpath'}="/usr/sbin/";                                                                   # use when needed

                     if ($t[3]=~/svn_151/)  { $sys{'auto_os'}="Solaris Express"; $sys{'auto_os_type'}="ill_x"; }      # opensol solaris dev
                     if ($t[3]=~/^11/)      {                                                                         # Oracle Solaris
                            $sys{'auto_os'}="Solaris 11";      $sys{'auto_os_type'}="sol_s";
                            $sys{'auto_os_ver'}=$t[3];
                     }

                     if ($t[3]=~/oi/is)       { $sys{'auto_os'}="OpenIndiana";   $sys{'auto_os_type'}="ill_o"; }      # OI/illumos based
                     if ($t[3]=~/illumos/is)  { $sys{'auto_os'}="OpenIndiana";   $sys{'auto_os_type'}="ill_h"; }      # OI/Hipster based


                     if ($t[3]=~/omni/is)     {
                          $sys{'auto_os'}="OmniOS";        $sys{'auto_os_type'}="ill_o";

                          # get auto_os_ver
                          $v=`cat /etc/release`;
                          if ($v=~/r151/s) {
                             $v=~s/.*r151/151/s;
                             $v=~s/\s.*//s;
                             $sys{'auto_os_ver'}=$v;
                          }

                     }                                                                                                # OI/illumos based
                     if ($t[3]=~/nexenta/is)  { $sys{'auto_os'}="Nexenta";       $sys{'auto_os_type'}="ill_n"; }      # nexenta/Illumos based

                     $t=$t[3]; $t=~s/\D//g;                                                                           # get snv release
                     $sys{'auto_build'}=$t;

                     $sys{'auto_minpool'}="21";                                                                       # min supported pool-version default
                     if ($t > 129 )     { $sys{'auto_minpool'}="21"  }                                                # dedup
                     if ($t > 150 )     { $sys{'auto_minpool'}="30"  }                                                # se11 x crypt

                                }

                #if ($t[3]=~/4c|illumian/i)   { $sys{'auto_os'}="Illumian";       $sys{'auto_os_type'}="ill_i"; }       # Illumian obsolet

                if ($sys{'auto_os_type'} =~/ill_/)     { $sys{'auto_minpool'}="28"; $sys{'auto_build'}="151";  }
                if ($sys{'auto_os'} eq "Solaris 11" )  { $sys{'auto_minpool'}="33"; $sys{'auto_build'}="175";  }       # Solaris 11

        }
}


sub getp {    #pro , call by auto

  $sys{'edition'}="Free";
  my $f="C:\xampp\\napp-it\\_log\\appliance.log"; $cfg{'ver'}="";
  my ($k,$r,@v,@w,$t);
  if (-f $f) {
      $r=&exe("cat $f",'','0');
      my @w=split(/\n/,$r);
      foreach my $w (@w) {
              if (!($w=~/\w/)) { next; }
              chomp($w);
              if ($w=~/::/) {
                   @v=();
                   @v=split(/\|/,$w);
                   $t=&getv($v[1],"i");

                   if ($t=~/^ev/i) {
                        $k.="0";
                   } else {

                    #Command window
                       if ($w=~/\|complete/) {
                            if (!($cfg{'ver'}=~/c/)) { $cfg{'ver'}.="c"; }
                            if (!($cfg{'ver'}=~/a/)) { $cfg{'ver'}.="a"; }
                            if (!($cfg{'ver'}=~/r/)) { $cfg{'ver'}.="r"; }
                            if (!($cfg{'ver'}=~/m/)) { $cfg{'ver'}.="m"; }
                        }
                        if ($w=~/\_eval/) {
                            if (!($cfg{'ver'}=~/e/)) { $cfg{'ver'}.="e"; }
                        }
                        if ($w=~/\_home|\_noncom/) {
                            if (!($cfg{'ver'}=~/h/)) { $cfg{'ver'}.="h"; }
                        }
                        if (($w=~/\|acl/) && (!($cfg{'ver'}=~/a/))) {$cfg{'ver'}.="a";  }
                        if (($w=~/\|mon/) && (!($cfg{'ver'}=~/m/))) {$cfg{'ver'}.="m";  }
                        if (($w=~/\|rep/) && (!($cfg{'ver'}=~/r/))) {$cfg{'ver'}.="r";  }
                        if (!($cfg{'ver'}=~/p/)) {$cfg{'ver'}.="p";  }
                        $k.="1";
                   }
               }
      }
  }

  #print "Content-type: text/html\n\n"; &print_hash(%cfg);
  $cfg{'pver'}="";
  if (!($sys{'version'}=~/f/i)) { $cfg{'pver'}=$cfg{'ver'}; }   # pro and dev

  if ($cfg{'ver'}=~/p/) { $sys{'edition'}="PRO"; }
  #$sys{'edition'}="free";
  return();
}


sub getv {
#########
my $c=$_[0];
my $i=$_[1];
my ($r,$t,$h,$h2,$o,@w,@l);

#if ($c=~/h:/) {
#print &PrintHeader;
#&mess($c);
#}

if (($c=~/h:m/) && ($c=~/[lem]90[ \_]/)) {
   if ($c=~/m90[ \_]/) {
       $t=$c;
       $h2=$t;
       $h2=~s/.*h://;
       $h2=~s/ .*//;
       $h2=~s/\_.*//;      #h:xy_abc

       $t=~s/.*h:m//;
       $t=~s/m90 .*|m90\_.*//;
       $r=`sudo /usr/sbin/dladm show-phys -m`;
       $o="1";
       @l=split(/\n/,$r);
       foreach my $l (@l) {
          @w=();
          @w=split(/\s+/,$l);   # link slot adress
          $w[2]=~s/://g; $w[2]=lc($w[2]);

          #print ("$t -> $w[2]<br>");
          if ($t eq $w[2]) { $o="0"; last; }
       }
       if ($sys{'version'}=~/de/) { $o="0"; }       # dev release
       if ($o) {
            if ($i ne "") { return ("Evaluation only or wrong host"); }
            $h=&get_hid;
            &mess("This key is only valid on host with machine id $h2<br>If you have changed your system, request a new key.<br>Your old key may be moved to Extension > Register > oldkeys");
       }

   } else {

       $h=&get_hid("");
       $h=~s/h://;

       $h2=$c;
       $h2=~s/.*h://;
       $h2=~s/ .*//;
       $h2=~s/\_.*//;      #h:xy_abc
       if ($sys{'version'}=~/de/) { $h=$h2; }       # dev release
       if ($h ne $h2) {
            if ("h:$h2" ne $sys{'sysdef'}) { # oldkey
              if ($i ne "") { return ("Evaluation only or wrong host"); }
              &mess("This Key is only valid on host with machine id h:$h2<br>If you have changed your system, request a new key.<br>Your old key may be moved to Extension > Register > oldkeys");
            }
       }
   }
 }  else {
       if ($c=~/h:/) {
             $h=`hostname`;
             chomp($h);
             $h2=$c;
             $h2=~s/.*h://;
             $h2=~s/ .*//;
             $h2=~s/\_.*//;      #h:xy_abc
             if ($sys{'version'}=~/de/) { $h=$h2; }       # dev release
             if ($h ne $h2) {
                  if ($i ne "") { return ("Evaluation only or wrong host"); }
                  &mess("This key is only valid on server with hostname $h2<br>If you have changed your hostname, request a new key.<br>Your old key may be moved to Extension > Register > oldkeys");
             }
        }
 }


  $t=&getv2($c);

  if ($t=~/^Evaluation/i) {
    if (($c=~/h:m/) && ($c=~/[le]90\_|[le]90 /)) {
       $h=&get_hid;
       $h=~s/h://;

       $h2=$cfg{'app_complete'};
       $h2=~s/.*h://;
       $h2=~s/ .*//;
       $h2=~s/\_.*//;      #h:xy_abc
       if ($sys{'version'}=~/de/) { $h=$h2; }       # dev release
       if ($h ne $h2) {
            if ($i ne "") { return ("Evaluation only or wrong machine id"); }
            &mess("Evaluation only on host with machine id h:$h2<br>If you have changed your system, request a new key.<br>Your old key may be moved to Extension > Register > oldkeys");
       }
    }  else {

       if ($cfg{'app_complete'} =~/h:/) {
         $h=`hostname`;
         chomp($h);
         $h2=$cfg{'app_complete'};
         $h2=~s/.*h://;
         $h2=~s/ .*//;
         $h2=~s/\_.*//;      #h:xy_abc
         if ($sys{'version'}=~/de/) { $h=$h2; }       # dev release
         if ($h ne $h2) {
              if ($i ne "") { return ("Evaluation only or wrong host"); }
              #&mess("This key is only valid on host $h2");
              return ("Evaluation/ wrong host");
         }
       }
    }
     $t=&getv2($cfg{'app_complete'});
  }
  return ($t);
}


sub getc {       # hide:
#########
  my @p=split(/::/,$_[0]); $p[1]=~tr/mVsDhKlPOtqwT/0123456789\.\-\:/; return("$p[0]\::$p[1]");
}

sub getv2 {   # hide:
  # ex cfg of app_replicate

  my $c=&getc($_[0]);
  my @p=split(/\::/,$c);

 # print $_[0];
 # &print_array(@p);

  my $dif=&date2diff("$sys{'date'}",$p[2]);
  if (($dif < 0) || (&date2valid($p[2])==0)) { return("Evaluation only or valid key required: <br>");  }

  my $r=$p[0]."::".$p[1]."::".$p[2];
  my @c=split(//,$r);
  my $y;
  foreach my $w (@c)  { $y=$y+ord($w);}
  if ($y ne $p[3]) { return("Evaluation only: <br>$txt{'evaluate'}");  }
  my $t=$txt{'evaluate2'}; $t=~s/=from=/$p[1]/;
  if ($p[2]=~/2099/) { $t=~s/=until=/unlimited/;  }
  $t=~s/=until=/$p[2]/;
  if ($t eq "") { $t="$p[1] - $p[2]";}

  return("<br>$t");
}


########## KB-KiB-byte conerter ex &b_convert($data,"i","b");
sub b_convert {
##########
   #
   my $data=$_[0];     # data

   my $in=$_[1];       # in data [bid]:  (b)ytes = default, b(i)nary base like KiB, (d)eziman base like KB
   my $out=$_[2];      # out data [bsi(n)]: (b)ytes = default, (s)short like KB, b(i)nary base like KiB, n=number of dezimals

   my $dez=$out;  $dez=~s/[^0-9]//g;
   if ($dez eq "")     { $dez="2"; }  # default ex 10.15 MB
   if ($out eq "")     { $out="b"; }  # default exact bytes

   my $base="1000";
   if ($in=~/i/i)      {  $base="1024"; }  # or
   if ($data=~/iB/i)   {  $base="1024"; }

   $data=~s/,/./g;         # german 1,1 = 1.1
   my $bytes=$data; $bytes=~s/[^0-9\.\,]//g;
#&mess($bytes);

   if (($bytes eq "") || ($bytes eq "0"))     {  return("0"); }

   # 1. calculate bytes
   if ($data=~/k/i)     {  $bytes=$bytes * $base; }                                  # KB
   if ($data=~/m/i)     {  $bytes=$bytes * $base * $base; }                          # MB
   if ($data=~/g/i)     {  $bytes=$bytes * $base * $base * $base; }                  # GB
   if ($data=~/t/i)     {  $bytes=$bytes * $base * $base * $base * $base; }          # TB
   if ($data=~/p/i)     {  $bytes=$bytes * $base * $base * $base * $base * $base; }  # PB
   if ($data=~/e/i)     {  $bytes=$bytes * $base * $base * $base * $base * $base * $base; }  # EB

#&mess($bytes);
   if ($out=~/b/i)     {  return($bytes); }           # return exact number of Bytes

   # calculate output
   $base="1000";
   if ($out=~/i/i)          {  $base="1024";   }

   my $unit="";
   if ($bytes < 1000 )      {  return($bytes); }      # return exact number of Bytes

   $unit="K";
   if ($bytes >= 1000 )   {  $bytes=$bytes/$base; $unit="K"; } # < 1 MB (KB)
   if ($bytes >= 1000 )   {  $bytes=$bytes/$base; $unit="M"; } # < 1 GB (MB)
   if ($bytes >= 1000 )   {  $bytes=$bytes/$base; $unit="G"; } # < 1 TB (GB)
   if ($bytes >= 1000 )   {  $bytes=$bytes/$base; $unit="T"; } # < 1 PB (TB)
   if ($bytes >= 1000 )   {  $bytes=$bytes/$base; $unit="P"; } # < 1 EB (PB)
   if ($bytes >= 1000 )   {  $bytes=$bytes/$base; $unit="E"; } # < 1 ZB (EB)

   # calc dezimal
   $dez.="f";
   $bytes = sprintf ("%.$dez", $bytes);    # round  ex $bytes = sprintf ("%.1f", $bytes);
   $bytes=~s/\.0+$//;

   $bytes.=" $unit";
   if ($out=~/i/i) {  $bytes.="iB"; } else {  $bytes.="B";  }
   return($bytes);

}

############ host id
sub get_hid {
############
  #print "Content-type: text/html\n\n";
  my ($r,$tm,$f,@w,@l,$a,$e,$l,$x);

  $tm=time();
  $tm=sprintf("0x%X", $tm);
  $tm=~s/0x//;
  $tm=lc($tm);

  # solarish
  if (-f "/sbin/dladm") {

     &sysdef; # sysdef -> $current{'sysdef'}

     $r=`sudo /sbin/dladm show-phys -o DEVICE,LINK | grep -v DEVICE`;    # DEVICE  LINK

     @l=split(/\n/,$r);
     @l=sort(@l);                                               # ex e1000g0 first
     #@l=reverse(@l);
     @w=split(/ +/,$l[0]);

#&mess($w[0]);                                                                #
     if ($r=~/vmxnet/s)  {
          #                                                     # aio
          $r=`sudo /sbin/dladm show-phys -m -o ADDRESS $w[1] | grep :`;  # mac from DEVICE->LINK
          chomp($r);
          if ($r=~/:/s) {
              $r=~s/://g;
              $r=lc($r);
              if ($r=~/\w/) {
                   $r.="e90";
                   return("h:m$r");             # aio
              }
          }

     } else {

          $r=`sudo /sbin/dladm show-phys -m -o ADDRESS $w[1] | grep :`;  # mac from DEVICE->LINK
          chomp($r);
          if ($r=~/:/s) {
              $r=~s/://g;
              $r=lc($r);
              if ($r=~/\w/) {
                   $r.="m90";
                   return("h:m$r");             # barebone
              }
          }
     }
  }

  # linux

  $r=`sudo /sbin/ifconfig -a`;
  @t=split(/\n/,$r);


#&mess($r);

  $r="";
  foreach my $t (@t) {
      if ($t=~/^lo/) { next; }
      @w=split(/ +/,$t);
      foreach my $w (@w) {
        if ($w=~/:\w\w:\w\w:/)  { $r=$w; last; }   # mac  v4
      }
  }

  if ($r eq "") { return ("h:unknown"); }

  $r=~s/://g;
  $r=lc($r);
  $r.="l90";

#&mess("h:m$r");
# &print_array(@t); exit;

  return("h:m$r");
}



#############
sub sysdef {
  &mess("needed?");
  my ($k,$f,$tm,$r);

  $f="/var/web-gui/_log/report/time/tm.log";
  if (-f "/usr/sbin/sysdef") {

     $r=`/usr/sbin/sysdef -h`;                # ex 4f907e90
     chomp($r);
     $r=~s/.*\n +//s;

     if (-f "/etc/tm.cfg") {
         `sudo cp /etc/tm.cfg $f`;
     }  else {
         if (-f "$f") {
           `sudo cp $f /etc/tm.cfg`;
           `sudo chmod 644 /etc/tm.cfg`;
         } else {
          `sudo echo $tm > $f`;
          `sudo chmod 644 $f`;

          `sudo cp $f /etc/tm.cfg`;
          `sudo chmod 644 /etc/tm.cfg`;
         }
     }
     ###
     if (-f "$f") {
          $tm=`cat $f`;
          $tm=~s/\s//gs;
     }

     $tm=substr($tm,4,4);
     $r="$tm$r"."e90";
     $r=~s/e90e90/e90/;
     $r="h:m$r";
     $sys{'sysdef'}="$r";        # key from sysdef
  } else {
    $sys{'sysdef'}="not valid";
  }
}


################ ex converts TB to Tib
sub b2bib {
################
  my $data=$_[0];
  $data=~s/ +//g;

  my $dez=&b_convert("$data","d","1s");
  my $bin=&b_convert("$data","d","i1s");
  return("$dez","$bin");        # ex "3 TB"  "2,7 Tib"

  ####
  my $bytes=$_[0];                # needs nn (=Bytes), nn KB, nn MB, nn GB, nn TB, nn PB
  my ($ib,$iu,$b,$u,$k);

  $bytes=~s/ +//g;

  if ($bytes=~/K/i) { $bytes=~s/KB*//i; $bytes=$bytes * 1000; }             # KB or K
  if ($bytes=~/M/i) { $bytes=~s/MB*//i; $bytes=$bytes * 1000000; }
  if ($bytes=~/G/i) { $bytes=~s/GB*//i; $bytes=$bytes * 1000000000; }
  if ($bytes=~/T/i) { $bytes=~s/TB*//i; $bytes=$bytes * 1000000000000; }
  if ($bytes=~/P/i) { $bytes=~s/PB*//i; $bytes=$bytes * 1000000000000000; }

  # KB
  $ib=$bytes/1024;
  $bytes=$bytes/1000;

  $u=" KB";
  $iu=" KiB";

  if ($bytes == 0) {
       $u="";
       $iu="";
  }

  if ($bytes >= 1000) {
      # MB
      $bytes=$bytes/1000;
      $u=" MB";
  }
  if ($ib >= 1024) {
      $ib=$ib/1024;
      $iu=" MiB";
  }

  if ($bytes >= 1000) {
      # GB
      $u=" GB";
      $bytes=$bytes/1000;
  }
  if ($ib >= 1024) {
      $ib=$ib/1024;
      $iu=" GiB";
  }

  if ($bytes >= 1000) {
      # TB
      $bytes=$bytes/1000;
      $u=" TB";
  }
  if ($ib >= 1024) {
      $ib=$ib/1024;
      $iu=" TiB";
  }
  if ($bytes >= 1000) {
      # PB
      $bytes=$bytes/1000;
      $u=" PB";
  }
  if ($ib >= 1024) {
      $ib=$ib/1024;
      $iu=" PiB";
  }

  $bytes = sprintf ("%.1f", $bytes);    # round
  $ib = sprintf ("%.1f", $ib);          # round

  $bytes=~s/\.0$//;
  $ib=~s/\.0$//;

  if ($bytes >0) { $bytes="$bytes$u"; } else { $bytes="0"; }
  if ($ib >0)    { $ib="$ib$iu";      } else { $ib="0"; }

  return("$bytes","$ib");        # ex "3 TB"  "2,7 Tib"
}

sub getdate {                # setzt datumsvariablen
###########
 my ($min,$hour,$mday,$sec,$mon,$year,$wday,$ydat,$isdat);
 ($sec,$min,$hour,$mday,$mon,$year,$wday,$ydat,$isdst) = localtime(time);
 $mon++;
 if ($sec < 10) { $sec = "0$sec"; }
 if ($min < 10) { $min = "0$min"; }
 if ($hour < 10) { $hour = "0$hour"; }
 if ($mday < 10) { $mday = "0$mday"; }
 if ($mon < 10) { $mon = "0$mon"; }
 $year=1900+$year;

 $sys{'sec'}=$sec;
 $sys{'min'}=$min;
 $sys{'hour'}=$hour;
 $sys{'mday'}=$mday;
 $sys{'mon'}=$mon;
 $sys{'year'}=$year;
 $sys{'date_and_time'} = "$mday\.$mon\.$year, $hour:$min $sec s";
 $sys{'date'} = "$mday\.$mon\.$year";
 $sys{'date_us'}= "$mon\.$mday\.$year";
 $sys{'date_all'}="$year\.$mon\.$mday\.$hour\.$min\.$sec";
}



############# dow of today
sub getdow {
#day of week (tt.mm.jjjj)
 my $day="$mday.$mon.$year";

 my @day = ("sun","mon","tue","wed","thu","fri","sat");
 my @d = (0,3,2,5,0,3,5,1,4,6,2,4);
 my @tmp;
 @tmp=split(/\./,$day);
 my $t=$tmp[0];
 my $m=$tmp[1];
 my $j=$tmp[2];
 if ($j < 100) { $j =2000+$j; }
 if ($m < 3) { $j-- ; }
 $tmp[3] = $day[($j+int($j/4)-int($j/100)+int($j/400)+$d[$m-1]+$t) % 7];
 return ($tmp[3]);
}





sub date2valid {
##############
 # parameter: datum ("tt.mm.jjjj")
 # return datum oder -1
 my $d=$_[0];
 $d=~s/ +//g;
 my @tmp1=split(/\./,$d);
 my $t=$tmp1[0];
 my $m=$tmp1[1];
 my $j=$tmp1[2];

 if (($j >= 100) && ($j < 1900)) { return (0); }
 if ($j eq "") { $j=$sys{'year'}; }
 if ($j < 100) { $j =2000+$j; }


 if (($m > 12) || ($m < 1)) { return (0); }
 if (($t > 31) || ($t < 1) ) { return (0); }

  # if ($j eq "2099") {   &mess("$t.$m.$j"); }

 my $tmp3=&schaltjahr($j);
 $tmp3=$tmp3+28;
 if (($m == 2) && ($t > $tmp3) ) { return (0); }
 if ($m < 10) { $m=~s/^0+//;  $m="0$m"; }
 if ($t < 10) { $t=~s/^0+//; $t="0$t"; }



 return("$t\.$m\.$j");
}


#############
sub date2dow {
#day of week (tt.mm.jjjj)
 if (!(&date2valid($_[0]))) { return("--"); }
 my @day = ("So","Mo","Di","Mi","Do","Fr","Sa");
 my @d = (0,3,2,5,0,3,5,1,4,6,2,4);
 my @tmp;
 @tmp=split(/\./,$_[0]);
 my $t=$tmp[0];
 my $m=$tmp[1];
 my $j=$tmp[2];
 if ($j < 100) { $j =2000+$j; }
 if ($m < 3) { $j-- ; }
 $tmp[3] = $day[($j+int($j/4)-int($j/100)+int($j/400)+$d[$m-1]+$t) % 7];
 return ($tmp[3]);
}


sub schaltjahr {
##############
# (jahr)  ja = 1, nein = 0
 if ($_[0] < 100) { $_[0] =2000+$_[0]; }
 if (!($_[0] % 4)) {                        # schaltjahr 4 teilbar > ja aber
        if (!($_[0] % 100)) {                 # 100 teilbar >  nein aber
                if (!($_[0] % 400)) {         # 400 teilbar >  doch ja
                           return ("1");
                } else {
                           return ("0");
                }
        } else {
           return ("1");
        }
 } else {
        return ("0");
 }
}



sub date2doy {        #day of year -  konvertiert datum in eine zahl von tagen seit 1.1.des Jahres;
##############
 # parameter: datum ("tt.mm.jjjj")
 my $x=&date2valid($_[0]);
 if (!($x)) {  return("-1"); }
 my @tmp1=split(/\./,$x);
 my $t=$tmp1[0];
 my $m=$tmp1[1];
 my $j=$tmp1[2];
 my $tmp3=&schaltjahr($j);
 $tmp3=$tmp3+28;
 my @md = (31,$tmp3,31,30,31,30,31,31,30,31,30,31);
 my $days=0;
 my $n;
 for ($n=0;$n<$m-1;$n++) {
        $days=$days+$md[$n];         # monate
 }
 $days=$days+$t;
 return($days)
}


##############  konvertiert datum in eine zahl von tagen seit 1.1.1900;
sub date2dop {
############## day of period
 # parameter: datum ("tt.mm.jjjj")
 my $x=&date2valid($_[0]);
 if (!($x)) {  return("-1"); }
 my @tmp1=split(/\./,$x);
 my $t=$tmp1[0];
 my $m=$tmp1[1];
 my $j=$tmp1[2];
 if ($j < 1900) {  return("-1"); }
 my $days=0;
 my $n;
 for ($n=1900;$n<$j;$n++) {
        $x=&schaltjahr($n);
        $days=$days+365+$x;         # jahre
 }
 $x=&date2doy($_[0]);
 $days=$days+$x;
 return($days)
}

sub date2diff {        #day of period -  Datumsdifferenz in Tagen;
##############
 # parameter: datum ("tt.mm.jjjj","tt.mm.jjjj")
 my $x=&date2valid($_[0]);
 if (!($x)) {  return("d1error"); }
 my $y=&date2valid($_[1]);
 if (!($y)) {  return("d2error"); }

 my $d=&date2dop($y)-&date2dop($x);

 return($d);
}

sub date2add {        #  Datum in n Tagen;
##############
 # parameter: datum ("tt.mm.jjjj","n")
 my $x=&date2valid($_[0]);
 if (!($x)) {  return("-1"); }
 if (!(($_[1] >= 0) && ($_[1] < 300000))) {  return("-1"); }
 my @tmp1=split(/\./,$x);
 my $t=$tmp1[0];
 my $m=$tmp1[1];
 my $j=$tmp1[2];
 my @md = (31,(&schaltjahr($j)+28),31,30,31,30,31,31,30,31,30,31);
 my $n;
 for ($n=0;$n<$_[1];$n++) {
        $t++;
        if ($t > $md[($m-1)]) {
                $t =1;
                $m++;
        }
        if ($m eq "13") {
                $t =1;
                $m =1;
                $j++;
                 @md = (31,(&schaltjahr($j)+28),31,30,31,30,31,31,30,31,30,31);
        }
 }
 return(&date2valid("$t.$m.$j"));
}

sub date2sub {        #  Datum vor n Tagen;
##############
 # parameter: datum ("tt.mm.jjjj","n")
 my $x=&date2valid($_[0]);

 if (!($x)) {  return("-1"); }
 if (!(($_[1] >= 0) && ($_[1] < 300000))) {  return("-1"); }
 my @tmp1=split(/\./,$x);
 my $t=$tmp1[0];
 my $m=$tmp1[1];
 my $j=$tmp1[2];
 my @md = (31,(&schaltjahr($j)+28),31,30,31,30,31,31,30,31,30,31);
 my $n;
 for ($n=0;$n<$_[1];$n++) {
        $t--;
        if ($t==0) {
                $m--;
                if ($m==0) {
                        $m =12;
                        $j--;
                }
                $t=$md[($m-1)];
        }

 }
# return("$t.$m.$j");
 return(&date2valid("$t.$m.$j"));
}

sub date2ger {        #parameter jjjj.mm.tt xxx>> tt.mm.jjjj xxx (z.b. 01.12.2000 18:30 Uhr)
############
my $j;
my $m;
my $t;
my $opt=$_[0]; $opt=~s/^\d\d\d\d\.\d*\.\d*//;
my $date=$_[0]; $date=~s/ .*//;
my @werte=split(/\./,$date);
$j=$werte[0]; $m=$werte[1]; $t=$werte[2];
if (($t>0) && ($m>0) && ($j>1900)) { return ("$t.$m.$j$opt"); } else { return($_[0]); }
}

sub date2us {        #parameter  tt.mm.jjjj xxx >> jjjj.mm.tt xxx
#############
my $j;
my $m;
my $t;

my $opt=$_[0]; $opt=~s/\d*.\d*.\d\d\d\d//;
my $date=$_[0]; $date=~s/ .*//;
my @werte=split(/\./,$date);
$j=$werte[2]; $m=$werte[1]; $t=$werte[0];
if (($t>0) && ($m>0) && ($j>1900)) { return ("$j.$m.$t$opt"); } else { return($_[0]); }
}



sub parse_date { # ex  Jan  1:18 -> Jan 01:18 or  Tue  1  -> Tue 01 ex in zfs creation
#############
  # fix date problems
  # Problem creation date ex  Tue Oct  7  1:1  2014 (usually Tue Oct 07 01:01 2014), but problem on some configs
  # or date strings like Mo. Sep 29  0:48 2014

  my $s=$_[0];
  my @m=('Jan','Feb','Mar','Apr','May','Mai','Maj','Jun','Jul','Aug','Sep','Oct','Okt','Nov','Dec','Dez');
  for (my $i=0;$i<=9; $i++) { $s=~s/  $i:/ 0$i:/gs; }
  foreach my $m (@m) {  for (my $i=1;$i<=9; $i++) { $s=~s/ $m  $i / $m 0$i /igs; }  }
  return($s);

}



#sonstige funktionen
####################

sub SendSendmail {
 #&SendSendmail("$mailapp","from","to","subject","message");

 if (!(defined $_[0])){ $_[0]=""; }
 if (!(defined $_[1])){ $_[1]=""; }
 if (!(defined $_[2])){ $_[2]=""; }
 if (!(defined $_[3])){ $_[3]=""; }
 if (!(defined $_[4])){ $_[4]=""; }

 if (($_[0] eq "") || ($_[1] eq "") || ($_[2] eq "") || ($_[3] eq "") || ($_[4] eq "")) { return ("-1"); }
 if (-f $_[0]) {
  open (MAIL,"|$_[0] -t") || (return ("-2"));
  print MAIL "To: $_[2]\n";                        #webmaster
  print MAIL "From: $_[1]\n";
  print MAIL "Subject: $_[3]\n";
  #print MAIL "Message:\n\n";
  print MAIL "$_[4]\n\n";
  close (MAIL);
   return ("1");
 } else {
   return ("-3");
 }
}






sub filelock {
##############
#parameter datei z.B. /path/datei.txt
#datensatzsperre
#setze seite.plog:               plog=pre_lock
#id
#time() = sekunden seit 1.1.70

      my $time=time();
      my @fin=();
      my $t1;
      my $t2;

      my $dbfile="$sys{'cgipath'}"."$cgicms/pages/$in{'bereich'}/$in{'abteilung'}/$in{'seite'}";
      if ($in{'seite'} =~/^_sys/) {$dbfile=~s/$in{'bereich'}//; $dbfile=~s/$in{'abteilung'}//;} $dbfile=&name2file($dbfile);        #damit _sys.. gefunden wird
      if ($in{'seite'} =~/^_org/) {$dbfile=~s/$in{'abteilung'}//;} $dbfile=&name2file($dbfile);        #damit _org.. gefunden wird

      my $file="$dbfile".".plog";
      my $seite=$dbfile; $seite=~s/.*\///; $seite=~s/^..--//;
      $dbfile.=".txt";

      if (!(-f $dbfile)) { return; }

#1 keine sperre
      if (!(-f $file)) {
            open (PF, ">$file") || &error("(lib1 2778) Datensatzsperre (Seite $seite) konnte nicht erstellt werden");
                  print PF "$session{'id'}\n";
                  print PF "$time\n";
            close (PF);
      }

#2 einlesen

       open (PR, "<$file") || &error("(lib1 2779) Datensatzsperre (Seite $seite) konnte nicht gelesen werden");
          @fin = <PR>;
       close (PR);

      chomp($fin[0]); chomp($fin[1]);

#3 zu alt: neue sperre
if (($time > ($fin[1] +30)) || ($time < $fin[1])) {                        #abgelaufen nach 30s
        @fin=();
        open (PF, ">$file") || &error("(lib1 2791) Datensatzsperre (Seite $seite) konnte nicht erstellt werden");
            print PF "$session{'id'}\n";
            print PF "$time\n";
        close (PF);
        #und zur Kontrolle wieder lesen
        open (PR, "<$file") || &error("(lib1 2796) Datensatzsperre (Seite $seite) konnte nicht gelesen werden");
            @fin = <PR>;
        close (PR);
        chomp($fin[0]); chomp($fin[1]);
}

if ($fin[0] eq "$session{'id'}") {                #ok
        if ($time > ($fin[1] + 10)) {                #neu setzen
                open (PF, ">$file") || &error("(lib1 2778) Datensatzsperre (Seite $seite) konnte nicht erstellt werden");
                    print PF "$session{'id'}\n";
                    print PF "$time\n";
                close (PF);
        }

        return;
}
        $tmp=substr($session{'id'},0,2);
        if ($in{'bereich'} ne "demo") {
          &error("Die Seite $seite wird gerade von $tmp bearbeitet.<br>Bitte versuchen Sie es etwas sp\&auml\;ter nochmal!","mess");
        }
}        #/eof lock



sub radio {
########### formelement parameter: 'name','values (a,b,c)','selected (b)','options (long)'      or tablist or newlinelist
        #                           0      1                2              3
            my ($r,$v,$s,$n,$i,$c,$l);
            $n=$_[0]; $n=~s/^\s+//;

            $v=$_[1]; $v=~s/^\s+//;           # parse values
                                             # if values = "aa ....\nbb ... -> values =aa,bb  text=aa ..  and  bb ...


            if ($v=~/\n/s) {                # each value in a line
                 $v=~s/\t/ /gs;
                 $v=~s/\n/\t/gs;
                 $v=~s/,/ /gs;
            }
           $v=~s/,/\t/gs;                 # tablist


            $s=$_[2];  #selected

            if ($n eq "")  { return ('error in formelement radio: element needs name'); }
            if ($v eq "")  { return ('error in formelement radio $n: element needs values to select from'); }

            local @werte=split(/\t/,$v);
            foreach $i (@werte) {

                     $c="";
                     $l=$i;
                     if (!($_[3]=~/long/)) {
                             $i=~s/ .*//;            #value of 'xx additional text' is xx
                     }
                     if (($s ne "") && ($s eq $i))     { $c=" checked=\"checked\""; }

                     $r.="\n        <input type=\"radio\" name=\"$n\" value=\"$i\"$c> $l <br>";
            }
            return $r;
}

###########################   check if all needed local/remote processes of a job are running  -> needed for ZFS replication in grouplib, jobslib and auto.pl
sub check_running_ps  {        # hide:
###########################   parameter host or none (localhost), pslist like xxx ++ xyz or jobid   ->  return none,partly,all
    my $host=$_[0];
    my $ps=$_[1];             # local/remote pslist  or jobid (read processlist from this file)
    my $opt=$_[2];            # option: kill
    my $ref=$_[3];            # referrer for debugging

    my $t="";
    my %par;

    # default: localhost
    if ($host eq "")     {
       $host=&socks("hostname");
    }
    chomp($host);

    #local or remote?
    @t=();
    &read_dir("C:\xampp\\napp-it\\_log\\group\\","t");               # read grouplist hostname~ip.txt
    my $local="0";
    foreach my $h (@t) {
      if ($h eq "$host\~127.0.0.1.txt") { $local="1"; last; }
    }

    #parameter processlist (abc..++def..) or jobid (1234567)
    if ($ps=~/^\d+$/) {                                     # read from jobid
        $t="C:\xampp\\napp-it\\_log\\jobs\\$ps\.pl";                  # ..
        if (-f $t) {                                                       # my $id = "1234567"
             $t=&socks("type $t");                                         # my $src_interface     = "sudo /usr/bin/nc -w 10 192.168.88.23 57705";
                                                                           # my $dest_interface    = "sudo /usr/bin/nc -w 120 -d -l -p 57705";
                                                                           # my $local_running_ps  = "nc .* -p 57705++zfs receive .*backup3/box4-nfs3";
                                                                           # my $remote_running_ps = "nc .*192.168.88.23 53708++zfs send .*1309523188";
                                                                           # -> if not set, create from src interface and from nc .. port and zfs send ..jobid
             my @w=split(/\n/,$t);
             $t="";
             foreach my $w (@w) {
                if (!($w=~/=\s*\"/)) {  next; }                            # not a key=value

                $w=~s/.*\$//;                                              #    my $id = -> id
                $w=~s/\s+=/=/;                                             #    spaces = -> =
                $w=~s/=\s+\"/=/;                                           #    = spaces -> =
                $w=~s/\".*//;                                              #    12345";  -> 12345

                @t=();
                @t=split(/=/,$w);
                $par{$t[0]}=$t[1];
             }


             # remote_running_ps empty?  -> already created jobs missed this entry
             if ($par{'remote_running_ps'} eq "") {
                $par{'remote_running_ps'}=$par{'src_interface'};        # .. port
                $par{'remote_running_ps'}=~s/.* //;                         # port
                if ($par{'remote_running_ps'} > 0) {
                  $par{'remote_running_ps'}="nc .* $par{'remote_running_ps'}\++zfs send .*$par{'id'}";
                } else {
                  $par{'remote_running_ps'}="";
                }
             }

             #&print_array(@w);
             #&print_hash(%par);
        }

        if ($local) {
           $ps=$par{'local_running_ps'};
        } else {
           $ps=$par{'remote_running_ps'};
        }

    }

    my ($r,@l,@ps);
    @l=split(/\+\+/,$ps);                                          # list of needed ps of this job


    if ($local) {
         $r=`sudo ps axw`;
         @ps=split(/\n/,$r);                                      # running ps
    } else {
         $r=&grouplib_ask_remote_pslist("$host","noerr","$ref -> check_running_ps");         # all running zfs/nc remote processes of host
         if ($r=~/error/) { return ($r); }                        # return error
         @ps=split(/;;/,$r);                                      # return value with separator ;;
    }


#test
##  if (@l>0) {         &print_array(@l);         }

   #check if job is running

          my $run="";
          $t="";
          foreach my $l (@l) {           # list of needed processes
                   foreach $p (@ps) {    # running processes
                              if ($l=~/nc /) {
                                       if ($p=~/$l\b/) {
                                            $run.="1"; $t="1";    # searched proc found
                                       }
                              } else {
                                       if ($p=~/$l/)   { $run.="1"; $t="1";  }  # proclist cuts at 80 char, \b not possible
                              }
                              #kill
                              if (($opt=~/kill/) && ($t eq "1"))        {       # kill only local netcat or zfs commands
                                  #print "p=$p\n";
                                  if ($p=~/sudo |grep /)                 { next; }
                                  if ($local) {
                                        if (!($p=~/nc |zfs receive/))       { next; }
                                        &log("check ps: kill $p","lp");
                                        $p=~s/^\s+//;
                                        $p=~s/ .*//;
                                        `kill -9 $p  >/dev/null 2>&1`;                  # kill id
                                   } else {
                                         if (!($p=~/nc |zfs send/))     { next; }
                                         $r=&grouplib_ask_remote_destroy_proc("host","$p","$ref -> check_running_ps");
                                   }
                              }

                    }
                    if ($t eq "") {
                              $run.="0";                       # missing !!
                    }


                    $t="";
         }

#if (!($local)) {
# &mess($run);
#}


 if ($local) {
   &log("\n########## check_running_ps 1996 local ##########\nrun=$run (11=all running)\nhost= $host \nopt=$opt \nps=$ps\n##########\n",'p');
 } else {
   &log("\n########## check_running_ps 1998 remote ##########\nrrun=$run (11=all)\nhost= $host \nopt=$opt \nps=$ps\n##########\n",'p');
 }


# killed
if ($opt=~/kill/) { return ("none"); }

# all/partly
   if ($run=~/1/) {
      if ($run=~/0/) { return ("partly"); }
      return("running");
   }

 # return no success
   return ("none");


}



sub checkbox {          # parameter: 'name','values (a,b,c)','selected (a,b)','options (long)'      or tablist or newlinelist
###########             #             0      1                2                3

                                        # values = 'a text.., b  text' -> returns only a or b
            my ($r,$v,$s,$n,$i,$c,$l);
            local @werte;
            $n=$_[0];
            $n=~s/^\s+//;

            $v=$_[1]; $v=~s/^\s+//;           # parse values
                                             # if values = "aa ....\nbb ... -> values =aa,bb  text=aa ..  and  bb ...


            if ($v=~/\n/s) {                # each value in a line
                 $v=~s/\t/ /gs;
                 $v=~s/\n/\t/gs;
                 $v=~s/,/ /gs;
            }
           $v=~s/,/\t/gs;                 # tablist


            $s=$_[2];  #selected

            $s=~s/[^a-zA-Z0-9\-\_\,\.\+\:\/]//gs;                 # remove special char from reload, allow /:
            $s=~s/\.\+/,/gs;                                      # a.+b.+c -> a,b,c from %in, reload
            $s.=","; $s=~s/,+/,/g;

            if ($n eq "")  { return ('error in formelement checkbox: element needs name'); }
            if ($v eq "")  { return ('error in formelement checkbox $n: element needs values to select from'); }

            @werte=split(/\t/,$v);

            foreach $i (@werte) {
                     $c="";
                     $l=$i;
                     if (!($_[3]=~/long/)) {
                             $i=~s/ .*//;            #value of 'xx additional text' is xx
                     }
                     $i=~s/\//\\\//g;                    # allows :/
                     if (($s ne "") && ($s =~/$i,/))     { $c=" checked=\"checked\""; }
                     $r.="\n        <input type=\"checkbox\" name=\"$n\" value=\"$i\.+\"$c> $l <br>";
            }
            return $r;
}



sub select {        #erzeuge select-form
############  &select(name,werte(Kommaliste),selected,css,"auto",maxwerte,style) -> return htmlcode
# werte auch als z.b. 01_val,   -> <option value="01_val" selected>val</option>
my $werte=$_[1];
$werte=~s/\t/,/gs;
my $sel=$_[2];


if (!(defined $_[0])){ $_[0]=""; }
if (!(defined $_[1])){ $_[1]=""; }
if (!(defined $_[2])){ $_[2]=""; }
if (!(defined $_[3])){ $_[3]=""; }
if (!(defined $_[4])){ $_[4]=""; }

if ($_[0] eq "selvar") {        # seite editieren vorschau
        $werte=~s/\%\%//g;
        $sel=~s/\%\%//g;
}

if ($sel=~/^\%\% /) {
    if ($werte=~/\%\%/)  { $werte=~s/\%\% *//;  $werte=~s/,+/,/; $werte=~s/,$//;  }         # falls %% in Auswahl  z.b. a,b,c,%%
    $werte.=",$sel";                                                                        # damit select vererbt werden knnen
}

local @werte=split(/,/,$werte); my $i=0;
my ($wert,$html,$tmp,$max,$auto,$css,$value,$style);

$style="";   $css=""; $auto="";
if (!(defined $_[6])) { $_[6]=""; }
if ($_[6] ne "") { $style=" style=\"$_[6]\""; }

if (($_[0] eq "") || (@werte==0)) { return (); }

if (!(defined $_[5])){ $_[5]="0"; }
$max=$_[5];
if ((!($max>0)) || ($max>20000)) { $max=20000; }                                            # opt max 20000 werte

if ($_[4] eq "auto") { $auto=" onChange=\"this.form.submit()\""; } else {$auto="";}

if ($_[3] ne "") { $css=" class=\"$_[3]\""; } else { $css=" class=\"select\""; }
$html="\t<select name=\"$_[0]\"$css$auto$style>\n";
        foreach $wert (@werte) {
                $wert=~s/\r|\n//gs;
                chomp($wert); if (($wert=~/^#/) || ($wert eq "")) {next;}
                $value=$wert;

                if (($value=~/^\d+_/) || ($value=~/==/)) {            # 01_text -> text    bzw menue: 01_text.=xx -> text
                      $wert=~s/^\d+_//; $wert=~s/==.*//;
                } else {
                      $value="";
                }

                if ($value eq "") {
                   if ($wert eq "$sel")  { $tmp=" selected"; } else { $tmp=""; }
                } else {
                   if ($value eq "$sel") { $tmp=" selected"; } else { $tmp=""; }
                   $value=" value=\"$value\"";
                }

                $html.="\t\t<option$value$tmp>$wert</option>\n";
                $i++; if ($i>$max) { last; }
        }
        $html.="\t</select>\n";
}
##

###########     return <select>..</select> with <select selected>..</select>
sub sel_selected {    # name, class, list, sel
###########
  my $name=$_[0];
  my $class=$_[1];
  my $list=$_[2];  # A,B,C   or with values A;a,B;b,C;c
  my $sel=$_[3];   # B
#&mess($sel);

  $list=~s/,+$//;
  my @l=split(/,/,$list);
  my $r="<select name=\"$name\" class=\"$class\">";
  my @v;
  foreach my $l (@l) {
      @v=split(/;/,$l);
      if ($v[0] eq $sel) {
          if ($v[1] ne "") {
             $r.="<option value=\"$v[1]\" selected>$v[0]</option>\n";
          } else {
             $r.="<option selected>$v[0]</option>\n";
          }
      } else {
          if ($v[1] ne "") {
             $r.="<option value=\"$v[1]\">$v[0]</option>\n";
          } else {
             $r.="<option>$v[0]</option>\n";
          }
      }
  }
  $r.="</select>\n";
  return ($r);
}

############# validate entry par data, type, options
sub validate {
############
my $data=$_[0];
my $type=$_[1];  # i (integer), i+ (positive), r (real), r+, r+
my $ref=$_[2];   # reference
    #

    if ($data eq "")                { &mess("$txt{'invalid'} $txt{$ref} $data (1 .. n)"); }

    if ($type =~/i/) {
         if ($type =~/\+/) {
              if (!($data > 0))     { &mess("$txt{'invalid'} $txt{$ref} $data (1 .. n)"); }
         }

         if (!($data=~/^[0-9]+$/))  { &mess("$txt{'invalid'} $txt{$ref} $data (0 .. n)"); }
         return($data);
    }
    if ($type =~/r/) {
         $data=~s/\,/./;
         if ($type =~/\+/) {
              if (!($data > 0))     { &mess("$txt{'invalid'} $txt{$ref} $data (0.1 .. n)"); }
         }

         if ($data=~/^\d+$/)        { return($data);   }  # integer

         if (!($data=~/^[0-9]+\.[0-9]+$/))  {
                                    &mess("$txt{'invalid'} $txt{$ref} $data (0 .. n)");
         }

         return($data);
    }
    &mess("invalid type to validate $data");

}

##############
sub ample_state {
##############
# this is now ajax pull data (dec, 8th 2016)


my $default=$_[1];
if ($default ne "") { # request default status
      $default=<<EoF;
 <div id='status'>
    <div id='status_time' style='display:inline;z-index: 111'> PRO Monitor: <i>inactive</i>  &nbsp;</div>
    <div id='status_info' style='display:inline; margin-left:8px'>
      <a title="Pool state"> Pool <img src='/_doc/grey.png' alt='Pool state'>&nbsp;</a>
      &nbsp;<a title="ZFS pool capacity"> Cap <img src='/_doc/grey.png' alt='Capacity'>&nbsp;</a>
      &nbsp;<a title="ZFS disk activity"> Disk <img src='/_doc/grey.png' alt='Disk activity'>&nbsp;</a>
      &nbsp;<a title="Network activity"> Net <img src='/_doc/grey.png' alt='Network activity'>&nbsp;</a>
      &nbsp;<a title="CPU activity"> CPU <img src='/_doc/grey.png' alt='CPU activity'>&nbsp;</a>
      &nbsp;<a title="Job state"> Job <img src='/_doc/grey.png' alt='Job state'>&nbsp;</a>
    </div>
 </div>
EoF
      return ($default);
}


## or reload
my $par=$_[0];
my $timer="1000";      # 1s


my $reload=<<EoF;

 <script type="text/javascript">
     function auto_load(){
         \$.ajax({
              url: "/cgi-bin/admin-mon.pl",
              cache: false,
              data: {
                  id : "$in{'id'}",
                  l1 : "$in{'l1'}",
                  l2 : "$in{'l2'}",
                  l3 : "$in{'l3'}",
                  prop : "$in{'prop'}",
                  prop2 : "$in{'prop2'}",
                  list : "$in{'list'}"
              },
              success: function(data){
                  \$("#auto_load_div").html(data);

                  var mess=[];
                  mess = data;

                  var arr=["","","","","","","","","","",""];
                  arr = mess.split("<~>");
                  if (arr[0] != '') { \$('#status_time').html(arr[0]);  }         // time
                  if (arr[1] != '') { \$('#status_info').html(arr[1]);  }         // status line
                  if (arr[2] != '') { \$('#content').html(arr[2]);      }         // div  #content update
                  if (arr[3] != '') { \$('#wsarea').html(arr[3]);       }         // div  #wsarea update
                  if (arr[4] != '') { \$('#wsdata').html(arr[4]);       }         // div  #wsdata update
                  if (arr[5] != '') { rgraph_iostat(arr[5]);           }         // draw rgraph with ws data5
                  if (arr[6] != '') { rgraph_arczilstat(arr[6]);       }         // draw rgraph with ws data
                  if (arr[7] != '') { rgraph_netstat(arr[7]);          }         // draw rgraph with ws data
                  if (arr[8] != '') { rgraph_cpustat(arr[8]);          }         // draw rgraph with ws data
                  if (arr[9] != '') { rgraph_servicestat(arr[9]);      }         // draw rgraph with ws data
              }
         });

     }

     \$(document).ready(function(){
            auto_load(); //Call auto_load() function when DOM is Ready
     });

     //Refresh auto_load() function after $timer milliseconds
     setInterval(auto_load,$timer);
 </script>

EoF


return ($reload);

##################
# old #

# $in{'id'} next session id, add webocketid: S_WID|123|
$par.=":wid=$current{'S_WID'}";

my $host=$ENV{'HTTP_HOST'};
$host=~s/:.*//;

#wss
my $https_port = -f "/var/web-gui/_log/https.443" ? 443 : 82;
my $is_https = exists $ENV{'SERVER_PORT'} && $ENV{'SERVER_PORT'} == $https_port;
my $morbo_tls_port = $is_https ? 63000 : 3000;
my $websocket_scheme = $is_https ? "wss" : "ws";
#/wss

my ($wsmess);
my $socket=<<EoF;

<script type="text/javascript">
window.onload = function() {
    if ("WebSocket" in window)  {
        // var ws = new  WebSocket("ws://$host:3000/echo");  old
        var ws = new  WebSocket("$websocket_scheme://$host:$morbo_tls_port/echo");

        // Incoming messages
        ws.onmessage = function(msg) {
          // split msg to status,main,..
          //  \$('#logws').html(msg.data);

          var mess = msg.data;
          var arr = mess.split("<~>");
          if (arr[0] != '') { \$('#status_time').html(arr[0]); }
          if (arr[1] != '') { \$('#status_info').html(arr[1]); }
          if (arr[2] != '') { \$('#content').html(arr[2]);     }
          if (arr[3] != '') { \$('#wsarea').html(arr[3]);      }
          if (arr[4] != '') { \$('#wsdata').html(arr[4]);      }
          if (arr[5] != '') { rgraph_iostat(arr[5]);           }         // draw rgraph with ws data5
          if (arr[6] != '') { rgraph_arczilstat(arr[6]);       }         // draw rgraph with ws data
          if (arr[7] != '') { rgraph_netstat(arr[7]);          }         // draw rgraph with ws data
          if (arr[8] != '') { rgraph_cpustat(arr[8]);          }         // draw rgraph with ws data
          if (arr[9] != '') { rgraph_servicestat(arr[9]);      }         // draw rgraph with ws data

          //alert(arr[0]);
        };

        // Outgoing messages
        window.setInterval(function() {
          ws.send("$par");
        }, 1000)
     }
}
</script>
EoF

return ($socket);
}

 ############
 sub perm_read {      # read when perm=700, data also in @t
     ##
     my ($file,$r,$d,$data,@data,$nomess,$perm,$opt);
     $file=$_[0];
     $nomess=$_[1];
     $perm=$_[2];

     if (!($perm=~/\d/)) { $perm="700"; } # default

     $r=`sudo test -f "$file" && echo 1 || echo 2`;  chomp($r);
     if ($r eq "2") {
          if ($nomess eq "") {
            return();
          } else {
            &mess("cannot access file $file");
          }
     }

     `sudo /usr/bin/chmod 755 $file 2>/dev/null`;
     @t=();
     open (F,"<$file");
        flock(F, LOCK_EX);
        @data=<F>;
     close (F);
     `sudo /usr/bin/chmod $perm $file 2>/dev/null`;
     $data=join("",@data);
     return($data);
 }

 ############         # write when file perm=700 (dir=755)
 sub perm_write {
     ##
     my ($file,$r,$dir,$data,$add);
     $file=$_[0];
     $data=$_[1];
     $perm=$_[2];
     $add=$_[3];
     if (!($perm=~/\d/)) { $perm="700"; } # default

     $r=rindex($file,'/');
     $dir=substr($file,0,$r);
     $r=`sudo test -d "$dir" && echo 1 || echo 2`;  chomp($r);
     if ($r eq "2") { &mess("cannot access directory $dir"); }

     `sudo touch $file`;
     `sudo /usr/bin/chmod 766 $file 2>/dev/null`;

     #&mess($data);

     if ($add=~/add/i) {
          open (PF, ">>$file");
            flock(PF, LOCK_EX);
            print PF "$data";
          close (PF);
     } else {
          open (PF, ">$file");
            flock(PF, LOCK_EX);
            print PF "$data";
          close (PF);
     }
     `sudo /usr/bin/chmod $perm $file 2>/dev/null`;
 }

 sub str2hash {
#############           ex %t() opt when calling
  my ($str,$h,@l,@w);
  $str=$_[0];
  $h=$_[1];      # public hash ex t -> %t
  @l=split(/\n/,$str);
  foreach my $l (@l) {
       @w=split(/\|/,$l);
       ${$h}{$w[0]}=$w[1];
  }
}
sub hash2str {
#############
  my ($k,$v,$s);
  my %h = @_;
  while (($k,$v) = each %h) {
      $s.="$k\|$v\|\n";
  }
  return($s);
}

##########
sub socket {
##########

my $input = $_[0];
$input=~s/^ +//;
$input="'".$input."'";

# Erstelle einen Socket, der mit dem Server auf Port 1234 verbindet
my $socket = IO::Socket::INET->new(
    PeerHost => 'localhost',
    PeerPort => '63000',
    Proto    => 'tcp',
) or return("<tt>Kann keine socket Verbindung herstellen: $!</tt>\n");

    # sende sie an den Server
    $socket->syswrite($input);
    # Warte auf die Antwort vom Server
    my $output;
    $socket->sysread($output, 16384);
    # Gib die Antwort aus
    # print "Antwort vom Server: $output";

    # Schliee den Socket
    $socket->close or die "close: $!";
    return ($output);
}

#return true
1;